/*
 * $Id: BaseLauncher.java,v 1.39 2005/11/20 14:23:49 lehmann Exp $
 * 
 * Copyright (C) 2005 Karsten Lehmann
 * 
 * Software Engineering Chair, Saarland University
 * Prof. Dr. Andreas Zeller
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 *
 * Created on 28.07.2005 14:15:39 by
 * @author Karsten Lehmann 
 * Mail to: lehmann@cs.uni-sb.de or KarstenLehmann@web.de
 */
package org.deltadebugging.ddstate.launcher.client;

import java.util.Date;

import org.deltadebugging.ddcore.DD;
import org.deltadebugging.ddcore.DeltaSet;
import org.deltadebugging.ddstate.comparator.exceptions.GraphComparatorException;
import org.deltadebugging.ddstate.extractor.GraphExtractor;
import org.deltadebugging.ddstate.extractor.GraphExtractorSemaphore;
import org.deltadebugging.ddstate.extractor.exceptions.GraphExtractorException;
import org.deltadebugging.ddstate.extractor.exceptions.WaitTimeExpiredException;
import org.deltadebugging.ddstate.launcher.BreakpointHitListenerAdapter;
import org.deltadebugging.ddstate.launcher.DDStateLaunchPlugin;
import org.deltadebugging.ddstate.launcher.bcel.instrumenter.ModifiedJarCreationException;
import org.deltadebugging.ddstate.launcher.client.exception.DeltaComputationException;
import org.deltadebugging.ddstate.launcher.client.exception.FailingRunPassedException;
import org.deltadebugging.ddstate.launcher.client.exception.PassingRunFailedException;
import org.deltadebugging.ddstate.launcher.client.testrunner.TestRunnerJob;
import org.deltadebugging.ddstate.launcher.ddrun.DDRunResult;
import org.deltadebugging.ddstate.launcher.ddrun.IDDLaunchDescription;
import org.deltadebugging.ddstate.launcher.ddrun.IDDRunDescription;
import org.deltadebugging.ddstate.launcher.ddrun.IDDRunResult;
import org.deltadebugging.ddstate.launcher.ddrun.IDDRunStatusCodes;
import org.deltadebugging.ddstate.launcher.matcher.IMatchPointFinder;
import org.deltadebugging.ddstate.launcher.tester.DeltaProcessor;
import org.deltadebugging.ddstate.launcher.tester.MemGraphDeltaTester;
import org.deltadebugging.ddstate.launcher.tester.exception.TesterException;
import org.deltadebugging.ddstate.launcher.tester.predicate.IPredicate;
import org.deltadebugging.ddstate.logger.DDLogger;
import org.deltadebugging.ddstate.memorygraph.IRootNode;
import org.deltadebugging.ddstate.memorygraph.util.MemoryGraphUtil;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.NullProgressMonitor;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.SubProgressMonitor;
import org.eclipse.debug.core.DebugException;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.debug.core.IJavaBreakpoint;
import org.eclipse.jdt.debug.core.IJavaBreakpointListener;
import org.eclipse.jdt.debug.core.IJavaDebugTarget;
import org.eclipse.jdt.debug.core.IJavaThread;
import org.eclipse.jdt.debug.core.JDIDebugModel;
import org.eclipse.jdt.launching.IVMInstall;
import org.eclipse.swt.widgets.Display;

/**
 * COMMENTME Comment type DDBaseLauncher
 * 
 * @author <a href="mailto:KarstenLehmann@web.de">Karsten Lehmann </a>
 */
public abstract class BaseLauncher extends AbstractLauncher {

    /** Personal logger for this class */
    private final static DDLogger logger = DDStateLaunchPlugin
            .getLogger(BaseLauncher.class);

    public BaseLauncher(IMatchPointFinder matchPointFinder) {
        super(matchPointFinder);
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.deltadebugging.ddstate.launcher.client.IDDLauncher#start(org.deltadebugging.ddstate.launcher.ddrun.IDDLaunchDescription,
     *      org.eclipse.core.runtime.IProgressMonitor)
     */
    public void start(IDDLaunchDescription ddLaunchDesc,
            IProgressMonitor monitor) {

        if (monitor == null)
            monitor = new NullProgressMonitor();
        
        if (ddLaunchDesc == null)
            throw new IllegalArgumentException(
                    "Given dd launch description must not be null");

        this.setLaunchDescription(ddLaunchDesc);

        int ddLaunchID = this.getNextDDLaunchID();
        this.getLaunchDescription().setDDLaunchID(ddLaunchID);

        logger.traceAndTimeStart(DDStateLaunchPlugin.DEBUG, "" + ddLaunchID,
                "Delta Debugging Launch (ID = " + ddLaunchID + ")");

        IJavaBreakpoint[] userBreakpoints = null;

        // Enable the breapoint point manager if it is disabled
        boolean bpManagerEnabled = DDStateLaunchPlugin.getBreakpointManager()
                .isEnabled();
        if (!bpManagerEnabled) {
            DDStateLaunchPlugin.getBreakpointManager().setEnabled(true);
        }

        IVMInstall vmInstall = ddLaunchDesc.getVMInstall();

        try {
            monitor.beginTask("Start DDstate ...", 100);
            
            // 
            // check if modified runtime jar exists, if not create it
            //
            if (!monitor.isCanceled()) {
                try {
                    boolean result = RuntimeJarUtil.checkAndCreate(vmInstall,
                            new SubProgressMonitor(monitor, 10));

                    // the runtime jar could not succesful created or the user
                    // canceld the creation
                    if (result == false) {
                        monitor.setCanceled(true);
                    }

                } catch (ModifiedJarCreationException e) {
                    LaunchExceptionHandler.handleException(e);

                    monitor.setCanceled(true);
                }
            }

            //
            // remember the breakpoints set by the user
            //
            if (!monitor.isCanceled()) {
                try {
                    userBreakpoints = DDStateLaunchPlugin
                            .getBreakpointManager()
                            .disableAllExistingBreakpoints();
                } catch (CoreException e) {
                    LaunchExceptionHandler.handleError(
                            "Cannot disable the breakpoints set by the user.",
                            e);
                    monitor.setCanceled(true);
                }
            }

            //
            // in this method the launcher can implement their
            // different strategies.
            //
            if (!monitor.isCanceled()) {
                this.startLauncherStrategy(new SubProgressMonitor(monitor, 90));
            }

        } finally {
            monitor.done();

            // enable all user breakpoints
            if (userBreakpoints != null) {
                // Outcomment for OOPSLA
//                try {
//                    DDStateLaunchPlugin.getBreakpointManager()
//                            .enabledBreakpoints(userBreakpoints);
//                } catch (CoreException e) {
//                    // do nothing
//                }
            }

            // if the breakpoint manager was disabled disable it again
            if (!bpManagerEnabled) {
                DDStateLaunchPlugin.getBreakpointManager().setEnabled(false);
            }

            logger.traceAndTimeEnd(DDStateLaunchPlugin.DEBUG, "" + ddLaunchID,
                    "Delta Debugging Launch (ID = " + ddLaunchID + ")");
        }
    }

    public abstract void startLauncherStrategy(IProgressMonitor monitor);

    public IDDRunResult computeRelevantDeltas(IDDRunDescription ddRunDesc,
            IProgressMonitor monitor) throws GraphComparatorException,
            DeltaComputationException, TestRunnerException, TesterException {

        IDDRunResult result = null;

        if (ddRunDesc == null)
            throw new IllegalArgumentException(
                    "Given test description must not be null");

        // assign the next available free dd run id
        ddRunDesc.setDDRunID(this.getNextDDRunID());

        //
        // notify all listeners that the delta debugging run has startetd
        //
        DDStateLaunchPlugin.getDDRunListenerManager().notifyDDRunStarted(
                ddRunDesc);

        logger.trace(DDStateLaunchPlugin.DEBUG, "Passing"
                + ddRunDesc.getPassTestDescription().toString());
        logger.trace(DDStateLaunchPlugin.DEBUG, "Failing"
                + ddRunDesc.getFailTestDescription().toString());

        if (monitor == null) {
            monitor = new NullProgressMonitor();
        }


        result = new DDRunResult(ddRunDesc);

        MemGraphDeltaTester tester = null;

        IJavaProject javaProject = ddRunDesc.getPassTestDescription()
                .getJavaProject();
        IVMInstall vmInstall = ddRunDesc.getPassTestDescription()
                .getVMInstall();

        // Create a DDRemoteTestLauncher for the passing and the failing run
        JUnitTestLauncher passDDRemoteLauncher = null;
        JUnitTestLauncher failDDRemoteLauncher = null;

        try {
            monitor.beginTask("Delta Debugging on java states...", 100);

            //
            // start the test launcher for the passing and the failing test run
            //
            if (!monitor.isCanceled()) {
                monitor.subTask("Launch JUnit tests...");
                passDDRemoteLauncher = this.startTestLauncher(javaProject,
                        vmInstall, new SubProgressMonitor(monitor, 0));
                failDDRemoteLauncher = this.startTestLauncher(javaProject,
                        vmInstall, new SubProgressMonitor(monitor, 0));

                try {
                    this
                            .setPassTestRunner(passDDRemoteLauncher
                                    .getTestRunner());
                    this
                            .setFailTestRunner(failDDRemoteLauncher
                                    .getTestRunner());
                } catch (CoreException e) {
                    LaunchExceptionHandler.handleException(e);
                    monitor.setCanceled(true);
                }
                
                monitor.worked(10);
            }

            //
            // compute deltas
            //
            DeltaCompResult deltaCompResult = null;
            if (!monitor.isCanceled()) {
                monitor.subTask("Compute state deltas...");
                logger.trace(DDStateLaunchPlugin.DEBUG, "Compute state deltas...");

                deltaCompResult = this.computeDeltas(result,
                        new SubProgressMonitor(monitor, 20));

                result.setComputedDeltaSet(deltaCompResult.computedDeltaSet);
            }
            
            // shut down the remote test launcher
            if (passDDRemoteLauncher != null)
                passDDRemoteLauncher.shutDown(false);

            if (failDDRemoteLauncher != null)
                failDDRemoteLauncher.shutDown(false);

            //
            // apply deltas
            //
            DeltaSet cMin = null;

            if (!monitor.isCanceled()
                    && deltaCompResult.computedDeltaSet.size() == 0) {

                IStatus status = new Status(IStatus.INFO,
                        DDStateLaunchPlugin.PLUGIN_ID,
                        IDDRunStatusCodes.NO_DELTAS_FOUND, "No deltas found",
                        null);
                result.addStatus(status);
                return result;
            }

            if (!monitor.isCanceled()) {
                logger.trace(DDStateLaunchPlugin.DEBUG, "Minimize state deltas...");

                // create and start the tester:
                // for reverse delta computation use the failing test
                // description and the passing run predicate
                ITestDescription testDesc = ddRunDesc.getPassTestDescription();
                IPredicate testPredicate = deltaCompResult.failRunPredicate;

                tester = new MemGraphDeltaTester(result, // this.getFailTestRunner(),
                        testDesc, deltaCompResult.computedDeltaSet,
                        testPredicate, new SubProgressMonitor(monitor, 0));
                tester.start();
                
                
            }

            if (!monitor.isCanceled()) {
                monitor.subTask("Minimize state deltas...");
                
                // start delta debugging
                DD dd = new DD(tester);
                dd.setAssumeAxiomsHold(true);
                cMin = dd.ddMin(deltaCompResult.computedDeltaSet);

                // evaluate the result of the tester, if something went wrong
                // throw an exception
                this.evaluateTesterResult(tester);
                
                monitor.worked(70);
            }

            // do not set the minized result if either the monitor is canceled,
            // the dd run result has errors, or the dd run result indicates that
            // if all deltas are applied the passing run does not fail
            if (!monitor.isCanceled()
                    && !result.hasErrorStatus()
                    && !result
                            .hasStatus(IDDRunStatusCodes.PASSING_RUN_NOT_FAIL)
                    && !result
                            .hasStatus(IDDRunStatusCodes.FAILING_RUN_NOT_PASS)) {
                logger.trace(DDStateLaunchPlugin.DEBUG, "Minimized Result: "
                        + cMin.toString());

                logger.trace(DDStateLaunchPlugin.DEBUG, "Execute "
                        + tester.getNumberOfExecutedTests() + " test(s)");

                result.setMinimizedDeltaSet(cMin);
            }

        } finally {
            // shut down the remote test launcher
            if (passDDRemoteLauncher != null)
                passDDRemoteLauncher.shutDown(false);

            if (failDDRemoteLauncher != null)
                failDDRemoteLauncher.shutDown(false);

            //
            // look if the run was aborted by the user
            //
            if (!result.hasErrorStatus() && monitor.isCanceled()) {
                IStatus status = new Status(IStatus.INFO,
                        DDStateLaunchPlugin.PLUGIN_ID,
                        IDDRunStatusCodes.TEST_ABORTED,
                        "Monitor canceld - test run aborted.", null);

                result.addStatus(status);
            }

            //
            // set the stop time to the result
            //
            result.setStopTime(new Date(System.currentTimeMillis()));

            // shutdown the tester
            if (tester != null)
                tester.shutDown();

            // monitor done
            monitor.done();
        }

        return result;
    }

    private DeltaSet compareMemoryGraphs(IDDRunResult ddRunResult,
            IRootNode passGraph, IRootNode failGraph, IProgressMonitor monitor)
            throws GraphComparatorException {

        DeltaSet result = null;

        int ddRunID = ddRunResult.getDDRunDescription().getDDRunID();

        result = this.getGraphComparator().compare(ddRunID, passGraph,
                failGraph, monitor);

        if (monitor.isCanceled()) {
            return result;
        }

        return result;
    }

    private class DeltaCompResult {

        public IPredicate failRunPredicate = null;

        public IPredicate passRunPredicate = null;

        public DeltaSet computedDeltaSet = null;

    }

    private DeltaCompResult computeDeltas(IDDRunResult ddRunResult,
            IProgressMonitor monitor) throws GraphComparatorException,
            PassingRunFailedException, FailingRunPassedException,
            TestRunnerException {
        DeltaCompResult result = new DeltaCompResult();

        // Read the passing and the failing test description from the
        // ddRunResult;
        IDDRunDescription ddRunDesc = ddRunResult.getDDRunDescription();
        ITestDescription passTestDesc = ddRunDesc.getPassTestDescription();
        ITestDescription failTestDesc = ddRunDesc.getFailTestDescription();

        // Create a SubProgressMonitor to show the progress of the delta
        // extracting process
        // SubProgressMonitor subMonitor = new SubProgressMonitor(monitor, 1);
        monitor.beginTask("Compute state deltas...", 100);

        // Breakpoints
        IJavaBreakpoint passBP = null;
        IJavaBreakpoint failBP = null;

        // BreakpointListener
        BreakpointHitListenerAdapter passBPListener = null;
        BreakpointHitListenerAdapter failBPListener = null;

        // TestRunnerJobs for the passing and the failing test runner
        TestRunnerJob passTestRunnerJob = null;
        TestRunnerJob failTestRunnerJob = null;

        // Create a semaphore to synchronize the wait of extracting the passing
        // and the wait of extracting failing graph
        GraphExtractorSemaphore semaphore = new GraphExtractorSemaphore();

        try {
            //
            // create and register the breakpoints
            //
            if (!monitor.isCanceled()) {
                monitor.subTask("Create and register breakpoints...");
                logger.trace(DDStateLaunchPlugin.DEBUG,
                        "Create and register breakpoints ...");

                try {
                    passBP = DDStateLaunchPlugin
                            .getBreakpointManager()
                            .createBreakpoint(
                                    passTestDesc.getBreakpointLocation(), false);
                } catch (CoreException e) {
                    logger.log(IStatus.ERROR, "Failed to create breakpoint"
                            + " for the pass breakpoint description", e);
                    throw new TestRunnerException(
                            "Failed to create breakpoint", e);
                }

                try {
                    failBP = DDStateLaunchPlugin
                            .getBreakpointManager()
                            .createBreakpoint(
                                    failTestDesc.getBreakpointLocation(), false);
                } catch (CoreException e) {
                    logger.log(IStatus.ERROR, "Failed to create breakpoint"
                            + " for the fail breakpoint description", e);
                    throw new TestRunnerException(
                            "Failed to create breakpoint", e);
                }

            }

            // 
            // create breakpoint listeners and register them to the
            // JDIDebugModel
            //
            if (!monitor.isCanceled()) {
                monitor.subTask("Register BreakpointListener...");
                logger.trace(DDStateLaunchPlugin.DEBUG,
                        "Register BreakpointListener...");

                passBPListener = this.createBPListener(ddRunResult, passBP,
                        true, semaphore, new SubProgressMonitor(monitor, 30));
                failBPListener = this.createBPListener(ddRunResult, failBP,
                        false, semaphore, new SubProgressMonitor(monitor, 30));

                JDIDebugModel.addJavaBreakpointListener(passBPListener);
                JDIDebugModel.addJavaBreakpointListener(failBPListener);

                monitor.worked(5);
            }

            //
            // extract memory graph
            //
            if (!monitor.isCanceled()) {
                monitor.subTask("Extract memory graphs...");
                logger.trace(DDStateLaunchPlugin.DEBUG,
                        "Extract memory graphs ...");

                passTestRunnerJob = this.launchTestRun(
                        this.getPassTestRunner(), passTestDesc
                                .getJUnitTestDescription(), passBP,
                        IPredicate.STATE_TEST_SUCCESSFUL, monitor);
                failTestRunnerJob = this
                        .launchTestRun(this.getFailTestRunner(), failTestDesc
                                .getJUnitTestDescription(), failBP,
                                IPredicate.STATE_TEST_WITH_FAILURES
                                        | IPredicate.STATE_TEST_WITH_ERRORS,
                                monitor);

                //
                // wait until passing and failing graphs are extraced
                //
                try {
                    semaphore.waitForGraphs(monitor);

                } catch (WaitTimeExpiredException e) {
                    IStatus status = new Status(
                            IStatus.ERROR,
                            DDStateLaunchPlugin.PLUGIN_ID,
                            IDDRunStatusCodes.BP_NOT_HIT,
                            "Wait time expired. Either the breakpoints are not reached "
                                    + "or the memory graphs are not extracted.",
                            e);

                    logger.log(status);

                    monitor.setCanceled(true);
                    ddRunResult.addStatus(status);
                }

                if (passBPListener.hasErros()) {
                    IStatus status = new Status(IStatus.ERROR,
                            DDStateLaunchPlugin.PLUGIN_ID,
                            IDDRunStatusCodes.PASS_GRAPH_EXTRACTION_FAILS,
                            "The extraction of the passing graph fails.",
                            passBPListener.getError());
                    logger.log(status);

                    monitor.setCanceled(true);
                    ddRunResult.addStatus(status);
                }

                if (failBPListener.hasErros()) {
                    IStatus status = new Status(IStatus.ERROR,
                            DDStateLaunchPlugin.PLUGIN_ID,
                            IDDRunStatusCodes.FAIL_GRAPH_EXTRACTION_FAILS,
                            "The extraction of the failing graph fails.",
                            passBPListener.getError());
                    logger.log(status);

                    monitor.setCanceled(true);
                    ddRunResult.addStatus(status);
                }

                monitor.worked(5);
            }

            IRootNode passGraph = semaphore.getPassGraph();
            IRootNode failGraph = semaphore.getFailGraph();

            //
            // notify all listeners that the memory graphs are extracted
            //
            if (!monitor.isCanceled() && passGraph != null
                    && failGraph != null) {
                DDStateLaunchPlugin.getDDRunListenerManager()
                        .notifyMemoryGraphsExtracted(ddRunDesc.getDDRunID(),
                                passGraph, failGraph);
            }

            //
            // compare the graphs
            //
            if (!monitor.isCanceled()) {
                monitor.subTask("Compare memory graphs...");
                logger.trace(DDStateLaunchPlugin.DEBUG,
                        "Compare memory graphs...");

                result.computedDeltaSet = this.compareMemoryGraphs(ddRunResult,
                        passGraph, failGraph, new SubProgressMonitor(monitor, 30));

                if (result != null) {
                    logger
                            .trace(DDStateLaunchPlugin.DEBUG,
                                    result.computedDeltaSet.size()
                                            + " deltas computed");
                }

                // FIXME
                // Test the computed delta set
                if (false) {
                    IJavaThread currentThread = passBPListener
                            .getCurrentThread();

                    DeltaProcessor deltaProcessor = new DeltaProcessor(this
                            .getPassTestRunner());
                    deltaProcessor.setTestSet(result.computedDeltaSet);
                    deltaProcessor.process(currentThread);

                    GraphExtractor graphExtractor = new GraphExtractor();
                    try {
                        int ddRunID = this.getNextDDRunID();

                        // reset informations in the failing graph
                        MemoryGraphUtil.resetCompareInformatios(failGraph);

                        IRootNode testPassGraph = graphExtractor
                                .createMemoryGraph(ddRunID, currentThread,
                                        true, passTestDesc
                                                .getJUnitTestDescription(),
                                        monitor);                        
                        DeltaSet testSet = this.compareMemoryGraphs(
                                new DDRunResult(ddRunResult
                                        .getDDRunDescription()), testPassGraph,
                                failGraph, monitor);

                        DDStateLaunchPlugin.getDDRunListenerManager()
                                .notifyMemoryGraphsExtracted(10001, passGraph,
                                        null);

                        if (testSet.size() > 0) {
                            logger.log(IStatus.ERROR,
                                    "Failed to transfer the memory state. Compute "
                                            + testSet.size()
                                            + " deltas after state transfer.");
                        } else {
                            logger.log(IStatus.OK,
                                    "Transfer of the memory state successful");
                        }
                        monitor.setCanceled(true);

                    } catch (GraphExtractorException e) {
                        throw new RuntimeException(
                                "Failed to check the computed delta set");
                    }

                }

                // subMonitor.worked(1);
            }

            //
            // notify all listeners that the deltas
            //
            if (!monitor.isCanceled() && result != null) {
                DDStateLaunchPlugin.getDDRunListenerManager()
                        .notifyMemoryGraphsCompared(ddRunDesc.getDDRunID(),
                                passGraph, failGraph, result.computedDeltaSet);
            }

        } finally {
            // commit the progress monitor
            monitor.done();

            // unregister and delete the breakpoints
            if (passBP != null) {
                try {
                    this.getPassTestRunner().getDebugTarget()
                            .breakpointRemoved(passBP, null);
                    passBP.delete();
                } catch (CoreException e) {
                    // do nothing
                }
            }

            if (failBP != null) {
                try {
                    this.getFailTestRunner().getDebugTarget()
                            .breakpointRemoved(failBP, null);
                    failBP.delete();
                } catch (CoreException e) {
                    // do nothing
                }
            }

            // resume the current threads
            if (passBPListener != null) {
                try {
                    passBPListener.resumeCurrentThread();
                } catch (DebugException e) {
                    // do nothing
                }
            }

            if (failBPListener != null) {
                try {
                    failBPListener.resumeCurrentThread();
                } catch (DebugException e) {
                    // do nothing
                }
            }

            // evaluate the result of both test runners
            if (!monitor.isCanceled()) {
                this.evaluateTestRunnerResult(passTestRunnerJob,
                        failTestRunnerJob);
            }

            // remove the breakpoint listeners
            JDIDebugModel.removeJavaBreakpointListener(passBPListener);
            JDIDebugModel.removeJavaBreakpointListener(failBPListener);

            // extract the predicates
            result.passRunPredicate = passTestRunnerJob.getTestPredicate();
            result.failRunPredicate = failTestRunnerJob.getTestPredicate();
        }

        return result;
    }

    /**
     * Evaluate the result of the tester.
     * 
     * @param tester
     * @throws TesterException
     *             if the tester execution fails.
     */
    private void evaluateTesterResult(MemGraphDeltaTester tester)
            throws TesterException {
        if (tester == null)
            return;

        IStatus result = tester.getResult();

        switch (result.getCode()) {
        case MemGraphDeltaTester.STATUS_CODE_UNEXPECTED_ERROR:
            throw new TesterException(
                    "Unexpected error on running passing test", result
                            .getException());
        }

    }

    /**
     * Evaluate the result of the pass test runner job and the fail test runner
     * job.
     * 
     * @param passTestRunnerJob
     * @param failTestRunnerJob
     * @throws TestRunnerException
     *             if the pass or the fail test runner job has an error status
     *             and the status code equals
     *             <code>TestRunnerJob.STATUS_CODE_UNEXPECTED_ERROR</code>.
     * @throws PassingRunFailedException
     *             if the pass test runner produce not the expected result.
     * @throws FailingRunPassedException
     *             if the fail test runner produce not the expected result.
     */
    private void evaluateTestRunnerResult(TestRunnerJob passTestRunnerJob,
            TestRunnerJob failTestRunnerJob) throws TestRunnerException,
            PassingRunFailedException, FailingRunPassedException {
        if (passTestRunnerJob == null || failTestRunnerJob == null)
            return;

        // wait until both tests are executed succesfully
        this.waitForTestRunners(passTestRunnerJob, failTestRunnerJob);

        IStatus passRunnerStatus = passTestRunnerJob.getResult();
        IStatus failRunnerStatus = failTestRunnerJob.getResult();

        switch (passRunnerStatus.getCode()) {
        case TestRunnerJob.STATUS_CODE_UNEXPECTED_ERROR:
            throw new TestRunnerException(
                    "Unexpected error on running passing test",
                    passRunnerStatus.getException());
        case TestRunnerJob.STATUS_CODE_UNEXPECTED_TEST_RESULT:
            throw new PassingRunFailedException();
        }

        switch (failRunnerStatus.getCode()) {
        case TestRunnerJob.STATUS_CODE_UNEXPECTED_ERROR:
            throw new TestRunnerException(
                    "Unexpected error on running failing test",
                    failRunnerStatus.getException());
        case TestRunnerJob.STATUS_CODE_UNEXPECTED_TEST_RESULT:
            throw new FailingRunPassedException();
        }
    }

    /**
     * Waits until the pass test runner job and the fail test runner job is
     * finished. This method will block the calling thread until the jobs has
     * finished executing, or until this thread has been interrupted. If the job
     * has not been scheduled, this method returns immediately
     * 
     * @param passTestRunnerJob
     * @param failTestRunnerJob
     * @throws TestRunnerException
     *             if one of the jobs was interrupted.
     */
    private void waitForTestRunners(TestRunnerJob passTestRunnerJob,
            TestRunnerJob failTestRunnerJob) throws TestRunnerException {
        if (passTestRunnerJob == null || failTestRunnerJob == null)
            return;

        try {
            passTestRunnerJob.join();
        } catch (InterruptedException e) {
            throw new TestRunnerException(
                    "Error occured on waiting for finish of the pass test runner job.",
                    e);
        }
        try {
            failTestRunnerJob.join();
        } catch (InterruptedException e) {
            throw new TestRunnerException(
                    "Error occured on waiting for finish of the fail test runner job.",
                    e);
        }

    }

    private BreakpointHitListenerAdapter createBPListener(
            final IDDRunResult ddRunResult, IJavaBreakpoint bp,
            final boolean passingRun, final GraphExtractorSemaphore semaphore,
            final IProgressMonitor monitor) {
        BreakpointHitListenerAdapter result = new BreakpointHitListenerAdapter(
                bp) {
            protected int givenBreakpointHit(final IJavaThread thread,
                    IJavaBreakpoint breakpoint) {

                Display.getDefault().asyncExec(new Runnable() {
                    public void run() {
                        try {
                            if (passingRun) {
                                extractPassingGraph(ddRunResult, thread,
                                        semaphore, monitor);

                            } else {
                                extractFailingGraph(ddRunResult, thread,
                                        semaphore, monitor);
                            }
                        } catch (Throwable t) {
                            setError(t);
                        }

                    }
                });

                return IJavaBreakpointListener.SUSPEND;
            }
        };

        return result;
    }

    private TestRunnerJob launchTestRun(final IExtTestRunner testRunner,
            final IJUnitTestDescription testDesc, IJavaBreakpoint breakpoint,
            int expectedResult, final IProgressMonitor monitor) {
        TestRunnerJob result = null;

        String jobName = "... Run " + testDesc.getTestName();

        // install the breakpoints in the debug target
        IJavaDebugTarget debugTarget = testRunner.getDebugTarget();
        debugTarget.breakpointAdded(breakpoint);

        if (!monitor.isCanceled()) {

            // Thread the call
            result = new TestRunnerJob(jobName, testRunner, testDesc);
            result.setSystem(true);
            result.installPostLaunchCheck(expectedResult);

            result.schedule();
        }

        return result;
    }

    void extractFailingGraph(IDDRunResult ddRunResult, IJavaThread thread,
            GraphExtractorSemaphore semaphore, IProgressMonitor monitor)
            throws GraphExtractorException {
        IDDRunDescription ddRunDesc = ddRunResult.getDDRunDescription();
        int ddRunID = ddRunDesc.getDDRunID();

        IRootNode failGraph = null;
        try {
            failGraph = new GraphExtractor().createMemoryGraph(ddRunID, thread,
                    false, ddRunDesc.getFailTestDescription()
                            .getJUnitTestDescription(), monitor);

        } finally {
            semaphore.failGraphExtracted(failGraph);
        }
    }

    void extractPassingGraph(IDDRunResult ddRunResult, IJavaThread thread,
            GraphExtractorSemaphore semaphore, IProgressMonitor monitor)
            throws GraphExtractorException {
        IDDRunDescription ddRunDesc = ddRunResult.getDDRunDescription();
        int ddRunID = ddRunDesc.getDDRunID();

        IRootNode passGraph = null;
        try {
            passGraph = new GraphExtractor().createMemoryGraph(ddRunID, thread,
                    true, ddRunDesc.getPassTestDescription()
                            .getJUnitTestDescription(), monitor);
        } finally {
            semaphore.passGraphExtracted(passGraph);
        }
    }

    JUnitTestLauncher startTestLauncher(final IJavaProject javaProject,
            final IVMInstall vmInstall, final IProgressMonitor monitor) {
        JUnitTestLauncher result = new JUnitTestLauncher();

        try {
            result.start(javaProject, vmInstall, monitor);
        } catch (CoreException e) {
            LaunchExceptionHandler.handleException(e);
            monitor.setCanceled(true);
        }

        return result;
    }

}
