/*
 * $Id: LaunchExceptionHandler.java,v 1.13 2005/08/15 12:24:48 lehmann Exp $
 * 
 * Copyright (C) 2005 Karsten Lehmann
 * 
 * Software Engineering Chair, Saarland University
 * Prof. Dr. Andreas Zeller
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 *
 * Created on Aug 4, 2005 12:14:28 PM by
 * @author Karsten Lehmann 
 * Mail to: lehmann@cs.uni-sb.de or KarstenLehmann@web.de
 */
package org.deltadebugging.ddstate.launcher.client;

import org.deltadebugging.ddstate.comparator.exceptions.GraphComparatorException;
import org.deltadebugging.ddstate.launcher.DDStateLaunchPlugin;
import org.deltadebugging.ddstate.launcher.bcel.instrumenter.ModifiedJarCreationException;
import org.deltadebugging.ddstate.launcher.client.exception.DeltaComputationException;
import org.deltadebugging.ddstate.launcher.client.exception.FailingRunPassedException;
import org.deltadebugging.ddstate.launcher.client.exception.PassingRunFailedException;
import org.deltadebugging.ddstate.launcher.matcher.exception.MatchPointComputationException;
import org.deltadebugging.ddstate.launcher.tester.exception.TesterException;
import org.deltadebugging.ddstate.launcher.tester.exception.TesterStartException;
import org.deltadebugging.ddstate.logger.DDLogger;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;

/**
 * COMMENTME Comment type LaunchExceptionHandler
 * 
 * @author <a href="mailto:KarstenLehmann@web.de">Karsten Lehmann </a>
 */
public class LaunchExceptionHandler {

    private final static String ERROR_DIALO_SUFFIX = "\n\nFor more information take a look in the error log view";

    // Unexpected Throwable
    private final static String EXMSG_UNEXPECTED_THROWABLE = "Unexcepted exception occured.";

    // ModifiedJarCreationException
    private final static String EXMSG_MOD_JAR_OOM = "Not enough memory to create the modified runtime jar. "
            + "Please increase the initial memory of your eclipse application.";

    private final static String EXMSG_MOD_JAR = "Error on modification of the java runtime jar.";

    // MatchPointComputationException
    private final static String EXMSG_MP_COMPUTATION = "Error on computation the match points "
            + "of the passing and the failing run.";

    // GraphComparatorException
    private final static String EXMSG_GRAPH_COMPARATOR = "Error on comparing memory graphs.";

    // TestRunnerException
    private final static String EXMSG_TEST_RUNNER = "Error on executing test run.";

    // DeltaComputationException
    private final static String EXMSG_DELTA_COMP = "Unexpected error on delta computation.";

    private final static String EXMSG_PASSING_RUN_FAIL = "Given passing run fails.";

    private final static String EXMSG_FAILING_RUN_PASS = "Given failing run pass.";

    // TesterException
    private final static String EXMSG_TESTER = "Unexcpected error on running delta tester.";

    private final static String EXMSG_TESTER_START = "Cannot start the tester to apply the computed deltas.";

    /** Personal logger for this class */
    private final static DDLogger logger = DDStateLaunchPlugin
            .getLogger(LaunchExceptionHandler.class);

    // This class is a util class - use the provided static methods.
    private LaunchExceptionHandler() {
        super();
    }

    /**
     * Log the given error message and exception to the eclipse error log and
     * trace it to the console. After that present the user an error dialog with
     * the given error message and a hint to look at the eclipse error log view
     * for more information.
     * 
     * @param errorMessage
     *            the error message you want to present the user
     * @param excpetion
     *            the exception occured
     * @return the status object logged to the console and displayed to the
     *         user.
     */
    public static IStatus handleError(String errorMessage, Throwable excpetion) {

        IStatus result = new Status(IStatus.ERROR,
                DDStateLaunchPlugin.PLUGIN_ID, IStatus.OK, errorMessage,
                excpetion);
        logger.log(result);

        errorMessage += ERROR_DIALO_SUFFIX;
        result = new Status(IStatus.ERROR, DDStateLaunchPlugin.PLUGIN_ID,
                IStatus.OK, errorMessage, excpetion);
        DDStateLaunchPlugin.showErrorDialog(result,
                "Error on Delta Debugging Run");

        return result;
    }

    /**
     * Log the given info message to the console and present the user an
     * inforamtion dialog with the given message.
     * 
     * @param infoMessage
     *            the message you want to present the user.
     */
    public static void handleInformation(String infoMessage) {
        IStatus status = new Status(IStatus.INFO,
                DDStateLaunchPlugin.PLUGIN_ID, IStatus.OK, infoMessage, null);

        logger.trace(DDStateLaunchPlugin.DEBUG, infoMessage);
        DDStateLaunchPlugin.showErrorDialog(status,
                "Delta Degugging Run Information");
    }

    public static void handleException(Throwable t) {
        if (t instanceof ModifiedJarCreationException) {
            handleModifiedJarCreationException((ModifiedJarCreationException) t);

        } else if (t instanceof MatchPointComputationException) {
            handleMatchPointComputationException((MatchPointComputationException) t);

        } else if (t instanceof GraphComparatorException) {
            handleGraphComparatorException((GraphComparatorException) t);

        } else if (t instanceof DeltaComputationException) {
            handleDeltaComputationException((DeltaComputationException) t);

        } else if (t instanceof TestRunnerException) {
            handleTestRunnerException((TestRunnerException) t);

        } else if (t instanceof TesterException) {
            handleTesterException((TesterException) t);

        } else if (t instanceof CoreException) {
            handleCoreException((CoreException) t);

        } else {
            handleUnexpectedThrowable(t);

        }

    }

    private static void handleTesterException(TesterException e) {
        if (e instanceof TesterStartException) {
            LaunchExceptionHandler.handleError(EXMSG_TESTER_START, e);
        } else {
            LaunchExceptionHandler.handleError(EXMSG_TESTER, e);
        }
    }

    private static void handleDeltaComputationException(
            DeltaComputationException e) {

        if (e instanceof PassingRunFailedException) {
            LaunchExceptionHandler.handleError(EXMSG_PASSING_RUN_FAIL, null);
        } else if (e instanceof FailingRunPassedException) {
            LaunchExceptionHandler.handleError(EXMSG_FAILING_RUN_PASS, null);
        } else {
            LaunchExceptionHandler.handleError(EXMSG_DELTA_COMP, e);
        }

    }

    private static void handleTestRunnerException(TestRunnerException e) {
        LaunchExceptionHandler.handleError(EXMSG_TEST_RUNNER, e);
    }

    private static void handleGraphComparatorException(
            GraphComparatorException e) {
        LaunchExceptionHandler.handleError(EXMSG_GRAPH_COMPARATOR, e);
    }

    private static void handleMatchPointComputationException(
            MatchPointComputationException e) {
        LaunchExceptionHandler.handleError(EXMSG_MP_COMPUTATION, e);
    }

    private static void handleCoreException(CoreException e) {
        IStatus status = e.getStatus();
        DDStateLaunchPlugin.showErrorDialog(status,
                "Error on Delta Debugging Run");
    }

    private static void handleUnexpectedThrowable(Throwable t) {
        LaunchExceptionHandler.handleError(EXMSG_UNEXPECTED_THROWABLE, t);
    }

    private static void handleModifiedJarCreationException(
            ModifiedJarCreationException e) {
        if (e.getCause() instanceof OutOfMemoryError) {
            LaunchExceptionHandler.handleError(EXMSG_MOD_JAR_OOM, e.getCause());
        } else {
            LaunchExceptionHandler.handleError(EXMSG_MOD_JAR, e);
        }
    }
}
