/*
 * $Id: CTSRunListenerManager.java,v 1.2 2005/08/17 17:25:45 lehmann Exp $
 * 
 * Copyright (C) 2005 Karsten Lehmann
 * 
 * Software Engineering Chair, Saarland University
 * Prof. Dr. Andreas Zeller
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 *
 * Created on 10.04.2005 17:12:48 by
 * @author Karsten Lehmann 
 * Mail to: lehmann@cs.uni-sb.de or KarstenLehmann@web.de
 */
package org.deltadebugging.ddstate.launcher.client.cts;

import org.deltadebugging.ddstate.DDStatePlugin;
import org.deltadebugging.ddstate.launcher.DDStateLaunchPlugin;
import org.deltadebugging.ddstate.launcher.ddrun.IDDLaunchDescription;
import org.deltadebugging.ddstate.logger.DDLogger;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IConfigurationElement;
import org.eclipse.core.runtime.IExtensionPoint;
import org.eclipse.core.runtime.ISafeRunnable;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Platform;
import org.eclipse.jface.util.ListenerList;

/**
 * The CTSRunListenerManager manage all listeners for the cause transition runs.
 * Also this manager informs all listeners if something changed at the cts run.
 * 
 * @author <a href="mailto:KarstenLehmann@web.de">Karsten Lehmann </a>
 */
public class CTSRunListenerManager implements ICTSRunListenerManager {

	/** Personal logger for this class */
	final static DDLogger logger = DDStatePlugin
			.getLogger(CTSRunListenerManager.class);

	/** The list of all cause transition run listeners */
	private ListenerList fCTSRunListener = null;

	/** The instance of this CTSRunListenerManager */
	private static CTSRunListenerManager instance = null;

	public abstract class ListenerSafeRunnable implements ISafeRunnable {
		public void handleException(Throwable t) {
			logger.log(IStatus.ERROR, "Exception on notify listener", t);
		}
	}

	private CTSRunListenerManager() {
		this.loadCTSRunListeners();
	}

	public synchronized static ICTSRunListenerManager getInstance() {
		if (instance == null) {
			instance = new CTSRunListenerManager();
		}

		return instance;
	}

	/**
	 * Initializes CTSRunListener Listener extensions
	 */
	private void loadCTSRunListeners() {
		this.fCTSRunListener = new ListenerList();
		IExtensionPoint extensionPoint = Platform
				.getExtensionRegistry()
				.getExtensionPoint(
						DDStateLaunchPlugin.ID_EXTENSION_POINT_CTSRUN_LISTENERS);
		if (extensionPoint == null) {
			return;
		}
		IConfigurationElement[] configs = extensionPoint
				.getConfigurationElements();

		for (int i = 0; i < configs.length; i++) {
			try {
				ICTSRunListener ctsRunListener = (ICTSRunListener) configs[i]
						.createExecutableExtension("class"); //$NON-NLS-1$
				this.fCTSRunListener.add(ctsRunListener);
			} catch (CoreException e) {
				DDStatePlugin.getLogger(DDStatePlugin.class).log(IStatus.ERROR,
						"Could not load some ctsRunListeners extension points",
						e);
			}
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.launcher.client.cts.ICTSRunListenerManager#addCTSRunListener(org.deltadebugging.ddstate.launcher.client.cts.ICTSRunListener)
	 */
	public void addCTSRunListener(ICTSRunListener listener) {
		if (this.fCTSRunListener == null) {
			this.fCTSRunListener = new ListenerList(5);
		}
		this.fCTSRunListener.add(listener);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.launcher.client.cts.ICTSRunListenerManager#removeCTSRunListener(org.deltadebugging.ddstate.launcher.client.cts.ICTSRunListener)
	 */
	public void removeCTSRunListener(ICTSRunListener listener) {
		if (this.fCTSRunListener != null) {
			this.fCTSRunListener.remove(listener);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.launcher.client.cts.ICTSRunListenerManager#getCTSRunListeners()
	 */
	public ICTSRunListener[] getCTSRunListeners() {
		ICTSRunListener[] result = null;
		Object[] listeners = this.fCTSRunListener.getListeners();

		result = new ICTSRunListener[listeners.length];

		for (int i = 0; i < listeners.length; i++) {
			result[i] = (ICTSRunListener) listeners[i];
		}

		return result;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.launcher.client.cts.ICTSRunListenerManager#notifyCTSRunTerminated(org.deltadebugging.ddstate.launcher.client.cts.ICTSResult)
	 */
	public void notifyCTSRunTerminated(final ICTSRunResult result) {
		ICTSRunListener[] listeners = this.getCTSRunListeners();

		for (int i = 0; i < listeners.length; i++) {
			final ICTSRunListener listener = listeners[i];

			Platform.run(new ListenerSafeRunnable() {
				public void run() {
					listener.ctsRunTerminated(result);
				}
			});

		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.launcher.client.cts.ICTSRunListenerManager#notifyCTSRunStarted(org.deltadebugging.ddstate.launcher.ddrun.IDDLaunchDescription)
	 */
	public void notifyCTSRunStarted(final IDDLaunchDescription desc) {
		ICTSRunListener[] listeners = this.getCTSRunListeners();

		for (int i = 0; i < listeners.length; i++) {
			final ICTSRunListener listener = listeners[i];

			Platform.run(new ListenerSafeRunnable() {
				public void run() {
					listener.ctsRunStarted(desc);
				}
			});

		}
	}

}