/*
 * $Id: MatchPointInterval.java,v 1.5 2005/09/11 13:31:27 lehmann Exp $
 * 
 * Copyright (C) 2005 Karsten Lehmann
 * 
 * Software Engineering Chair, Saarland University
 * Prof. Dr. Andreas Zeller
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 *
 * Created on Aug 10, 2005 8:31:58 PM by
 * @author Karsten Lehmann 
 * Mail to: lehmann@cs.uni-sb.de or KarstenLehmann@web.de
 */
package org.deltadebugging.ddstate.launcher.client.cts;

import org.deltadebugging.ddstate.launcher.matcher.IMatchPoint;

/**
 * The match point interval defines an interval over a match point array.
 * 
 * @author <a href="mailto:KarstenLehmann@web.de">Karsten Lehmann </a>
 */
public class MatchPointInterval {

	/** the match points defining this interval */
	private final IMatchPoint[] fMatchPoints;

	/** the start index in this interval */
	private final int fStartIndex;

	/** the end index in this interval */
	private final int fEndIndex;

	public MatchPointInterval(IMatchPoint[] matchPoints) {

		if (matchPoints == null) {
			throw new IllegalArgumentException(
					"Given matchpoint must not be null.");
		}

		if (matchPoints.length < 2) {
			throw new IllegalArgumentException(
					"The given match point array must contain at least two elements.");
		}

		this.fMatchPoints = matchPoints;

		this.fStartIndex = 0;
		this.fEndIndex = this.fMatchPoints.length - 1;
	}

	private MatchPointInterval(IMatchPoint[] matchPoints, int startIndex,
			int endIndex) {

		if (matchPoints == null) {
			throw new IllegalArgumentException(
					"Given matchpoint must not be null.");
		}

		if (matchPoints.length < 2) {
			throw new IllegalArgumentException(
					"The given match point array must contain at least two elements.");
		}

		if (startIndex < 0 || endIndex < 0) {
			throw new IllegalArgumentException(
					"Given indices must greater equal 0.");
		}

		if (startIndex > endIndex) {
			throw new IllegalArgumentException(
					"Given start index must less or equal than end index.");
		}

		if (endIndex > matchPoints.length - 1) {
			throw new IllegalArgumentException(
					"Given end index is out of range. "
							+ "It must less than the length of the given match point array.");
		}

		this.fMatchPoints = matchPoints;
		this.fStartIndex = startIndex;
		this.fEndIndex = endIndex;

	}

	/**
	 * Returns the match point at the given index in the match point interval.
	 * 
	 * @param index
	 * @return the match point at the given index in the match point interval.
	 */
	public IMatchPoint getMatchPointAtIndex(int index) {
		IMatchPoint result = null;

		if (index < 0) {
			throw new IllegalArgumentException(
					"Given index must greater equal than 0.");
		}

		if (index >= this.length()) {
			throw new IllegalArgumentException(
					"Given index is out of range. "
							+ "It must less than the number of match points containing this interval.");
		}

		result = this.fMatchPoints[this.fStartIndex + index];

		return result;
	}

	/**
	 * Returns the match point at the start position of this interval.
	 * 
	 * @return the match point at the start position of this interval.
	 */
	public IMatchPoint getStartMatchPoint() {
		return this.fMatchPoints[this.fStartIndex];
	}

	/**
	 * Returns the match point of the end position of this interval.
	 * 
	 * @return the match point of the end position of this interval.
	 */
	public IMatchPoint getEndMatchPoint() {
		return this.fMatchPoints[this.fEndIndex];
	}

	/**
	 * Get the match point in the interval for the given ratio. The ratio is a
	 * number between 0 and 1. 0 define the start point in the interval, 1 the
	 * end.
	 * 
	 * @param ratio
	 *            a number between 0 and 1.
	 * @return the matchpoint of this interval defined by the given ratio.
	 */
	public IMatchPoint getMatchPointAt(float ratio) {
		IMatchPoint result = null;

		int index = this.computeIndex(ratio);
		result = this.fMatchPoints[index];

		return result;
	}

	private int computeIndex(float ratio) {
		int result = -1;

		if (ratio < 0 || ratio > 1) {
			throw new IllegalArgumentException(
					"Given ratio must between 0 and 1.");
		}

		int count = this.length() - 1;

		result = this.fStartIndex + Math.round(count * ratio);

		return result;
	}

	/**
	 * Returns true if it is possible two divide this interval in the middle.
	 * That means that this interval must contain at least 3 elements.
	 * 
	 * @return true if it is possible two divide this interval, otherwise false.
	 */
	public boolean canDivideInterval() {
		return this.length() >= 3;
	}

	/**
	 * Returns two match point intervals if it possible to divide this interval.
	 * The first interval range from the start index of this interval to the
	 * computed divider. The second interval range from the divider to the end
	 * index of this interval.
	 * 
	 * @return two match point intervals if it possible to divide this interval,
	 *         otherwise <code>null</code>.
	 * 
	 * @see MatchPointInterval#canDivideInterval()
	 */
	public MatchPointInterval[] divideInTwoIntervals() {
		MatchPointInterval[] result = null;

		if (this.canDivideInterval()) {
			result = new MatchPointInterval[2];

			int divider = this.computeIndex(0.5f);

			assert this.fStartIndex < divider;
			assert divider < this.fEndIndex;

			result[0] = new MatchPointInterval(this.fMatchPoints,
					this.fStartIndex, divider);
			result[1] = new MatchPointInterval(this.fMatchPoints, divider,
					this.fEndIndex);
		}

		return result;
	}

	/**
	 * Returns a sub interval for the given indices.
	 * 
	 * @param startIndex
	 *            the start index of the new subinterval.
	 * @param endIndex
	 *            the end index of the new subninterval.
	 * @return a sub interval for the given indices.
	 */
	public MatchPointInterval getSubInterval(int startIndex, int endIndex) {

		return new MatchPointInterval(this.fMatchPoints, this.fStartIndex
				+ startIndex, this.fStartIndex + endIndex);
	}

	/**
	 * Count all match points in this interval including the start and the end
	 * match point.
	 * 
	 * @return the number of match points in this interval.
	 */
	public int length() {
		return (this.fEndIndex - this.fStartIndex) + 1;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		StringBuffer result = new StringBuffer("MatchPointInterval:\n");
		result.append("\tStart match point (" + this.fStartIndex + "): "
				+ this.fMatchPoints[this.fStartIndex] + "\n");
		result.append("\tEnd match point (" + this.fEndIndex + "): "
				+ this.fMatchPoints[this.fEndIndex]);

		return result.toString();
	}

	public static void main(String[] args) {
		System.out.println(Math.round(0.5 * 4));
	}
}
