/*
 * $Id: DDRunListenerManager.java,v 1.3 2005/08/16 09:14:11 lehmann Exp $
 * 
 * Copyright (C) 2005 Karsten Lehmann
 * 
 * Software Engineering Chair, Saarland University
 * Prof. Dr. Andreas Zeller
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 *
 * Created on 10.04.2005 17:12:48 by
 * @author Karsten Lehmann 
 * Mail to: lehmann@cs.uni-sb.de or KarstenLehmann@web.de
 */
package org.deltadebugging.ddstate.launcher.ddrun;

import org.deltadebugging.ddcore.DeltaSet;
import org.deltadebugging.ddstate.DDStatePlugin;
import org.deltadebugging.ddstate.launcher.DDStateLaunchPlugin;
import org.deltadebugging.ddstate.logger.DDLogger;
import org.deltadebugging.ddstate.memorygraph.IRootNode;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IConfigurationElement;
import org.eclipse.core.runtime.IExtensionPoint;
import org.eclipse.core.runtime.ISafeRunnable;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Platform;
import org.eclipse.jface.util.ListenerList;

/**
 * The DDRunListenerManager manage all listeners for the delta debugging runs.
 * Also this manager informs all listeners if something changed at the delta
 * debugging run.
 * 
 * @author <a href="mailto:KarstenLehmann@web.de">Karsten Lehmann </a>
 */
public class DDRunListenerManager implements IDDRunListenerManager {

	/** Personal logger for this class */
	final static DDLogger logger = DDStatePlugin
			.getLogger(DDRunListenerManager.class);

	/** The list of all delta debugging run listeners */
	private ListenerList fDDRunListener = null;

	/** The instance of this DDRunListenerManager */
	private static DDRunListenerManager instance = null;

	public abstract class ListenerSafeRunnable implements ISafeRunnable {
		public void handleException(Throwable t) {
			logger.log(IStatus.ERROR, "Exception on notify listener", t);
		}
	}

	private DDRunListenerManager() {
		this.loadDDRunListeners();
	}

	public synchronized static IDDRunListenerManager getInstance() {
		if (instance == null) {
			instance = new DDRunListenerManager();
		}

		return instance;
	}

	/**
	 * Initializes DDRunListener Listener extensions
	 */
	private void loadDDRunListeners() {
		this.fDDRunListener = new ListenerList();
		IExtensionPoint extensionPoint = Platform.getExtensionRegistry()
				.getExtensionPoint(
						DDStateLaunchPlugin.ID_EXTENSION_POINT_DDRUN_LISTENERS);
		if (extensionPoint == null) {
			return;
		}
		IConfigurationElement[] configs = extensionPoint
				.getConfigurationElements();

		for (int i = 0; i < configs.length; i++) {
			try {
				IDDRunListener ddRunListener = (IDDRunListener) configs[i]
						.createExecutableExtension("class"); //$NON-NLS-1$
				this.fDDRunListener.add(ddRunListener);
			} catch (CoreException e) {
				DDStatePlugin.getLogger(DDStatePlugin.class).log(IStatus.ERROR,
						"Could not load some ddRunListeners extension points",
						e);
			}
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.launcher.ddrun.IDDRunListenerManager#addDDRunListener(org.deltadebugging.ddstate.launcher.ddrun.IDDRunListener)
	 */
	public void addDDRunListener(IDDRunListener listener) {
		if (this.fDDRunListener == null) {
			this.fDDRunListener = new ListenerList(5);
		}
		this.fDDRunListener.add(listener);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.launcher.ddrun.IDDRunListenerManager#removeDDRunListener(org.deltadebugging.ddstate.launcher.ddrun.IDDRunListener)
	 */
	public void removeDDRunListener(IDDRunListener listener) {
		if (this.fDDRunListener != null) {
			this.fDDRunListener.remove(listener);
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.launcher.ddrun.IDDRunListenerManager#getDDRunListeners()
	 */
	public IDDRunListener[] getDDRunListeners() {
		IDDRunListener[] result = null;
		Object[] listeners = this.fDDRunListener.getListeners();

		result = new IDDRunListener[listeners.length];

		for (int i = 0; i < listeners.length; i++) {
			result[i] = (IDDRunListener) listeners[i];
		}

		return result;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.launcher.ddrun.IDDRunListenerManager#notifyDDRunTerminated(org.deltadebugging.ddstate.launcher.ddrun.IDDRunResult)
	 */
	public void notifyDDRunTerminated(final IDDRunResult result) {
		IDDRunListener[] listeners = this.getDDRunListeners();

		for (int i = 0; i < listeners.length; i++) {
			final IDDRunListener listener = listeners[i];

			Platform.run(new ListenerSafeRunnable() {
				public void run() {
					listener.ddRunTerminated(result);
				}
			});

		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.launcher.ddrun.IDDRunListenerManager#notifyMemoryGraphsCompared(int,
	 *      org.deltadebugging.ddstate.memorygraph.IRootNode,
	 *      org.deltadebugging.ddstate.memorygraph.IRootNode,
	 *      org.deltadebugging.ddcore.DeltaSet)
	 */
	public void notifyMemoryGraphsCompared(final int ddRunID,
			final IRootNode passGraph, final IRootNode failGraph,
			final DeltaSet deltas) {
		IDDRunListener[] listeners = this.getDDRunListeners();

		for (int i = 0; i < listeners.length; i++) {
			final IDDRunListener listener = listeners[i];

			Platform.run(new ListenerSafeRunnable() {
				public void run() {
					listener.memoryGraphsCompared(ddRunID, passGraph,
							failGraph, deltas);
				}
			});

		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.launcher.ddrun.IDDRunListenerManager#notifyDDRunStarted(org.deltadebugging.ddstate.launcher.ddrun.IDDRunDescription)
	 */
	public void notifyDDRunStarted(final IDDRunDescription desc) {
		IDDRunListener[] listeners = this.getDDRunListeners();

		for (int i = 0; i < listeners.length; i++) {
			final IDDRunListener listener = listeners[i];

			Platform.run(new ListenerSafeRunnable() {
				public void run() {
					listener.ddRunStarted(desc);
				}
			});

		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.launcher.ddrun.IDDRunListenerManager#notifyMemoryGraphsExtracted(int,
	 *      org.deltadebugging.ddstate.memorygraph.IRootNode,
	 *      org.deltadebugging.ddstate.memorygraph.IRootNode)
	 */
	public void notifyMemoryGraphsExtracted(final int ddRunID,
			final IRootNode passGraph, final IRootNode failGraph) {
		IDDRunListener[] listeners = this.getDDRunListeners();

		for (int i = 0; i < listeners.length; i++) {
			final IDDRunListener listener = listeners[i];

			Platform.run(new ListenerSafeRunnable() {
				public void run() {
					listener
							.memoryGraphsExtracted(ddRunID, passGraph, failGraph);
				}
			});

		}
	}

}