/*
 * $Id: AllMethodCallMatcher.java,v 1.9 2005/10/25 11:14:19 lehmann Exp $
 * 
 * Copyright (C) 2005 Karsten Lehmann
 * 
 * Software Engineering Chair, Saarland University
 * Prof. Dr. Andreas Zeller
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 *
 * Created on Aug 10, 2005 6:23:12 PM by
 * @author Karsten Lehmann 
 * Mail to: lehmann@cs.uni-sb.de or KarstenLehmann@web.de
 */
package org.deltadebugging.ddstate.launcher.matcher;

import org.deltadebugging.ddstate.launcher.DDStateLaunchPlugin;
import org.deltadebugging.ddstate.launcher.ddrun.IDDLaunchDescription;
import org.deltadebugging.ddstate.launcher.matcher.breakpoint.IBreakpointLocation;
import org.deltadebugging.ddstate.launcher.matcher.breakpoint.MethodEntryBreakpointLocation;
import org.deltadebugging.ddstate.launcher.matcher.exception.MatchPointComputationException;
import org.deltadebugging.ddstate.launcher.server.instrumenter.MethodIdentifier;
import org.eclipse.compare.rangedifferencer.RangeDifference;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.NullProgressMonitor;

/**
 * This Matcher find all method match points between the given test runs.
 * 
 * @author <a href="mailto:KarstenLehmann@web.de">Karsten Lehmann</a>
 */
public class AllMethodCallMatcher extends BaseMethodCallMatcher implements
        IMatchPointFinder {

    /*
     * (non-Javadoc)
     * 
     * @see org.deltadebugging.ddstate.launcher.matcher.IMatchPointFinder#computeMatchPoints(org.deltadebugging.ddstate.launcher.ddrun.IDDLaunchDescription,
     *      org.eclipse.core.runtime.IProgressMonitor)
     */
    public IMatchPoint[] computeMatchPoints(IDDLaunchDescription ddLaunchDesc,
            IProgressMonitor monitor) throws MatchPointComputationException {
        IMatchPoint[] result = null;
        
        try {
            monitor.beginTask("Compute match points...", 100);
            monitor.subTask("Compute method trace...");
            BaseMatcherResult matcherBaseResult = this.computeBaseInformations(
                    ddLaunchDesc, new NullProgressMonitor()); 
                    
                    // monitor);
            monitor.worked(50);

            //
            // Compute the equal ranges over the stacktrace. Thereby handle the
            // passing test method and the failing test method as equal
            //
            RangeDifference[] equalRanges = null;
            if (!monitor.isCanceled()) {
                monitor.subTask("Compute equal stack traces in method both method traces...");
                equalRanges = computeEqualStacktraces(matcherBaseResult,
                        monitor);
                
                monitor.worked(30);
            }

            //
            // Compute the match points
            //
            if (!monitor.isCanceled()) {
                monitor.subTask("Generate match points...");
                result = this.generateMatchPoints(equalRanges,
                        matcherBaseResult.getPassingStackTraces(),
                        matcherBaseResult.getFailingStackTraces());
                
                monitor.worked(20);
            }

            if (result != null) {
                logger.trace(DDStateLaunchPlugin.DEBUG, "Find " + result.length
                        + " match points between both test runs.");
            }
            monitor.worked(100);

        } finally {
            monitor.done();
        }

        return result;
    }

    /**
     * Generate all match points for the given range differences and the given
     * stack traces. The returned match point array comprised all match points
     * in the order they appear in the test runs.
     * 
     * @param equalRanges
     *            all sites in both test runs where the stack traces are equal.
     * @param passStackTraces
     *            the stacktraces of the passing test run.
     * @param failStackTraces
     *            the stacktraces of the failing test run.
     * @return all match points defined by the range differences.
     */
    private IMatchPoint[] generateMatchPoints(RangeDifference[] equalRanges,
            StackTrace[] passStackTraces, StackTrace[] failStackTraces) {
        IMatchPoint[] result = null;

        IMatchPoint[] helpResult = new IMatchPoint[Math.min(
                passStackTraces.length, failStackTraces.length)];
        int nrOfResult = 0;

        for (int i = 0; i < equalRanges.length; i++) {
            RangeDifference range = equalRanges[i];

            for (int j = 0; j < range.leftLength(); j++) {

                int leftIndex = range.leftStart() + j;
                int rightIndex = range.rightStart() + j;

                StackTrace passST = passStackTraces[leftIndex];
                StackTrace failST = failStackTraces[rightIndex];

                MethodIdentifier passMI = passST.getMethodOnTop();
                MethodIdentifier failMI = failST.getMethodOnTop();

                // FIXME outsource method identifier
                IBreakpointLocation passBD = new MethodEntryBreakpointLocation(
                        new org.deltadebugging.ddstate.memorygraph.MethodIdentifier(
                                passMI.getClassName(), passMI.getMethodName(),
                                passMI.getSignature()), passST.getCountOnTop());

                IBreakpointLocation failBD = new MethodEntryBreakpointLocation(
                        new org.deltadebugging.ddstate.memorygraph.MethodIdentifier(
                                failMI.getClassName(), failMI.getMethodName(),
                                failMI.getSignature()), failST.getCountOnTop());

                helpResult[nrOfResult] = new MatchPoint(passBD, failBD);

                nrOfResult++;
            }
        }

        result = new IMatchPoint[nrOfResult];
        System.arraycopy(helpResult, 0, result, 0, nrOfResult);

        return result;
    }
}
