/*
 * $Id: BaseMethodCallMatcher.java,v 1.9 2005/09/29 18:30:57 lehmann Exp $
 * 
 * Copyright (C) 2005 Karsten Lehmann
 * 
 * Software Engineering Chair, Saarland University
 * Prof. Dr. Andreas Zeller
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 *
 * Created on Aug 4, 2005 3:07:10 PM by
 * @author Karsten Lehmann 
 * Mail to: lehmann@cs.uni-sb.de or KarstenLehmann@web.de
 */
package org.deltadebugging.ddstate.launcher.matcher;

import org.deltadebugging.ddstate.launcher.DDStateLaunchPlugin;
import org.deltadebugging.ddstate.launcher.client.IExtTestRunner;
import org.deltadebugging.ddstate.launcher.client.JUnitTestLauncher;
import org.deltadebugging.ddstate.launcher.client.LaunchExceptionHandler;
import org.deltadebugging.ddstate.launcher.client.TestRunnerException;
import org.deltadebugging.ddstate.launcher.ddrun.IDDLaunchDescription;
import org.deltadebugging.ddstate.launcher.matcher.exception.MatchPointComputationException;
import org.deltadebugging.ddstate.launcher.server.GenerateTraceException;
import org.deltadebugging.ddstate.launcher.server.instrumenter.IdentifierMap;
import org.deltadebugging.ddstate.logger.DDLogger;
import org.eclipse.compare.rangedifferencer.RangeDifference;
import org.eclipse.compare.rangedifferencer.RangeDifferencer;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.launching.IVMInstall;

/**
 * Extend the BaseMethodCallMatcher if you want to compute matching method calls
 * over two program excecutions. This util provides some usefull functionality
 * for the subclass. <br>
 * Take a look at the methods <code>computeBaseInformations</code> and
 * <code>computetEqualStacktraces</code> if you implent a further method call
 * matcher.
 * 
 * @author <a href="mailto:KarstenLehmann@web.de">Karsten Lehmann </a>
 */
public abstract class BaseMethodCallMatcher implements IMatchPointFinder {

	/** Personal logger for this class */
	final static DDLogger logger = DDStateLaunchPlugin
			.getLogger(BaseMethodCallMatcher.class);

	/** The method argument signature for methods with no arguments */
	public static final String SIGNATUR_NO_ARGS = "()V";

	/**
	 * Compute some base informations for the method call matcher. This includes
	 * the method id of the passing and the failing test method in the junit
	 * class. Also the result includes the passing and the failing method trace
	 * and the stack traces array for the passing and the failing run.
	 * 
	 * @param ddLaunchDesc
	 *            the description for the delta debugging launch you want to
	 *            compute the base informations.
	 * @param monitor
	 * @return an object with some usefull data for the further computation of
	 *         the match points.
	 * @throws MatchPointComputationException
	 *             if something went wrong on the computation of the base
	 *             informations.
	 */
	protected BaseMatcherResult computeBaseInformations(
			IDDLaunchDescription ddLaunchDesc,
			IProgressMonitor monitor) throws MatchPointComputationException {

		BaseMatcherResult result = new BaseMatcherResult();

		String logKey = "Compute base informations"
				+ ddLaunchDesc.getDDLaunchID();
		logger.traceAndTimeStart(DDStateLaunchPlugin.DEBUG, logKey,
				"Compute base informations");

		MethodTrace passMethodTrace = null;
		MethodTrace failMethodTrace = null;

		IdentifierMap identiefierMap = new IdentifierMap();

		IJavaProject javaProject = ddLaunchDesc.getJavaProject();
		IVMInstall vmInstall = ddLaunchDesc.getVMInstall();

		// Create a DDRemoteTestLauncher for the passing and the failing run
		JUnitTestLauncher passDDRemoteLauncher = null;
		JUnitTestLauncher failDDRemoteLauncher = null;

		//
		// generate the method trace for the passing and the failing run
		//
		try {

			IExtTestRunner passTestRunner = null;
			IExtTestRunner failTestRunner = null;
			
			//
			// start the test launcher for the passing and the failing test run
			//
			if (!monitor.isCanceled()) {
				passDDRemoteLauncher = this.startTestLauncher(javaProject,
						vmInstall, monitor);
				failDDRemoteLauncher = this.startTestLauncher(javaProject,
						vmInstall, monitor);

				try {
					passTestRunner = passDDRemoteLauncher.getTestRunner();
					failTestRunner = failDDRemoteLauncher.getTestRunner();
				} catch (CoreException e) {
					LaunchExceptionHandler.handleException(e);
					monitor.setCanceled(true);
				}
			}

			if (!monitor.isCanceled()) {
				passMethodTrace = passTestRunner.generateTrace(ddLaunchDesc
						.getPassJUnitTestDescription(), identiefierMap);
				identiefierMap = passMethodTrace.getIdentifierMap();
				result.setPassingMethodTrace(passMethodTrace);

				logger.trace(DDStateLaunchPlugin.DEBUG_STACKTRACE_MATCHER,
						"Method trace passing run generated: "
								+ passMethodTrace.getLength()
								+ " number of method enter and exit calls.");
			}

			if (!monitor.isCanceled()) {
				failMethodTrace = failTestRunner.generateTrace(ddLaunchDesc
						.getFailJUnitTestDescription(), identiefierMap);
				identiefierMap = failMethodTrace.getIdentifierMap();
				result.setFailingMethodTrace(failMethodTrace);

				logger.trace(DDStateLaunchPlugin.DEBUG_STACKTRACE_MATCHER,
						"Method trace failing run generated: "
								+ failMethodTrace.getLength()
								+ " number of method enter and exit calls.");
			}

		} catch (GenerateTraceException e) {
			throw new MatchPointComputationException("Cannot generate trace", e);

		} catch (TestRunnerException e) {
			throw new MatchPointComputationException(
					"Unexcpected exception occured on generating method trace");

		} finally {
			// shut down the remote test launcher
			if (passDDRemoteLauncher != null)
				passDDRemoteLauncher.shutDown(false);

			if (failDDRemoteLauncher != null)
				failDDRemoteLauncher.shutDown(false);
			
		}

		//
		// compute the method id for the passing and the failing test method
		//        
		int passMethodID = -1;
		int failMethodID = -1;

		if (!monitor.isCanceled()) {
			String className = ddLaunchDesc.getPassJUnitTestDescription()
					.getTestCaseClassName();
			String methodName = ddLaunchDesc.getPassJUnitTestDescription()
					.getTestName();
			passMethodID = this.getMethodID(className, methodName,
					identiefierMap);
			result.setPassingJUnitMethodID(passMethodID);

			className = ddLaunchDesc.getFailJUnitTestDescription()
					.getTestCaseClassName();
			methodName = ddLaunchDesc.getFailJUnitTestDescription()
					.getTestName();
			failMethodID = this.getMethodID(className, methodName,
					identiefierMap);
			result.setFailingJUnitMethodID(failMethodID);
		}

		//
		// generate the stacktraces for the passing and the failing test run
		//
		StackTrace[] passStackTraces = null;
		StackTrace[] failStackTraces = null;

		if (!monitor.isCanceled()) {
			passMethodTrace = this
					.cutMethodTrace(passMethodTrace, passMethodID);
			passStackTraces = this.generateStackTraces(passMethodTrace);
			result.setPassingStackTraces(passStackTraces);

			logger.trace(DDStateLaunchPlugin.DEBUG_STACKTRACE_MATCHER,
					passStackTraces.length
							+ " number of stacktraces generated (passing run)");
		}

		if (!monitor.isCanceled()) {
			failMethodTrace = this
					.cutMethodTrace(failMethodTrace, failMethodID);
			failStackTraces = this.generateStackTraces(failMethodTrace);
			result.setFailingStackTraces(failStackTraces);

			logger.trace(DDStateLaunchPlugin.DEBUG_STACKTRACE_MATCHER,
					failStackTraces.length
							+ " number of stacktraces generated (failing run)");
		}

		logger.traceAndTimeEnd(DDStateLaunchPlugin.DEBUG, logKey,
				"Matcher base informations computed");

		return result;
	}

	/**
	 * Returns the method id for the given class and method name from the
	 * identifier.
	 * 
	 * @param className
	 *            the classname for the method id
	 * @param methodName
	 *            the methodname for the method id
	 * @param identifierMap
	 * @return the id of the defined method from the identifier map.
	 * @throws MatchPointComputationException
	 */
	private int getMethodID(String className, String methodName,
			IdentifierMap identifierMap) throws MatchPointComputationException {
		int result = -1;

		if (identifierMap.existsIdentifierFor(className, methodName,
				SIGNATUR_NO_ARGS)) {
			result = identifierMap.getIdentifier(className, methodName,
					SIGNATUR_NO_ARGS);
		}

		if (result == -1) {
			throw new MatchPointComputationException("Cannot find the method "
					+ className + "." + methodName
					+ "() within the given identifier map");
		}

		return result;
	}

	/**
	 * Returns a new method trace from the first occurance of the given test
	 * method id until the last occurance of the inverse test method id. This
	 * means we cut the method trace from the first call of the test method
	 * until the last exit call of the test method.
	 * 
	 * @param methodTrace
	 *            the orginal method trace.
	 * @param testMethodID
	 * @return a new method trace from the first occurance of the given test
	 *         method id until the end.
	 */
	private MethodTrace cutMethodTrace(MethodTrace methodTrace, int testMethodID) {
		MethodTrace result = new MethodTrace(methodTrace.getIdentifierMap());

		int traceLength = methodTrace.getLength();

		// cut until the call of the test method
		for (int i = 0; i < traceLength; i++) {
			if (testMethodID == methodTrace.getMethodIDAt(i)) {
				result = methodTrace.getSubTrace(i);
				break;
			}
		}

		// cut until you leave the test method
		int newTraceLength = result.getLength();
		int exitTestMethodID = testMethodID * -1;
		for (int i = newTraceLength - 1; i >= 0; i--) {
			if (exitTestMethodID == result.getMethodIDAt(i)) {
				result = result.getSubTrace(0, i);
				break;
			}
		}

		return result;
	}

	/**
	 * Generate for every method enter entry in the method trace the stacktrace
	 * at this position. So be get a array which reflect the stacktraces during
	 * the test execution.
	 * 
	 * @param methodTrace
	 *            a method trace with all methods called during the test
	 *            execution.
	 * @return the generated stack traces of the method trace.
	 */
	private StackTrace[] generateStackTraces(MethodTrace methodTrace) {
		StackTrace[] result = null;

		int[] methodIDTrace = methodTrace.getMethodIDTrace();
		int[] methodCountTrace = methodTrace.getMethodCountTrace();

		int nrOfResult = 0;

		StackTrace actualTrace = new StackTrace(methodTrace.getIdentifierMap());
		StackTrace[] helpResult = new StackTrace[methodIDTrace.length];

		for (int i = 0; i < methodTrace.getLength(); i++) {
			int methodID = methodIDTrace[i];
			int count = methodCountTrace[i];

			if (methodID > 0) {
				actualTrace = actualTrace.pushMethod(methodID, count);
				helpResult[nrOfResult] = actualTrace;
				nrOfResult++;

			} else {
				actualTrace = actualTrace.popMethod();
			}
		}

		result = new StackTrace[nrOfResult];
		System.arraycopy(helpResult, 0, result, 0, nrOfResult);

		return result;
	}

	/**
	 * Compute the range differences over the stacktraces include in the given
	 * baseMatcherResult. Return only this range differences that are of kind
	 * <code>RangeDifference.NOCHANGE</code>.
	 * 
	 * @param baseMatcherResult
	 *            the base matcher result including the stacktraces and the
	 *            method id's for the passing and the failing run.
	 * @param monitor
	 * @return all range differences over the two stack traces in the base
	 *         matcher result.
	 */
	protected RangeDifference[] computeEqualStacktraces(
			BaseMatcherResult baseMatcherResult, IProgressMonitor monitor) {
		RangeDifference[] result = null;

		StackTraceRangeComparator passSTRC = new StackTraceRangeComparator(
				baseMatcherResult.getPassingStackTraces(), baseMatcherResult
						.getPassingJUnitMethodID());
		StackTraceRangeComparator failSTRC = new StackTraceRangeComparator(
				baseMatcherResult.getFailingStackTraces(), baseMatcherResult
						.getFailingJUnitMethodID());

		if (!monitor.isCanceled())
			result = RangeDifferencer.findRanges(passSTRC, failSTRC);

		if (!monitor.isCanceled())
			result = this.eleminitaDifferences(result);

		return result;
	}

	/**
	 * Eleminate in the given range difference array all differrences that are
	 * not of kind <code>RangeDifference.NOCHANGE</code>.
	 * 
	 * @param rangeDifferences
	 *            an array of range differences.
	 * @return an array of range differences containing all range differences
	 *         from the given array that are of kind
	 *         <code>RangeDifference.NOCHANGE</code>.
	 */
	private RangeDifference[] eleminitaDifferences(
			RangeDifference[] rangeDifferences) {
		RangeDifference[] result = null;

		int nrOfResult = 0;
		RangeDifference[] helpResult = new RangeDifference[rangeDifferences.length];

		for (int i = 0; i < rangeDifferences.length; i++) {
			RangeDifference difference = rangeDifferences[i];

			if (difference.kind() == RangeDifference.NOCHANGE) {
				helpResult[nrOfResult] = difference;
				nrOfResult++;
			}
		}

		result = new RangeDifference[nrOfResult];
		System.arraycopy(helpResult, 0, result, 0, nrOfResult);

		return result;
	}

	JUnitTestLauncher startTestLauncher(final IJavaProject javaProject,
			final IVMInstall vmInstall, final IProgressMonitor monitor) {
		JUnitTestLauncher result = new JUnitTestLauncher();

		try {
			result.start(javaProject, vmInstall, monitor);
		} catch (CoreException e) {
			LaunchExceptionHandler.handleException(e);
			monitor.setCanceled(true);
		}

		return result;
	}

}
