/*
 * $Id: SimpleMethodCallMatcher.java,v 1.20 2005/10/25 11:14:28 lehmann Exp $
 * 
 * Copyright (C) 2005 Karsten Lehmann
 * 
 * Software Engineering Chair, Saarland University
 * Prof. Dr. Andreas Zeller
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 *
 * Created on Jul 20, 2005 3:30:30 PM by
 * @author Karsten Lehmann 
 * Mail to: lehmann@cs.uni-sb.de or KarstenLehmann@web.de
 */
package org.deltadebugging.ddstate.launcher.matcher;

import org.deltadebugging.ddstate.launcher.DDStateLaunchPlugin;
import org.deltadebugging.ddstate.launcher.ddrun.IDDLaunchDescription;
import org.deltadebugging.ddstate.launcher.matcher.breakpoint.IBreakpointLocation;
import org.deltadebugging.ddstate.launcher.matcher.breakpoint.MethodEntryBreakpointLocation;
import org.deltadebugging.ddstate.launcher.matcher.exception.MatchPointComputationException;
import org.deltadebugging.ddstate.launcher.server.instrumenter.MethodIdentifier;
import org.eclipse.compare.rangedifferencer.RangeDifference;
import org.eclipse.core.runtime.IProgressMonitor;

/**
 * COMMENTME Comment type SimpleMethodCallMatcher
 * 
 * @author <a href="mailto:KarstenLehmann@web.de">Karsten Lehmann </a>
 */
public class SimpleMethodCallMatcher extends BaseMethodCallMatcher implements
		IMatchPointFinder {

	private final double[] fSearchPoints;

	public SimpleMethodCallMatcher(double[] searchPoints) {
		this.fSearchPoints = searchPoints;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.launcher.matcher.IMatchPointFinder#computeMatchPoints(org.deltadebugging.ddstate.launcher.ddrun.IDDLaunchDescription,
	 *      org.eclipse.core.runtime.IProgressMonitor)
	 */
	public IMatchPoint[] computeMatchPoints(IDDLaunchDescription ddLaunchDesc,
			IProgressMonitor monitor) throws MatchPointComputationException {

        IMatchPoint[] result = null;
        
        try {
            monitor.beginTask("Compute match points", 100);
            
            BaseMatcherResult matcherBaseResult = this.computeBaseInformations(
                    ddLaunchDesc, monitor);
            monitor.worked(50);

            //
            // Compute the equal ranges over the stacktrace. Thereby handle the
            // passing test method and the failing test method as equal
            //
            RangeDifference[] equalRanges = null;
            if (!monitor.isCanceled()) {
                equalRanges = computeEqualStacktraces(matcherBaseResult, monitor);
                monitor.worked(30);
            }

            //
            // Compute the match points
            //
            if (!monitor.isCanceled()) {
                result = this.computeMatchPoints(equalRanges, matcherBaseResult
                        .getPassingStackTraces(), matcherBaseResult
                        .getFailingStackTraces());
                monitor.worked(20);
            }
            
            
        } finally {
            monitor.done();
        }


		return result;
	}

	private IMatchPoint[] computeMatchPoints(RangeDifference[] equalRanges,
			StackTrace[] passStackTraces, StackTrace[] failStackTraces) {
		IMatchPoint[] result = null;

		IMatchPoint[] helpResult = new IMatchPoint[this.fSearchPoints.length];
		int nrOfResult = 0;

		//
		// compute the real indexes for the search points and eleminate the
		// duplicated of the computed index array
		//
		int[] indexes = this.computeIndexes(this.fSearchPoints,
				passStackTraces.length);
		indexes = this.eliminateEqualIndexes(indexes);

		for (int i = 0; i < indexes.length; i++) {
			int[] indexMatch = this.getIndexMatch(indexes[i], equalRanges);

			if (indexMatch != null) {
				logger.trace(DDStateLaunchPlugin.DEBUG_STACKTRACE_MATCHER,
						"Find comparabled traces for match point");
				logger.trace(DDStateLaunchPlugin.DEBUG_STACKTRACE_MATCHER,
						"Passing stack trace:\n"
								+ passStackTraces[indexMatch[0]]);
				logger.trace(DDStateLaunchPlugin.DEBUG_STACKTRACE_MATCHER,
						"Failing stack trace:\n"
								+ failStackTraces[indexMatch[1]]);

				StackTrace passST = passStackTraces[indexMatch[0]];
				StackTrace failST = failStackTraces[indexMatch[1]];

				MethodIdentifier passMI = passST.getMethodOnTop();
				MethodIdentifier failMI = failST.getMethodOnTop();

				// FIXME outsource method identifier
				IBreakpointLocation passBD = new MethodEntryBreakpointLocation(
						new org.deltadebugging.ddstate.memorygraph.MethodIdentifier(
								passMI.getClassName(), passMI.getMethodName(),
								passMI.getSignature()), passST.getCountOnTop());
				IBreakpointLocation failBD = new MethodEntryBreakpointLocation(
						new org.deltadebugging.ddstate.memorygraph.MethodIdentifier(
								failMI.getClassName(), failMI.getMethodName(),
								failMI.getSignature()), failST.getCountOnTop());

				helpResult[nrOfResult] = new MatchPoint(passBD, failBD);
				nrOfResult++;
			}
		}

		result = new IMatchPoint[nrOfResult];
		System.arraycopy(helpResult, 0, result, 0, nrOfResult);

		return result;
	}

	private int[] getIndexMatch(int index, RangeDifference[] rangeDifferences) {
		int[] result = null;
		for (int i = 0; i < rangeDifferences.length; i++) {
			RangeDifference difference = rangeDifferences[i];

			if (difference.rightStart() <= index
					&& index < difference.rightEnd()) {
				result = new int[2];
				result[1] = index;

				result[0] = difference.leftStart()
						+ (index - difference.rightStart());
			}
		}

		return result;
	}

	/**
	 * Compute with the given prozentual search points and the array length the
	 * real indexes in the array.
	 * 
	 * @param searchPoints
	 *            The search points are number between 0.0 and 1.0.
	 * @param arraylength
	 *            The length of the array you want to compute the indexes.
	 * @return an array of indexes corresponding to the prozentual search
	 *         points.
	 */
	private int[] computeIndexes(double[] searchPoints, int arraylength) {
		int[] result = new int[searchPoints.length];

		for (int i = 0; i < searchPoints.length; i++) {
			long index = Math.round(searchPoints[i] * (arraylength - 1));

			result[i] = new Long(index).intValue();
		}

		return result;
	}

	/**
	 * Eliminate equal entries in the given index array. The precondition is
	 * that the array have to be sorted ascending.
	 * 
	 * @param indexes
	 *            The index array you want to clean of duplicated entries.
	 * @return The cleaned index array with no equal entries.
	 */
	private int[] eliminateEqualIndexes(int[] indexes) {
		int[] result = null;
		int nrOfResult = 0;

		int[] helpResult = new int[indexes.length];

		if (indexes.length > 0) {
			// Initialize the help array with the first value
			helpResult[0] = indexes[0];
			nrOfResult = 1;

			for (int i = 1; i < indexes.length; i++) {

				if (indexes[i] > helpResult[i - 1]) {
					helpResult[nrOfResult] = indexes[i];
					nrOfResult++;
				}
			}
		}

		result = new int[nrOfResult];
		System.arraycopy(helpResult, 0, result, 0, nrOfResult);

		return result;

	}
}
