/*
 * $Id: ExtTestCaseClassLoader.java,v 1.6 2006/01/10 19:01:41 lehmann Exp $
 * 
 * Copyright (C) 2005 Karsten Lehmann
 * 
 * Software Engineering Chair, Saarland University
 * Prof. Dr. Andreas Zeller
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 *
 * Created on 17.03.2005 10:02:00 by
 * @author Karsten Lehmann 
 * Mail to: lehmann@cs.uni-sb.de or KarstenLehmann@web.de
 */
package org.deltadebugging.ddstate.launcher.server;

import java.io.IOException;
import java.io.InputStream;
import java.net.URL;
import java.util.Enumeration;
import java.util.Properties;
import java.util.StringTokenizer;
import java.util.Vector;

import org.deltadebugging.ddstate.launcher.bcel.instrumenter.IClassInstrumenter;
import org.deltadebugging.ddstate.launcher.bcel.instrumenter.InstrumentationClassLoader;

/**
 * COMMENTME Comment type ExtTestCaseClassLoader
 * 
 * @author <a href="mailto:KarstenLehmann@web.de">Karsten Lehmann </a>
 */
public class ExtTestCaseClassLoader extends InstrumentationClassLoader {

    private static int id = 1;

    private final int currentClassLoaderID = ExtTestCaseClassLoader.getNextID();

    private final static synchronized int getNextID() {
        return id++;
    }

    public int getID() {
        return this.currentClassLoaderID;
    }

    /**
     * Constructs a TestCaseLoader. It scans the class path and the excluded
     * package paths
     */
    public ExtTestCaseClassLoader(IClassInstrumenter classInstrumenter) {
        this(classInstrumenter, System.getProperty("java.class.path"));
    }

    /**
     * Constructs a TestCaseLoader. It scans the class path and the excluded
     * package paths
     */
    public ExtTestCaseClassLoader(IClassInstrumenter classInstrumenter,
            String classPath) {
        super(classInstrumenter);
        scanPath(classPath);
        readExcludedPackages();
    }

    /** scanned class path */
    private Vector fPathItems;
    /** default excluded paths */
    private String[] defaultExclusions = { "junit.framework.",
            "junit.extensions.", "junit.runner." };
    /** name of excluded properties file */
    static final String EXCLUDED_FILE = "excluded.properties";
    /** excluded paths */
    private Vector fExcluded;

    private void scanPath(String classPath) {
        String separator = System.getProperty("path.separator");
        this.fPathItems = new Vector(10);
        StringTokenizer st = new StringTokenizer(classPath, separator);
        while (st.hasMoreTokens()) {
            this.fPathItems.addElement(st.nextToken());
        }
    }

    public URL getResource(String name) {
        return java.lang.ClassLoader.getSystemResource(name);
    }

    public InputStream getResourceAsStream(String name) {
        return java.lang.ClassLoader.getSystemResourceAsStream(name);
    }

    public boolean isExcluded(String name) {
        for (int i = 0; i < this.fExcluded.size(); i++) {
            if (name.startsWith((String) this.fExcluded.elementAt(i))) {
                return true;
            }
        }
        return false;
    }

    public synchronized Class loadClass(String name, boolean resolve)
            throws ClassNotFoundException {

        Class c = findLoadedClass(name);
        if (c != null) {
            return c;
        }

        //
        // Delegate the loading of excluded classes to the
        // standard class loader.
        //
        if (isExcluded(name)) {
            try {
                c = findSystemClass(name);
                // c = super.loadClass(name, resolve);
                return c;
            } catch (ClassNotFoundException e) {
                // keep searching
            }
        }

        if (c == null) {
            c = super.loadClass(name, resolve);
        }

        return c;
    }

    boolean isJar(String pathEntry) {
        return pathEntry.endsWith(".jar") || pathEntry.endsWith(".zip");
    }

    private void readExcludedPackages() {
        this.fExcluded = new Vector(10);
        for (int i = 0; i < this.defaultExclusions.length; i++) {
            this.fExcluded.addElement(this.defaultExclusions[i]);
        }

        InputStream is = getClass().getResourceAsStream(EXCLUDED_FILE);
        if (is == null)
            return;
        Properties p = new Properties();
        try {
            p.load(is);
        } catch (IOException e) {
            return;
        } finally {
            try {
                is.close();
            } catch (IOException e) {
                // do nothing
            }
        }
        for (Enumeration e = p.propertyNames(); e.hasMoreElements();) {
            String key = (String) e.nextElement();
            if (key.startsWith("excluded.")) {
                String path = p.getProperty(key);
                path = path.trim();
                if (path.endsWith("*"))
                    path = path.substring(0, path.length() - 1);
                if (path.length() > 0)
                    this.fExcluded.addElement(path);
            }
        }
    }
}