/*
 * $Id: MemGraphDeltaTester.java,v 1.23 2005/10/25 13:38:38 lehmann Exp $
 * 
 * Copyright (C) 2005 Karsten Lehmann
 * 
 * Software Engineering Chair, Saarland University
 * Prof. Dr. Andreas Zeller
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 *
 * Created on 03.02.2005 13:31:33 by
 * @author Karsten Lehmann 
 * Mail to: lehmann@cs.uni-sb.de or KarstenLehmann@web.de
 */
package org.deltadebugging.ddstate.launcher.tester;

import java.util.Iterator;

import org.deltadebugging.ddcore.DD;
import org.deltadebugging.ddcore.DeltaSet;
import org.deltadebugging.ddcore.tester.ITester;
import org.deltadebugging.ddstate.launcher.DDStateLaunchPlugin;
import org.deltadebugging.ddstate.launcher.client.IExtTestRunner;
import org.deltadebugging.ddstate.launcher.client.ITestDescription;
import org.deltadebugging.ddstate.launcher.client.JUnitTestLauncher;
import org.deltadebugging.ddstate.launcher.client.TestRunnerException;
import org.deltadebugging.ddstate.launcher.ddrun.DDTestResult;
import org.deltadebugging.ddstate.launcher.ddrun.IDDRunResult;
import org.deltadebugging.ddstate.launcher.ddrun.IDDRunStatusCodes;
import org.deltadebugging.ddstate.launcher.tester.exception.TesterStartException;
import org.deltadebugging.ddstate.launcher.tester.predicate.IPredicate;
import org.deltadebugging.ddstate.logger.DDLogger;
import org.deltadebugging.ddstate.memorygraph.delta.IMemoryGraphDelta;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.jdt.debug.core.IJavaBreakpoint;
import org.eclipse.jdt.debug.core.JDIDebugModel;

/**
 * COMMENTME Comment type MemGraphDeltaTester
 * 
 * @author <a href="mailto:KarstenLehmann@web.de">Karsten Lehmann </a>
 */
public class MemGraphDeltaTester implements ITester {

	/**
	 * Indicate that the tester run was aborted without any unexpected error
	 * (e.g. the postlaunch check fails).
	 */
	public final static int STATUS_CODE_OK = 0;

	public final static int STATUS_CODE_UNEXPECTED_ERROR = 1;

	public final static int STATUS_CODE_ABORTED = 2;

	/** Personal logger for this class */
	final static DDLogger logger = DDStateLaunchPlugin
			.getLogger(MemGraphDeltaTester.class);

	private final IProgressMonitor fMonitor;

	private final IDDRunResult fDDRunResult;

	private final boolean fComputeReverseDelta = false;

	private final IPredicate fTestPredicate;

	/**
	 * The test description that will be run by the test runner.
	 */
	private final ITestDescription fTestDesc;

	/**
	 * The whole delta set used for the post launch check.
	 */
	private final DeltaSet fAllDeltas;

	/**
	 * The test launcher that will be initialized, if this tester was
	 * initialized with an test runner.
	 */
	private JUnitTestLauncher fTestLauncher = null;

	/**
	 * The test runner used to start the junit tests.
	 */
	private IExtTestRunner fTestRunner = null;

	/**
	 * The delta processor works as breakpoint listener and apply the delta set
	 * if the given breakpoint is hit.
	 */
	private DeltaProcessor fDeltaProcessor = null;

	/**
	 * The result of the last run of this tester.
	 */
	private IStatus fTesterResult = Status.OK_STATUS;

	/**
	 * Simple outcome cache to remember allready executed tests. This is to
	 * speed up the algorithm.
	 */
	private SimpleOutcomeCache fOutcomeCache = null;

	/**
	 * Counter for the tests executed by the tester. This counter not include
	 * the tests taken from the outcome cache.
	 */
	private int fSerialNumber = 0;

	private IJavaBreakpoint fBreakpoint;

	public MemGraphDeltaTester(IDDRunResult ddRunResult,
			ITestDescription testDesc, DeltaSet allDeltas,
			IPredicate testPredicate, IProgressMonitor monitor) {
		this(ddRunResult, null, testDesc, allDeltas, testPredicate, monitor);
	}

	public MemGraphDeltaTester(IDDRunResult ddRunResult,
			IExtTestRunner testRunner, ITestDescription testDesc,
			DeltaSet allDeltas, IPredicate testPredicate,
			IProgressMonitor monitor) {
		this.fDDRunResult = ddRunResult;
		this.fTestRunner = testRunner;
		this.fTestDesc = testDesc;
		this.fAllDeltas = allDeltas;
		this.fTestPredicate = testPredicate;
		this.fMonitor = monitor;
	}

	public void start() throws TesterStartException {

		if (this.fTestRunner == null) {
			this.fTestLauncher = new JUnitTestLauncher();
			try {
				this.fTestLauncher.start(this.fTestDesc.getJavaProject(),
						this.fTestDesc.getVMInstall(), this.fMonitor);
			} catch (CoreException e) {
				throw new TesterStartException("Cannot start test launcher.", e);
			}

			try {
				this.fTestRunner = this.fTestLauncher.getTestRunner();
			} catch (CoreException e) {
				throw new TesterStartException(
						"Cannot retrieve the test runner of the test launcher.",
						e);
			}
		}

		try {
			// try to add the breakpoint direct to the debugtarget
			this.fBreakpoint = DDStateLaunchPlugin.getBreakpointManager()
					.createBreakpoint(this.fTestDesc.getBreakpointLocation(),
							false);

			if (!this.fMonitor.isCanceled()) {
				this.fTestRunner.getDebugTarget().breakpointAdded(
						this.fBreakpoint);
			}

		} catch (CoreException e) {
			throw new TesterStartException(
					"Cannot create and register the breakpoint of the given test description.",
					e);
		}

		this.fDeltaProcessor = new DeltaProcessor(this.fTestRunner,
				this.fBreakpoint);
		JDIDebugModel.addJavaBreakpointListener(this.fDeltaProcessor);

		this.fSerialNumber = 0;
		this.fOutcomeCache = new SimpleOutcomeCache();

		this.fTesterResult = Status.OK_STATUS;
	}

	/**
	 * Returns the result of this tester's last run.
	 * 
	 * @return the result of this tester's last run.
	 */
	public IStatus getResult() {
		return this.fTesterResult;
	}

	/**
	 * Returns true if the test run was aborted. This can caused by cancel the
	 * progress monitor or if the test run itself reports an error.
	 * 
	 * @return true if the test run was aborted, otherwise false.
	 */
	public boolean isTestRunAborted() {
		boolean result = false;

		if (this.fMonitor != null && this.fMonitor.isCanceled())
			result = true;

		if (this.fTesterResult != null
				&& this.fTesterResult.getCode() != STATUS_CODE_OK)
			result = true;

		return result;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddcore.tester.ITester#test(org.deltadebugging.ddcore.DeltaSet)
	 */
	public int test(DeltaSet c) {
		int cSize = c.size();
		String deltas = cSize > 1 ? "deltas" : "delta";
		this.fMonitor.subTask("Minimize state deltas (apply " + cSize + " " + deltas + ")...");
		
		int result = DD.UNRESOLVED;

		// check if the test run is already aborted
		if (this.isTestRunAborted()) {
			logger.trace(DDStateLaunchPlugin.DEBUG_DELTA_TESTER, "-> "
					+ this.outcomeToString(DD.FAIL) + " <- (aborted)");

			return DD.FAIL;

		}

		// ... else
		logger.trace(DDStateLaunchPlugin.DEBUG_DELTA_TESTER, "Test " + c.size()
				+ " deltas");

		// check if the outcome for the given delta set is already computed
		// and stored in the outcome cache
		if (this.fOutcomeCache.existTestOutcome(c)) {
			result = this.fOutcomeCache.getTestOutcome(c);

			logger.trace(DDStateLaunchPlugin.DEBUG, "-> "
					+ this.outcomeToString(result) + " <- (cached)");

		} else {
			this.fSerialNumber++;

			// test if the given delta set is valid
			if (this.preLaunchCheck(c) == true) {
				result = startTest(c);
				logger.trace(DDStateLaunchPlugin.DEBUG, "-> "
						+ this.outcomeToString(result) + " <-");

			} else {
				result = DD.UNRESOLVED;
				logger.trace(DDStateLaunchPlugin.DEBUG_DELTA_TESTER, "-> "
						+ this.outcomeToString(result)
						+ " <- (delta set invalid)");

			}

			if (this.fTesterResult.getCode() == STATUS_CODE_OK) {
				// do some post launch checks
				this.postLaunchCheck(result, c);

				// add test result to dd run result
				DDTestResult testResult = new DDTestResult(this.fSerialNumber,
						c, result);
				this.fDDRunResult.addTestResult(testResult);

				// store the test outcome in the cache
				this.fOutcomeCache.addTestOutcome(c, result);
			}
		}

		if (this.fComputeReverseDelta) {
			result = switchOutcome(result);
		}

		return result;
	}

	/**
	 * Switch the test outcome.
	 * 
	 * <ul>
	 * <li>DD.FAIL -> DD.PASS</li>
	 * <li>DD.PASS -> DD.FAIL</li>
	 * <li>DD.UNRESOLVED -> DD.UNRESOLVED</li>
	 * </ul>
	 * 
	 * @param testOutcome
	 *            the test result to switch.
	 * @return the switched test outcome.
	 */
	private int switchOutcome(int testOutcome) {
		int result = testOutcome;

		switch (testOutcome) {
		case DD.FAIL:
			result = DD.PASS;
			break;

		case DD.PASS:
			result = DD.FAIL;
			break;

		default:
			result = DD.UNRESOLVED;
			break;
		}

		return result;
	}

	/**
	 * After the check of all preconditions this is the method which start the
	 * test for the given delta set.
	 * 
	 * @param c
	 *            the delta set to test
	 * @return the test outcome of the test.
	 */
	private int startTest(DeltaSet c) {
		int result = DD.UNRESOLVED;

		// Set the deltaset that should applied if the breakpoint is
		// reached.
		this.fDeltaProcessor.setTestSet(c);

		// Enable the delta processor to ensure that the JUnit run halts on
		// the given breakpoint.
		this.fDeltaProcessor.enable();

		try {
			// Ensure that the breakpoint is enabled too.
			if (!this.fBreakpoint.isEnabled()) {
				this.fBreakpoint.setEnabled(true);
			}

			IPredicate predicate = this.fTestRunner.run(this.fTestDesc
					.getJUnitTestDescription());
			
			result = this.evaluatePredicate(predicate); 

		} catch (TestRunnerException e) {
			this.fTesterResult = logger.log(IStatus.ERROR,
					STATUS_CODE_UNEXPECTED_ERROR,
					"Unexpected error on test execution", e);

			result = DD.FAIL;

		} catch (CoreException e) {
			this.fTesterResult = logger.log(IStatus.ERROR,
					STATUS_CODE_UNEXPECTED_ERROR, "Cannot enable breakpoint.",
					e);

			result = DD.FAIL;

		}

		if (this.fDeltaProcessor.hasErros()) {
			logger.trace(DDStateLaunchPlugin.DEBUG_DELTA_TESTER,
					"The delta processor reported an error",
					this.fDeltaProcessor.getError());
			result = DD.UNRESOLVED;
		}

		return result;
	}

	private int evaluatePredicate(IPredicate predicate) {
		int result = DD.UNRESOLVED;
		
		if (!this.fComputeReverseDelta) {
			if (predicate.wasSuccessful()) {
				result = DD.PASS;
			} else if (predicate.matches(this.fTestPredicate)) {
				result = DD.FAIL;
			}
			
		} else {
			if (predicate.wasSuccessful()) {
				result = DD.PASS;
			} else {
				result = DD.FAIL;
			}
			
		}
		
		return result;
	}

	/**
	 * Convert the outcome of a delta debugging run in a user readable form.
	 * 
	 * @param outcome
	 *            the outcome of a delta debugging run
	 * @return the outcome of a delta debugging run in a user readable form.
	 */
	private String outcomeToString(int outcome) {
		String result = null;

		switch (outcome) {
		case DD.FAIL:
			result = "FAIL";
			break;

		case DD.PASS:
			result = "PASS";
			break;

		case DD.UNRESOLVED:
			result = "UNRESOLVED";
			break;

		default:
			result = "Outcome undefined";
			break;
		}

		return result;
	}

	/**
	 * Check if all deltas was applied the passing run must fail or the failing
	 * run must pass (reverse delta computing).
	 * 
	 * @param result
	 *            computed result.
	 * @param testedSet
	 *            tested delta set.
	 */
	private void postLaunchCheck(int result, DeltaSet testedSet) {

		if (testedSet.size() == this.fAllDeltas.size()) { // only to speed up

			if (testedSet.containsAll(this.fAllDeltas)) {
				if (!this.fComputeReverseDelta && result != DD.FAIL) {
					logger.trace(DDStateLaunchPlugin.DEBUG_DELTA_TESTER,
							"Postlaunch check failed: "
									+ "passing run does not fail.");
					IStatus status = new Status(
							IStatus.INFO,
							DDStateLaunchPlugin.PLUGIN_ID,
							IDDRunStatusCodes.PASSING_RUN_NOT_FAIL,
							"After all deltas are applied the given passing run does not fail.",
							null);

					this.fDDRunResult.addStatus(status);

					this.fTesterResult = new Status(IStatus.INFO,
							DDStateLaunchPlugin.PLUGIN_ID, STATUS_CODE_ABORTED,
							"Postlaunch check failed", null);
				}

				if (this.fComputeReverseDelta && result != DD.PASS) {
					logger.trace(DDStateLaunchPlugin.DEBUG_DELTA_TESTER,
							"Postlaunch check failed: "
									+ "failing run does not pass.");

					IStatus status = new Status(
							IStatus.INFO,
							DDStateLaunchPlugin.PLUGIN_ID,
							IDDRunStatusCodes.FAILING_RUN_NOT_PASS,
							"After all deltas are applied the given failing run does not pass.",
							null);

					this.fDDRunResult.addStatus(status);

					this.fTesterResult = new Status(IStatus.INFO,
							DDStateLaunchPlugin.PLUGIN_ID, STATUS_CODE_ABORTED,
							"Postlaunch check failed", null);
				}
			}
		}
	}

	/**
	 * The number of executed tests since start was called
	 * 
	 * @return the number of executed tests since start was called.
	 */
	public int getNumberOfExecutedTests() {
		return this.fSerialNumber;
	}

	public void shutDown() {

		// unregister and delete breakpoint
		if (this.fBreakpoint != null)
			try {
				if (this.fTestRunner != null) {
					this.fTestRunner.getDebugTarget().breakpointRemoved(
							this.fBreakpoint, null);
				}

				if (this.fBreakpoint != null) {
					this.fBreakpoint.delete();
				}

			} catch (CoreException e) {
				// do nothing
			}

		// shutdown the test launcher
		if (this.fTestLauncher != null)
			this.fTestLauncher.shutDown(true);

		// remove delta processor
		if (this.fDeltaProcessor != null)
			JDIDebugModel.removeJavaBreakpointListener(this.fDeltaProcessor);

		this.fOutcomeCache = null;
	}

	/**
	 * Check if every delta in the given delta set is valid against the delta
	 * set. That means that every dependant deltas from all deltas in the given
	 * set must contain in the set.
	 * 
	 * @param deltaSet
	 *            the delta set you want to check.
	 * @return true if the given delta set is valid against itself, otherwise
	 *         false.
	 */
	private boolean preLaunchCheck(DeltaSet deltaSet) {
		boolean result = true;

		Iterator itDeltas = deltaSet.iterator();

		while (itDeltas.hasNext() && result == true) {
			IMemoryGraphDelta delta = (IMemoryGraphDelta) itDeltas.next();

			result = delta.containsAllDependencies(deltaSet);
		}

		return result;
	}
}