/*
 * $Id: PredicateFactory.java,v 1.6 2005/10/25 13:39:07 lehmann Exp $
 * 
 * Copyright (C) 2005 Karsten Lehmann
 * 
 * Software Engineering Chair, Saarland University
 * Prof. Dr. Andreas Zeller
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 *
 * Created on Sep 12, 2005 10:27:20 AM by
 * @author Karsten Lehmann 
 * Mail to: lehmann@cs.uni-sb.de or KarstenLehmann@web.de
 */
package org.deltadebugging.ddstate.launcher.tester.predicate;

import java.util.regex.Matcher;
import java.util.regex.Pattern;

import junit.framework.TestFailure;
import junit.framework.TestResult;

import org.deltadebugging.ddstate.launcher.client.IJUnitTestDescription;

/**
 * Use the predicate factory to generate a predicate for the test result of one
 * junit test run.
 * 
 * @author <a href="mailto:KarstenLehmann@web.de">Karsten Lehmann</a>
 */
public class PredicateFactory {

    private final static Pattern COMPARISON_PATTERN = Pattern
            .compile("^expected:<(.*)> but was:<(.*)>$");
    private final static Pattern OBJECT_REF = Pattern
            .compile("^.*@[\\dabcdef]*$");

    private PredicateFactory() {
        // use the static factory methods
    }

    /**
     * Creates a predicate for the given test result and the give junit test
     * description.
     * 
     * @param testResult
     *            the test result of a single junit test run.
     * @param testDesc
     *            the description of the single junit test run.
     * @return the created predicate for the result of the junit test run.
     */
    public static IPredicate createPredicate(TestResult testResult,
            IJUnitTestDescription testDesc) {
        IPredicate result = null;

        // Either an error occured or a failure or nothing
        if (testResult.failureCount() > 0) {
            TestFailure failure = (TestFailure) testResult.failures()
                    .nextElement();
            result = createPredicate(failure, testDesc);

        } else if (testResult.errorCount() > 0) {
            TestFailure failure = (TestFailure) testResult.errors()
                    .nextElement();
            result = createPredicate(failure, testDesc);

        } else {
            result = new PassPredicate();

        }

        return result;
    }

    /**
     * Creates a predicate for the failed junit test run.
     * 
     * @param failure
     *            the test failure occured on running the junit test run.
     * @param testDesc
     *            the description of the single junit test run.
     * @return the created predicate for the result of the junit test run.
     */
    public static IPredicate createPredicate(TestFailure failure,
            IJUnitTestDescription testDesc) {
        IPredicate result = null;

        result = PredicateFactory.computeComparisonPredicate(failure, testDesc);

        if (result == null) {
            result = new ExceptionPredicate(failure, testDesc);
        }

        return result;
    }

    /**
     * Compute and return the comparison predicate. If is computation was not
     * successful the functions returnx <code>null</code>.
     * 
     * @param failure
     *            the test failure occured on running the junit test run.
     * @param testDesc
     *            the description of the single junit test run.
     * @return the computed comparison predicate, or <code>null</code> if the
     *         computation failed.
     */
    private static ComparisonPredicate computeComparisonPredicate(
            TestFailure failure, IJUnitTestDescription testDesc) {
        ComparisonPredicate result = null;

        String excMessage = failure.exceptionMessage();

        if (excMessage != null) {
            if (!PredicateFactory.isCalledAssertEqualsOnObjects(failure
                    .thrownException())) {
                Matcher matcher = COMPARISON_PATTERN.matcher(excMessage);
                boolean matchPattern = matcher.matches();

                if (matchPattern) {
                    result = new ComparisonPredicate(failure, testDesc, matcher
                            .group(1), matcher.group(2));

                }
            }
        }

        return result;
    }

    /**
     * If this location is include in the stacktrace we cannot use the
     * comparison predicate.
     */
    private final static String ASSERTOBJECT_CLASS_NAME = "junit.framework.Assert";
    private final static String ASSERTOBJECT_METHOD_NAME = "assertEquals";
    private final static int ASSERTOBJECT_LINE_NUMBER = 71;

    /**
     * Returns true if the stacktrace of the exception contains the locations of
     * assertEquals on two objects, otherwise false.
     * 
     * @param exc
     * @return true if the stacktrace of the exception contains the locations of
     *         assertEquals on two objects, otherwise false.
     */
    private static boolean isCalledAssertEqualsOnObjects(Throwable exc) {
        boolean result = false;

        if (exc != null) {
            StackTraceElement[] stacktraceElements = exc.getStackTrace();

            for (int i = 0; i < stacktraceElements.length; i++) {
                StackTraceElement element = stacktraceElements[i];

                if (element.getClassName().equals(ASSERTOBJECT_CLASS_NAME)
                        && element.getMethodName().equals(
                                ASSERTOBJECT_METHOD_NAME)
                        && element.getLineNumber() == ASSERTOBJECT_LINE_NUMBER) {
                    result = true;
                    break;
                }
            }
        }

        return result;
    }

    /**
     * Creates a predicate for the exception.
     * 
     * @param exception
     *            the exception occured on test execution.
     * @param testDesc
     *            the description of the single junit test run.
     * @return the created predicate for the given exception.
     */
    public static ExceptionPredicate createPredicate(Throwable exception,
            IJUnitTestDescription testDesc) {
        return new ExceptionPredicate(exception, testDesc);
    }

    public static void main(String[] args) {
        // Test Pattern

    }
}
