/*
 * $Id: FormTextUtil.java,v 1.3 2005/08/24 17:32:01 lehmann Exp $
 * 
 * Copyright (C) 2005 Karsten Lehmann
 * 
 * Software Engineering Chair, Saarland University
 * Prof. Dr. Andreas Zeller
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 *
 * Created on Aug 17, 2005 3:21:05 PM by
 * @author Karsten Lehmann 
 * Mail to: lehmann@cs.uni-sb.de or KarstenLehmann@web.de
 */
package org.deltadebugging.ddstate.ui.views;

import org.deltadebugging.ddstate.logger.DDLogger;
import org.deltadebugging.ddstate.ui.DDStateUIPlugin;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.jdt.core.IField;
import org.eclipse.jdt.core.IJavaElement;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.IMethod;
import org.eclipse.jdt.core.IType;
import org.eclipse.jdt.core.JavaModelException;
import org.eclipse.jdt.core.Signature;
import org.eclipse.jdt.ui.JavaUI;
import org.eclipse.jface.text.BadLocationException;
import org.eclipse.jface.text.IDocument;
import org.eclipse.ui.IEditorPart;
import org.eclipse.ui.PartInitException;
import org.eclipse.ui.forms.events.HyperlinkAdapter;
import org.eclipse.ui.forms.events.HyperlinkEvent;
import org.eclipse.ui.forms.events.IHyperlinkListener;
import org.eclipse.ui.texteditor.IDocumentProvider;
import org.eclipse.ui.texteditor.ITextEditor;

/**
 * A util class with utility methods usefull in the form text.
 * 
 * @author <a href="mailto:KarstenLehmann@web.de">Karsten Lehmann </a>
 */
public class FormTextUtil {

    /** Personal logger for this class */
    final static DDLogger logger = DDStateUIPlugin
            .getLogger(FormTextUtil.class);

    private FormTextUtil() {
        // use the static methods
    }

    /**
     * Serial number to generate unambigious key for the hyperlinks.
     */
    private static long keySerialNumber = 0;

    public final static String FORM_BEGIN = "<form><p>";

    public final static String FORM_END = "</p></form>";

    /**
     * Cut the package name of the given type name.
     * 
     * @param typeName
     * @return only the type name without the package prefix.
     */
    public static String cutPackageOfTypeName(String typeName) {
        String result = typeName;

        if (result != null) {
            int lastIndex = result.lastIndexOf('.') + 1;
            result = result.substring(lastIndex, result.length());
        }

        return result;
    }

    /**
     * Open the type of the given type name in an editor. Use the given java
     * project to search the type.
     * 
     * @param javaProject
     *            the project, where the type must specified in the classpath.
     * @param typeName
     *            the full qualified name of the type.
     */
    public static void openType(IJavaProject javaProject, String typeName) {
        try {
            IType type = javaProject.findType(typeName);

            IEditorPart editor = JavaUI.openInEditor(type);
            JavaUI.revealInEditor(editor, type.getPrimaryElement());

        } catch (PartInitException e) {
            logger.log(IStatus.ERROR, "Failed open editor.", e);

        } catch (JavaModelException e) {
            logger.log(IStatus.ERROR, "Failed open editor.", e);

        }
    }

    public static void openField(IJavaProject javaProject, String typeName,
            String fieldName) {
        try {
            IType type = javaProject.findType(typeName);
            IField field = type.getField(fieldName);

            IEditorPart editor = JavaUI.openInEditor(type);
            JavaUI.revealInEditor(editor, field.getPrimaryElement());

        } catch (PartInitException e) {
            logger.log(IStatus.ERROR, "Failed open editor.", e);

        } catch (JavaModelException e) {
            logger.log(IStatus.ERROR, "Failed open editor.", e);

        }
    }

    /**
     * Open the type of the given type name in an editor. Use the given java
     * project to search the type. After that mark the given method in the open
     * editor.
     * 
     * @param javaProject
     *            the project, where the type must specified in the classpath.
     * @param typeName
     *            the full qualified name of the type.
     * @param methodName
     *            the name of the method which should marked.
     * @param methodParamTypeSignature
     *            the full type signature of all parameters.
     */
    public static void openMethod(IJavaProject javaProject, String typeName,
            String methodName, String[] methodParamTypeSignature) {
        try {
            IType type = javaProject.findType(typeName);

            IMethod method = type.getMethod(methodName,
                    methodParamTypeSignature);
            IMethod sourceMethods[] = type.findMethods(method);

            IEditorPart editor = JavaUI.openInEditor(type);

            if (sourceMethods != null && sourceMethods.length > 0) {
                JavaUI.revealInEditor(editor, (IJavaElement) sourceMethods[0]);
            }

        } catch (PartInitException e) {
            logger.log(IStatus.ERROR, "Failed open editor.", e);

        } catch (JavaModelException e) {
            logger.log(IStatus.ERROR, "Failed open editor.", e);

        }
    }

    /**
     * Open the type of the given type name in an editor. Use the given java
     * project to search the type. After that mark the given line in the open
     * editor.
     * 
     * @param javaProject
     *            the project, where the type must specified in the classpath.
     * @param typeName
     *            the full qualified name of the type.
     * @param lineNumber
     *            the line number of the line which should marked in the open
     *            editor.
     */
    public static void openLine(IJavaProject javaProject, String typeName,
            int lineNumber) {
        try {
            IType passType;
            passType = javaProject.findType(typeName);

            IEditorPart editor = JavaUI.openInEditor(passType);
            ITextEditor textEditor = (ITextEditor) editor
                    .getAdapter(ITextEditor.class);

            IDocumentProvider docProvider = textEditor.getDocumentProvider();
            if (docProvider == null)
                return;

            IDocument document = docProvider.getDocument(textEditor
                    .getEditorInput());
            if (document == null)
                return;

            int lastLine = document.getLineOfOffset(document.getLength()) + 1;

            // check if the given line number is valid
            if (lineNumber >= 1 && lineNumber <= lastLine) {
                int start = document.getLineOffset(lineNumber - 1);
                textEditor.selectAndReveal(start, 0);
            }

        } catch (PartInitException e) {
            logger.log(IStatus.ERROR, "Failed open editor.", e);

        } catch (JavaModelException e) {
            logger.log(IStatus.ERROR, "Failed open editor.", e);

        } catch (BadLocationException e) {
            logger.log(IStatus.ERROR, "Failed open editor.", e);

        }
    }

    /**
     * Announce the number of an enumeration. <br>
     * <ul>
     * <li>1 -> 1st</li>
     * <li>2 -> 2nd</li>
     * <li>3 -> 3rd</li>
     * <li>x -> xth</li>
     * </ul>
     * 
     * @param number
     *            the number to announce
     * @return the announced number.
     */
    public static String announceEnum(int number) {
        String result = null;

        switch (number) {
        case 1:
            result = "1st";
            break;

        case 2:
            result = "2nd";
            break;

        case 3:
            result = "3rd";
            break;

        default:
            result = number + "th";
            break;
        }

        return result;
    }

    /**
     * Use this method to append a hyperlink code used by the form text to the
     * string buffer. For this we need the href-key and the unquoted text to
     * display.
     * 
     * @param buf
     *            the buffer to append the hyperlink code
     * @param key
     *            the key of the hyperlink
     * @param text
     *            the text that should displayed by the form text
     * 
     * @see org.eclipse.ui.forms.widgets.FormText
     */
    public static void appendFormtTextHRef(StringBuffer buf, String key,
            String text) {
        if (buf == null) {
            throw new NullPointerException("Given buffer must not be null");
        }

        buf.append("<a href=\"" + key + "\">" + quoteFormTextString(text)
                + "</a>");
    }

    /**
     * Returns the quoted string so it can be used in a FormText field
     * 
     * @param text
     *            the text to quote
     * @return the quoted text for the FormText field
     * 
     * @see org.eclipse.ui.forms.widgets.FormText
     */
    public static String quoteFormTextString(String text) {
        String result = text;

        result = result.replaceAll("<", "&lt;");
        result = result.replaceAll(">", "&gt;");

        return result;
    }

    public static IHyperlinkListener addMethodHyperLink(StringBuffer buf,
            final IJavaProject javaProject, final String typeName,
            final String methodName, final String[] paramSignature) {

        String methodSignature = Signature.createMethodSignature(
                paramSignature, Signature.SIG_VOID);

        return addMethodHyperLink(buf, javaProject, typeName, methodName,
                methodSignature);
    }

    public static IHyperlinkListener addMethodHyperLink(StringBuffer buf,
            final IJavaProject javaProject, final String typeName,
            final String methodName, final String signature) {
        String strMethod = Signature.toString(signature, methodName, null,
                false, false);

        return addMethodHyperLink(buf, strMethod, javaProject, typeName,
                methodName, signature);
    }

    public static IHyperlinkListener addMethodHyperLink(StringBuffer buf,
            String text, final IJavaProject javaProject, final String typeName,
            final String methodName, final String signature) {
        IHyperlinkListener result = null;

        final String key = "methodname" + keySerialNumber++;

        FormTextUtil.appendFormtTextHRef(buf, key, text);

        result = new HyperlinkAdapter() {
            public void linkActivated(HyperlinkEvent e) {
                if (e.getHref().equals(key)) {
                    FormTextUtil.openMethod(javaProject, typeName, methodName,
                            Signature.getParameterTypes(signature.replace('/',
                                    '.')));
                }
            }
        };

        return result;
    }

    public static IHyperlinkListener addLineHyperLink(StringBuffer buf,
            final IJavaProject javaProject, final String typeName,
            final int lineNumber) {
        return addLineHyperLink(buf, "line " + lineNumber, javaProject,
                typeName, lineNumber);
    }

    public static IHyperlinkListener addLineHyperLink(StringBuffer buf,
            String text, final IJavaProject javaProject, final String typeName,
            final int lineNumber) {
        IHyperlinkListener result = null;

        final String key = "line" + keySerialNumber++;
        FormTextUtil.appendFormtTextHRef(buf, key, text);

        result = new HyperlinkAdapter() {
            public void linkActivated(HyperlinkEvent e) {
                if (e.getHref().equals(key)) {
                    FormTextUtil.openLine(javaProject, typeName, lineNumber);
                }
            }
        };

        return result;
    }

    public static IHyperlinkListener addTypeHyperLink(StringBuffer buf,
            final IJavaProject javaProject, final String typeName) {
        return addTypeHyperLink(buf, FormTextUtil
                .cutPackageOfTypeName(typeName), javaProject, typeName);
    }

    public static IHyperlinkListener addTypeHyperLink(StringBuffer buf,
            String text, final IJavaProject javaProject, final String typeName) {
        IHyperlinkListener result = null;

        final String key = "typename" + keySerialNumber++;
        FormTextUtil.appendFormtTextHRef(buf, key, text);

        result = new HyperlinkAdapter() {
            public void linkActivated(HyperlinkEvent e) {
                if (e.getHref().equals(key)) {
                    FormTextUtil.openType(javaProject, typeName);
                }
            }
        };

        return result;
    }

    public static IHyperlinkListener addFieldHyperLink(StringBuffer buf,
            final IJavaProject javaProject, final String typeName,
            final String fieldName) {
        String text = FormTextUtil.cutPackageOfTypeName(typeName) + "."
                + fieldName;
        return addFieldHyperLink(buf, text, javaProject, typeName, fieldName);
    }

    public static IHyperlinkListener addFieldHyperLink(StringBuffer buf,
            String text, final IJavaProject javaProject, final String typeName,
            final String fieldName) {
        IHyperlinkListener result = null;

        final String key = "fieldname" + keySerialNumber++;
        FormTextUtil.appendFormtTextHRef(buf, key, text);

        result = new HyperlinkAdapter() {
            public void linkActivated(HyperlinkEvent e) {
                if (e.getHref().equals(key)) {
                    FormTextUtil.openField(javaProject, typeName, fieldName);
                }
            }
        };

        return result;
    }

}
