/*
 * $Id: DeltaToRTextGenerator.java,v 1.22 2006/01/24 19:47:18 lehmann Exp $
 * 
 * Copyright (C) 2005 Karsten Lehmann
 * 
 * Software Engineering Chair, Saarland University
 * Prof. Dr. Andreas Zeller
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 *
 * Created on 18.08.2005 08:42:53 by
 * @author Karsten Lehmann 
 * Mail to: lehmann@cs.uni-sb.de or KarstenLehmann@web.de
 */
package org.deltadebugging.ddstate.ui.views.cts;

import org.deltadebugging.ddstate.memorygraph.IGraphEdge;
import org.deltadebugging.ddstate.memorygraph.ILocalVariableEdge;
import org.deltadebugging.ddstate.memorygraph.IMethodIdentifier;
import org.deltadebugging.ddstate.memorygraph.IStackFrame;
import org.deltadebugging.ddstate.memorygraph.delta.IMemoryGraphDelta;
import org.deltadebugging.ddstate.memorygraph.delta.create.CreateObjectDelta;
import org.deltadebugging.ddstate.memorygraph.delta.create.CreatePrimitiveValueDelta;
import org.deltadebugging.ddstate.memorygraph.delta.delete.DeleteDelta;
import org.deltadebugging.ddstate.memorygraph.delta.delete.DeleteObjectReferenceDelta;
import org.deltadebugging.ddstate.memorygraph.delta.location.CreateObjectDeltaLocation;
import org.deltadebugging.ddstate.memorygraph.delta.location.IDeltaLocation;
import org.deltadebugging.ddstate.memorygraph.delta.location.NullLocation;
import org.deltadebugging.ddstate.memorygraph.delta.location.PathLocation;
import org.deltadebugging.ddstate.memorygraph.delta.update.UpdateArraySizeDelta;
import org.deltadebugging.ddstate.memorygraph.delta.update.UpdateObjectReferenceDelta;
import org.deltadebugging.ddstate.memorygraph.delta.update.UpdateObjectValueDelta;
import org.deltadebugging.ddstate.memorygraph.type.IGlobalVariableEdge;
import org.deltadebugging.ddstate.memorygraph.type.IReferenceTypeNode;
import org.deltadebugging.ddstate.memorygraph.util.Path;
import org.deltadebugging.ddstate.memorygraph.value.IArrayElementEdge;
import org.deltadebugging.ddstate.memorygraph.value.IArrayNode;
import org.deltadebugging.ddstate.memorygraph.value.IObjectNode;
import org.deltadebugging.ddstate.memorygraph.value.IStringObjectNode;
import org.deltadebugging.ddstate.memorygraph.value.IValueNode;
import org.deltadebugging.ddstate.memorygraph.value.IObjectFieldEdge;
import org.deltadebugging.ddstate.ui.views.FormTextUtil;
import org.deltadebugging.ddstate.ui.views.GraphDeltaTableViewer;
import org.eclipse.jdt.core.IJavaProject;
import org.eclipse.jdt.core.Signature;
import org.eclipse.swt.widgets.Composite;
import org.eclipse.ui.forms.events.IHyperlinkListener;
import org.eclipse.ui.forms.widgets.FormText;
import org.eclipse.ui.forms.widgets.FormToolkit;

public class DeltaToRTextGenerator {

    IJavaProject fJavaProject;

    private final FormToolkit fToolkit;

    public DeltaToRTextGenerator(FormToolkit toolkit) {
        this.fToolkit = toolkit;
    }

    public void setJavaProject(IJavaProject javaProject) {
        this.fJavaProject = javaProject;
    }

    public FormText createFormText(Composite parent, IMemoryGraphDelta delta) {
        return this.createFormText(parent, delta, "", "");
    }

    public FormText createFormText(Composite parent, IMemoryGraphDelta delta,
            String prefix, String suffix) {

        FormText result = null;

        if (delta instanceof UpdateObjectValueDelta) {
            result = this.createUpdateValueFormText(parent,
                    (UpdateObjectValueDelta) delta, prefix, suffix);

        } else if (delta instanceof DeleteDelta) {
            result = this.createDeleteFormText(parent, (DeleteDelta) delta,
                    prefix, suffix);

        } else if (delta instanceof CreateObjectDelta) {
            result = this.createCreateObjectFormText(parent,
                    (CreateObjectDelta) delta, prefix, suffix);

        } else if (delta instanceof UpdateObjectReferenceDelta) {
            result = this.createUpdateObjectReferenceFormText(parent,
                    (UpdateObjectReferenceDelta) delta, prefix, suffix);

        } else if (delta instanceof CreatePrimitiveValueDelta) {
            result = this.createAddPrimitiveValueFormText(parent,
                    (CreatePrimitiveValueDelta) delta, prefix, suffix);

        }

        if (result == null) {
            result = this.fToolkit.createFormText(parent, true);
            result.setText("No presentation found for "
                    + delta.getClass().getName() + "\nLoc: "
                    + delta.getLocation().toString(), false, false);
        }

        return result;
    }

    private FormText createDeleteFormText(Composite parent,
            final DeleteDelta delta, String prefix, String suffix) {
        FormText result = this.fToolkit.createFormText(parent, true);
        IGraphEdge edgeTovalueToDelete = delta.getVariableToDelete();

        StringBuffer buf = new StringBuffer();
        buf.append(FormTextUtil.FORM_BEGIN);
        buf.append(prefix);

        if (delta instanceof DeleteObjectReferenceDelta) {
            buf.append("object reference of ");
            result.addHyperlinkListener(this.addVariable(buf,
                    edgeTovalueToDelete));
            buf.append(" is now null");

        } else {
            result.addHyperlinkListener(this.addVariable(buf,
                    edgeTovalueToDelete));
            buf.append(" is now the default value");
        }

        buf.append(suffix);
        buf.append(FormTextUtil.FORM_END);

        result.setText(buf.toString(), true, false);

        return result;
    }

    private FormText createUpdateObjectReferenceFormText(Composite parent,
            final UpdateObjectReferenceDelta delta, String prefix, String suffix) {
        FormText result = this.fToolkit.createFormText(parent, true);

        StringBuffer buf = new StringBuffer();
        buf.append(FormTextUtil.FORM_BEGIN);
        buf.append(prefix);
        buf.append("update object reference from ");

        IHyperlinkListener oldLocationListener = this.addDeltaLocation(buf,
                delta.getLocation());
        if (oldLocationListener != null) {
            result.addHyperlinkListener(oldLocationListener);
        }

        buf.append(" to ");

        IHyperlinkListener newLocationListener = this.addDeltaLocation(buf,
                delta.getNewLocation());
        if (newLocationListener != null) {
            result.addHyperlinkListener(newLocationListener);
        }

        buf.append(suffix);
        buf.append(FormTextUtil.FORM_END);

        result.setText(buf.toString(), true, false);

        return result;
    }

    private FormText createCreateObjectFormText(Composite parent,
            final CreateObjectDelta delta, String prefix, String suffix) {
        FormText result = this.fToolkit.createFormText(parent, true);

        StringBuffer buf = new StringBuffer();
        buf.append(FormTextUtil.FORM_BEGIN);
        buf.append(prefix);
        buf.append("new ");

        IHyperlinkListener listener = this.addCreatedVariable(buf, delta);
        if (listener != null) {
            result.addHyperlinkListener(listener);
        }

        buf.append(" was created");
        buf.append(suffix);
        buf.append(FormTextUtil.FORM_END);

        result.setText(buf.toString(), true, false);

        return result;
    }

    private FormText createAddPrimitiveValueFormText(Composite parent,
            final CreatePrimitiveValueDelta delta, String prefix, String suffix) {
        FormText rtext = this.fToolkit.createFormText(parent, true);

        StringBuffer buf = new StringBuffer();
        buf.append(FormTextUtil.FORM_BEGIN);
        buf.append(prefix);

        final boolean isLocalVariable = delta.getValueToCreate()
                .isLocalVariable();

        rtext.addHyperlinkListener(this.addVariable(buf, delta
                .getValueToCreate()));
        buf.append(" is now ");
        buf.append(FormTextUtil.quoteFormTextString(GraphDeltaTableViewer
                .getNewValueAsString(delta)));

        if (isLocalVariable) {
            ILocalVariableEdge localVarEdge = delta.getValueToCreate()
                    .getLocalVariablePrecessors()[0];
            IStackFrame stackFrame = localVarEdge.getStackFrame();
            IMethodIdentifier methodIdentifier = stackFrame
                    .getMethodIdentifier();

            String typeName = FormTextUtil
                    .cutPackageOfTypeName(methodIdentifier.getTypeName());
            String method = Signature.toString(methodIdentifier.getSignature(),
                    methodIdentifier.getMethodName(), null, false, false);

            String frameText = "frame #" + stackFrame.getFrameNumber() + " - "
                    + typeName + "." + method + " line "
                    + stackFrame.getLineNumber();

            buf.append(" (");
            rtext.addHyperlinkListener(FormTextUtil.addLineHyperLink(buf,
                    frameText, this.fJavaProject, methodIdentifier
                            .getTypeName(), stackFrame.getLineNumber()));
            buf.append(")");
        }

        buf.append(suffix);
        buf.append(FormTextUtil.FORM_END);

        rtext.setText(buf.toString(), true, false);

        return rtext;
    }

    private FormText createUpdateValueFormText(Composite parent,
            final UpdateObjectValueDelta delta, String prefix, String suffix) {
        FormText rtext = this.fToolkit.createFormText(parent, true);

        StringBuffer buf = new StringBuffer();
        buf.append(FormTextUtil.FORM_BEGIN);
        buf.append(prefix);

        final boolean isLocalVariable = delta.getNewValue().isLocalVariable();

        if (delta instanceof UpdateArraySizeDelta) {
            buf.append("array size of ");
        }

        rtext.addHyperlinkListener(this.addVariable(buf, delta.getNewValue()));
        buf.append(" is now ");
        buf.append(FormTextUtil.quoteFormTextString(GraphDeltaTableViewer
                .getNewValueAsString(delta)));

        if (isLocalVariable) {
            ILocalVariableEdge localVarEdge = delta.getNewValue()
                    .getLocalVariablePrecessors()[0];
            IStackFrame stackFrame = localVarEdge.getStackFrame();
            IMethodIdentifier methodIdentifier = stackFrame
                    .getMethodIdentifier();

            String typeName = FormTextUtil
                    .cutPackageOfTypeName(methodIdentifier.getTypeName());
            String method = Signature.toString(methodIdentifier.getSignature(),
                    methodIdentifier.getMethodName(), null, false, false);

            String frameText = "frame #" + stackFrame.getFrameNumber() + " - "
                    + typeName + "." + method + " line "
                    + stackFrame.getLineNumber();

            buf.append(" (");
            rtext.addHyperlinkListener(FormTextUtil.addLineHyperLink(buf,
                    frameText, this.fJavaProject, methodIdentifier
                            .getTypeName(), stackFrame.getLineNumber()));
            buf.append(")");
        }

        buf.append(suffix);
        buf.append(FormTextUtil.FORM_END);

        rtext.setText(buf.toString(), true, false);

        return rtext;
    }

    private IHyperlinkListener addDeltaLocation(StringBuffer buf,
            IDeltaLocation deltaLocation) {
        IHyperlinkListener result = null;

        if (deltaLocation instanceof NullLocation) {
            buf.append("null");

        } else if (deltaLocation instanceof CreateObjectDeltaLocation) {
            buf.append("new created ");
            result = this.addCreatedVariable(buf,
                    ((CreateObjectDeltaLocation) deltaLocation)
                            .getCreateOjbectDelta());

        } else if (deltaLocation instanceof PathLocation) {
            Path path = ((PathLocation) deltaLocation).getPath();

            if (path.size() > 0) {
                IGraphEdge lastEdge = path.getLastPathElement();

                result = this.addVariable(buf, lastEdge);
            } else {
                buf.append("<path location contains no elements>");
            }
        }

        return result;
    }

    private IHyperlinkListener addCreatedVariable(StringBuffer buf,
            CreateObjectDelta createDelta) {
        IHyperlinkListener result = null;

        final IValueNode valueToCreate = createDelta.getValueToCreate();

        if (valueToCreate instanceof IStringObjectNode) {
            buf.append("string \""
                    + ((IStringObjectNode) valueToCreate).getStringValue()
                    + "\"");

        } else if (valueToCreate instanceof IObjectNode) {
            buf.append("object of type ");
            result = FormTextUtil.addTypeHyperLink(buf, this.fJavaProject,
                    valueToCreate.getReferenceTypeName());

        } else if (valueToCreate instanceof IArrayNode) {
            IArrayNode arrRef = (IArrayNode) valueToCreate;

            buf.append("array with length " + arrRef.getLength() + " of type ");
            result = FormTextUtil.addTypeHyperLink(buf, this.fJavaProject,
                    valueToCreate.getReferenceTypeName());
        }

        return result;
    }

    private IHyperlinkListener addVariable(StringBuffer buf,
            IValueNode graphNode) {
        IHyperlinkListener result = null;

        IGraphEdge edgeToVariable = null;

        if (graphNode.isGlobalVariable()) {
            edgeToVariable = graphNode.getGlobalVariablePredecessors()[0];

        } else if (graphNode.isLocalVariable()) {
            edgeToVariable = graphNode.getLocalVariablePrecessors()[0];

        } else if (graphNode.isArrayElementVariable()) {
            edgeToVariable = graphNode.getArrayElementPredecessors()[0];

        } else if (graphNode.isObjectReferenceFieldVariable()) {
            edgeToVariable = graphNode.getObjectReferenceFieldPredecessors()[0];

        } else {
            edgeToVariable = (IGraphEdge) graphNode.getParents().iterator().next();

        }

        result = this.addVariable(buf, edgeToVariable);

        return result;
    }

    private IHyperlinkListener addVariable(StringBuffer buf,
            IGraphEdge graphEdge) {
        IHyperlinkListener result = null;

        if (graphEdge instanceof IGlobalVariableEdge) {
            buf.append("global variable ");

            IGlobalVariableEdge globalVarEdge = (IGlobalVariableEdge) graphEdge;
            IReferenceTypeNode sourceNode = globalVarEdge.getTypeNode();

            String text = FormTextUtil.cutPackageOfTypeName(sourceNode
                    .getTypeName())
                    + "." + globalVarEdge.getName();
            result = FormTextUtil.addFieldHyperLink(buf, text,
                    this.fJavaProject, sourceNode.getTypeName(), globalVarEdge
                            .getName());

        } else if (graphEdge instanceof ILocalVariableEdge) {
            buf.append("local variable ");

            ILocalVariableEdge localVarEdge = (ILocalVariableEdge) graphEdge;
            IStackFrame stackFrame = localVarEdge.getStackFrame();

            result = FormTextUtil.addLineHyperLink(buf, graphEdge.getName(),
                    this.fJavaProject, stackFrame.getMethodIdentifier()
                            .getTypeName(), stackFrame.getLineNumber());

        } else if (graphEdge instanceof IArrayElementEdge) {
            IArrayElementEdge arrayElementEdge = (IArrayElementEdge) graphEdge;

            result = this.addVariable(buf, (IValueNode) arrayElementEdge
                    .getSourceNode());
            buf.append("[" + arrayElementEdge.getArrayIndex() + "]");

        } else if (graphEdge instanceof IObjectFieldEdge) {
            IObjectFieldEdge fieldEdge = (IObjectFieldEdge) graphEdge;

            String declTypeName = fieldEdge.getDeclaringTypeName();
            String fieldName = fieldEdge.getFieldName();

            buf.append("field variable ");
            result = FormTextUtil.addFieldHyperLink(buf, this.fJavaProject,
                    declTypeName, fieldName);

        } else {
            buf.append("variable " + graphEdge.getName());
        }

        return result;
    }
}
