/*
 * $Id: AbstractGraphVisualizer.java,v 1.25 2005/10/27 16:10:14 lehmann Exp $
 * 
 * Copyright (C) 2004 Karsten Lehmann
 * 
 * Software Engineering Chair, Saarland University
 * Prof. Dr. Andreas Zeller
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 *
 * Created on Oct 1, 2004 11:01:14 AM by
 * @author Karsten Lehmann 
 * Mail to: lehmann@cs.uni-sb.de or KarstenLehmann@web.de
 */
package org.deltadebugging.ddstate.visualizer;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Vector;

import org.deltadebugging.ddstate.comparator.delta.GraphDecorator;
import org.deltadebugging.ddstate.logger.DDLogger;
import org.deltadebugging.ddstate.memorygraph.IGraphNode;
import org.deltadebugging.ddstate.memorygraph.IRootNode;
import org.deltadebugging.ddstate.memorygraph.filter.GraphFilterException;
import org.deltadebugging.ddstate.memorygraph.filter.GraphFilterManager;
import org.deltadebugging.ddstate.visualizer.attributes.GraphVizAttribute;
import org.deltadebugging.ddstate.visualizer.ui.preferences.GraphVizPreferencePage;
import org.deltadebugging.ddstate.visualizer.vizgraph.VizGraphEdge;
import org.deltadebugging.ddstate.visualizer.vizgraph.VizGraphNode;
import org.deltadebugging.ddstate.visualizer.vizgraph.VizRootNode;
import org.deltadebugging.ddstate.visualizer.vizgraph.exceptions.GraphVisualizerException;
import org.eclipse.core.resources.IProject;
import org.eclipse.core.resources.IProjectDescription;
import org.eclipse.core.resources.IResource;
import org.eclipse.core.resources.IWorkspace;
import org.eclipse.core.resources.ResourcesPlugin;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.NullProgressMonitor;
import org.eclipse.core.runtime.Path;
import org.eclipse.core.runtime.Preferences;
import org.eclipse.jface.dialogs.ErrorDialog;
import org.eclipse.jface.preference.IPreferenceNode;
import org.eclipse.jface.preference.IPreferencePage;
import org.eclipse.jface.preference.PreferenceDialog;
import org.eclipse.jface.preference.PreferenceManager;
import org.eclipse.jface.preference.PreferenceNode;
import org.eclipse.swt.widgets.Display;
import org.eclipse.ui.PartInitException;

import att.grappa.Edge;
import att.grappa.Element;
import att.grappa.Graph;
import att.grappa.Node;

/**
 * COMMENTME Comment type AbstractGraphVisualizer
 * 
 * @author <a href="mailto:KarstenLehmann@web.de">Karsten Lehmann </a>
 */
public abstract class AbstractGraphVisualizer {

    // Personal logger for this class
    private final static DDLogger logger = VisualizerPlugin
            .getLogger(AbstractGraphVisualizer.class);

    protected final static boolean PRESENTATION = false;
    public final static boolean THESIS = true;

    /**
     * COMMENTME
     */
    private static final String PROJECT_NAME_MEMORY_GRAPHS = "Generated Memory Graphs";

    /**
     * The directory where the output file will be written.
     */
    private String fOutputDir = null;

    /**
     * Determine if the full qualified names will be show or only a short
     * version.
     */
    private boolean fShowFullQualifiedNames = true;

    /**
     * If true the hidden successor nodes will combine in one node. The label of
     * the node is the count of the hidden nodes.
     */
    private boolean fShowCountForHiddenNodes = false;

    /**
     * The date format to parse a date for a graph comment.
     */
    public final static SimpleDateFormat DATE_FORMAT_COMMENT = new SimpleDateFormat(
            "yyyy-MM-dd HH:mm");

    /**
     * The date format "yyyy-MM-dd".
     */
    public final static SimpleDateFormat DATE_FORMAT_DATE = new SimpleDateFormat(
            "yyyy-MM-dd");

    /**
     * The date format "HH-mm-ss".
     */
    public final static SimpleDateFormat DATE_FORMAT_TIME = new SimpleDateFormat(
            "HH-mm-ss");

    /**
     * This Map is used to avoid cycles in the creation the visualization of the
     * graph. In the Map are all references to the Graphelements stored that
     * have already been visited. At the end of the visualisation the Map have
     * to be reset.
     */
    private Map fAlreadyVisitMap = new HashMap();

    /**
     * This Map is used to mark if a node has a special 'unfolded node' or not.
     * The unfolded nodes will be drawn if a node has successors which was
     * filtered out. But it should always draw one 'unfolded node' per node.
     */
    private Map fUnfoldedNodeMap = new HashMap();

    /**
     * Only for statistics.
     */
    private long fNodeCounter = -1;

    private long fEdgeCounter = -1;

    /**
     * A local instance of the <code>GraphFilterManager</code>.
     */
    private final GraphFilterManager fFilterManager = new GraphFilterManager();

    /**
     * Returns true if the full qualified names will be show and false if only a
     * short version will be show.
     * 
     * @return true for full qualified node names, else false.
     */
    abstract boolean showFullQualifiedNames();

    /**
     * @return true if the hidden successor nodes should combine in one node.
     *         The label of these node is the count of the hidden nodes.
     */
    abstract boolean showCountForHiddenNodes();

    /**
     * Returns a vector with all graph filters to use for the visualizations.
     * Note that every graph filter must implement the
     * <code>IGraphNodeFilter</code> -Interface.
     * 
     * @return a vector containing all graph filters.
     */
    abstract Vector getGraphFilters();

    /**
     * Returns the directory where the output file will be written.
     * 
     * @return the directory where the output file will be written.
     */
    abstract String getOutputDirectory();

    abstract String getNodeContent(VizGraphNode graphNode,
            boolean parShowFullQualifiedNames) throws GraphVisualizerException;

    abstract Iterator getNodeAttributes(VizGraphNode graphNode);

    abstract String getEdgeContent(VizGraphEdge graphEdge,
            boolean parShowFullQualifiedNames);

    abstract Iterator getEdgeAttributes(VizGraphEdge graphEdge);

    abstract Iterator getAllSuccessors(VizGraphNode graphNode)
            throws GraphVisualizerException;

    /**
     * Returns the preferences for the <code>VisualizerPlugin</code>.
     * 
     * @return the preferences for the <code>VisualizerPlugin</code>.
     */
    public static Preferences getPluginPreferences() {
        return VisualizerPlugin.getDefault().getPluginPreferences();
    }

    /**
     * Initialize the local attributes befor a new visualization will be start.
     * 
     */
    private void initialize() {
        // set the node and the edge counter to his intial values
        this.fNodeCounter = 0;
        this.fEdgeCounter = 0;

        // clear the Maps
        this.fAlreadyVisitMap.clear();
        this.fUnfoldedNodeMap.clear();

        // remember this setting for the visualization
        this.fShowFullQualifiedNames = this.showFullQualifiedNames();
        this.fOutputDir = this.getOutputDirectory();
        this.fShowCountForHiddenNodes = this.showCountForHiddenNodes()
                && !PRESENTATION;

        // reset the filter manager and read the new filters
        this.fFilterManager.removeAllFilter();
        this.fFilterManager.addAllFilter(this.getGraphFilters());

    }

    public void visualizeAndOpen(IRootNode rootNode, String fileName)
            throws GraphVisualizerException {

        this.initialize();

        String path = null;

        try {
            path = this.visualize(rootNode, fileName);

        } catch (GraphVisualizerException e) {
            throw e;

        }

        IProject project;
        try {
            project = this.getMemGraphProject();
            project.refreshLocal(IResource.DEPTH_INFINITE,
                    new NullProgressMonitor());
        } catch (PartInitException e) {
            logger.log(IStatus.WARNING,
                    "Cannot create or initialize external editor to open the file '"
                            + path + "'");
        } catch (CoreException e) {
            logger.log(IStatus.WARNING, "Cannot open file '" + path
                    + "' with system editor", e);

        }
    }

    // COMMENTME getMemGraphProject
    private IProject getMemGraphProject() throws CoreException {
        IWorkspace ws = ResourcesPlugin.getWorkspace();
        IProject project = ws.getRoot().getProject(PROJECT_NAME_MEMORY_GRAPHS);
        IProjectDescription projectDesc = ws
                .newProjectDescription("Memory Graph Project");
        projectDesc.setComment("Visualization of the extracted memory graphs");
        projectDesc.setLocation(new Path(this.getOutputDirectory()));

        if (!project.exists()) {
            project.create(projectDesc, null);
        }

        if (!project.isOpen())
            project.open(IResource.BACKGROUND_REFRESH, null);

        return project;
    }

    /**
     * Start the visualization of the given graph and write the output to the
     * specified directory. If the visualization is disabled in the preference
     * page this method returns <code>null</code>.
     * 
     * @param rootNode
     *            the root node of the graph to visualize.
     * @param fileName
     *            the name of the output file.
     * @return the file name (full qualified).
     * @throws GraphVisualizerException
     *             if the visualization of the graph failed.
     */
    public String visualize(IRootNode rootNode, String fileName)
            throws GraphVisualizerException {
        assert rootNode != null;

        this.initialize();

        boolean isPassingGraph = rootNode.isPassingGraph();

        logger.traceAndTimeStart(VisualizerPlugin.DEBUG, "GraphViz",
                "Start visualizing Graph ...");
        final String path = this.fOutputDir + "/" + fileName;

        Date now = new Date(System.currentTimeMillis());

        try {
            Graph vizGraph = createNewGraph(now);

            // Decorate the root node
            VizGraphNode vizRootNode = new VizRootNode(rootNode);

            // add the root node to the graph
            Node vizNode = this.createNewNode(vizGraph, vizRootNode);
            vizGraph.addNode(vizNode);

            // add all successors
            this.addAllSuccessors(vizGraph, vizRootNode, vizNode,
                    isPassingGraph);

            logger.traceAndTimeEnd(VisualizerPlugin.DEBUG, "GraphViz",
                    "Graph visualized (#nodes = " + this.fNodeCounter
                            + ", #edges = " + this.fEdgeCounter + ")");

            try {
                File dir = new File(path).getParentFile();
                if (!dir.exists()) {
                    dir.mkdirs();
                }

                vizGraph.printGraph(new BufferedWriter(new FileWriter(path)));
            } catch (IOException e) {
                final IStatus status = logger.log(IStatus.ERROR,
                        "IOException: Cannot write the graph in " + path, e);

                Display.getDefault().asyncExec(new Runnable() {

                    public void run() {
                        ErrorDialog
                                .openError(
                                        VisualizerPlugin
                                                .getActiveWorkbenchShell(),
                                        "Visualization problem",
                                        "The graph cannot store on the given location ("
                                                + path
                                                + "). Try to specify an other location in the preference page.",
                                        status);
                        showPreferencePage();
                    }

                });

            }

            logger.trace(VisualizerPlugin.DEBUG_GRAPH_VIUSALIZER,
                    "Write Graph in " + path);

        } catch (Throwable t) {
            // Handle unexpected exceptions. This was necessary because on the
            // MAC it occured an ClassNotFoundException that crashes the
            // application. If we handle this exception we can continue without
            // the whole application crashes.

            throw new GraphVisualizerException(
                    "Unexpected exception on visualizing graph.", t);
        }

        return path;
    }

    void showPreferencePage() {
        final IPreferencePage page = new GraphVizPreferencePage();
        page.setTitle("Graph-Visualizer");
        PreferenceManager manager = new PreferenceManager();
        IPreferenceNode node = new PreferenceNode("1", page);
        manager.addToRoot(node);
        final PreferenceDialog dialog = new PreferenceDialog(VisualizerPlugin
                .getActiveWorkbenchShell(), manager);

        Display.getDefault().syncExec(new Runnable() {
            public void run() {
                dialog.create();
                dialog.setMessage(page.getTitle());
                dialog.open();
            }
        });
    }

    /**
     * Return the target node of the edge. The Problem is that the edge can come
     * from two graphs if we merge two. If we don't want to draw nodes more than
     * one time we have to give the right node reference back.
     * 
     * @param graphEdge
     * @param isPassingGraph
     * @return target node of the graph edge.
     */
    private VizGraphNode getTargetNode(VizGraphEdge graphEdge,
            boolean isPassingGraph) {

        VizGraphNode result = graphEdge.getTargetNode();

        if (graphEdge.isMaskedAs(GraphDecorator.ONLY_IN_FIRSTGRAPH)
                && !isPassingGraph) {
            if (result.hasAssociatedElement()) {
                result = (VizGraphNode) result.getAssociatedElement();
            }
        } else if (graphEdge.isMaskedAs(GraphDecorator.ONLY_IN_SECONDGRAPH)
                && isPassingGraph) {
            if (result.hasAssociatedElement()) {
                result = (VizGraphNode) result.getAssociatedElement();
            }
        }

        return result;
    }

    // COMMENTME addAllSuccessors
    private void addAllSuccessors(Graph vizGraph, VizGraphNode graphNode,
            Node vizSourceNode, boolean isPassingGraph)
            throws GraphVisualizerException {

        Node vizTargetNode = null;

        // if graphNode is a String node create no successor if this was defined
        // in the preference page
        for (Iterator itSuccessors = this.getAllSuccessors(graphNode); itSuccessors
                .hasNext();) {
            VizGraphEdge successor = (VizGraphEdge) itSuccessors.next();

            VizGraphNode targetNode = this.getTargetNode(successor,
                    isPassingGraph);

            // look if the targetNode already visualized
            vizTargetNode = this.isAlreadyVisit((IGraphNode) targetNode
                    .getGraphElement());

            boolean filterNode = true;
            try {
                filterNode = this.fFilterManager.filterObject(targetNode
                        .getGraphElement());
            } catch (GraphFilterException e) {
                throw new GraphVisualizerException("Cannot apply graph filter",
                        e);
            }

            if (!filterNode) {

                // only draw a new node if the node not exists
                if (vizTargetNode == null) {

                    vizTargetNode = this.createNewNode(vizGraph, targetNode);

                    // We add the node to the viz graph before we create all
                    // successors!
                    // Because a successor of the node can point to this
                    // node too and we want to avoid the multiple creation
                    // of the same graph node.
                    vizGraph.addNode(vizTargetNode);

                    this.addAllSuccessors(vizGraph, targetNode, vizTargetNode,
                            isPassingGraph);

                }

                vizGraph.addEdge(createNewEdge(vizGraph, vizSourceNode,
                        vizTargetNode, successor));

            } else {

                if (this.fShowCountForHiddenNodes) {
                    this.addNodeForHiddenNodes(vizGraph, vizSourceNode);
                }
            }
        }
    }

    /**
     * Creates a new GraphViz node in the viGraph and use therefor the
     * informations described in the graphNode.
     * 
     * @param vizGraph
     *            the graph which the new node will be added.
     * @param graphNode
     *            the graph node specified in the memory graph.
     * @return a new graph node
     * @throws GraphVisualizerException
     *             if the creation of the node failed
     */
    private Node createNewNode(Graph vizGraph, VizGraphNode graphNode)
            throws GraphVisualizerException {

        Node vizNode = new Node(vizGraph);
        vizNode.setAttribute("label", this.getNodeContent(graphNode,
                this.fShowFullQualifiedNames));

        this.addAttributesToGraphElement(vizNode, this
                .getNodeAttributes(graphNode));

        this.registerNodeAllreadyVisit(
                (IGraphNode) graphNode.getGraphElement(), vizNode);

        logger.trace(VisualizerPlugin.DEBUG_GRAPH_VIUSALIZER, "Put node '"
                + graphNode.getGraphVizName(this.fShowFullQualifiedNames)
                + "' in the map");

        this.fNodeCounter++;

        return vizNode;
    }

    private void addNodeForHiddenNodes(Graph vizGraph, Node nodePredecessor) {
        assert vizGraph != null;
        assert nodePredecessor != null;

        Node vizUnfoldedNode = (Node) this.fUnfoldedNodeMap.get(nodePredecessor
                .getIdKey());

        // only add a pseudo node if not any exists for this node
        if (vizUnfoldedNode == null) {

            vizUnfoldedNode = new Node(vizGraph);
            vizUnfoldedNode.setAttribute("label", "1");

            // create a new edge from the predecessor to this node
            new Edge(vizGraph, nodePredecessor, vizUnfoldedNode);

            this.fUnfoldedNodeMap.put(nodePredecessor.getIdKey(),
                    vizUnfoldedNode);

            logger.trace(VisualizerPlugin.DEBUG_GRAPH_VIUSALIZER,
                    "Put special Unfolded-node (+) in the map");

        } else {
            int count = new Integer(vizUnfoldedNode.getAttribute("label")
                    .getStringValue()).intValue();
            vizUnfoldedNode.setAttribute("label", new Integer(count + 1)
                    .toString());
        }
    }

    /**
     * Creates a new GraphViz edge in the vizGraph from vizSourceNode to
     * vizTargetNode and use therefor the informations described in the
     * graphEdge.
     * 
     * @param vizGraph
     *            the graph which the new edge will be added.
     * @param vizSourceNode
     *            the source node of the edge.
     * @param vizTargetNode
     *            the target node of the edge.
     * @param graphEdge
     *            the graph edge specified in the memory graph.
     * @return a new graph edge.
     */
    private Edge createNewEdge(Graph vizGraph, Node vizSourceNode,
            Node vizTargetNode, VizGraphEdge graphEdge) {
        Edge vizEdge = new Edge(vizGraph, vizSourceNode, vizTargetNode);

        vizEdge.setAttribute("label", this.getEdgeContent(graphEdge,
                this.fShowFullQualifiedNames));

        this.addAttributesToGraphElement(vizEdge, this
                .getEdgeAttributes(graphEdge));

        this.fEdgeCounter++;

        return vizEdge;
    }

    /**
     * Creates a new graph specified in the GraphViz java library.
     * 
     * @param date
     *            the description date of the graph.
     * @return a new graph.
     */
    private Graph createNewGraph(Date date) {
        Graph vizGraph = new Graph("G", true, false);

        if (!AbstractGraphVisualizer.PRESENTATION && !AbstractGraphVisualizer.THESIS) {
            String label = "\\n\\nMemory graph created by DDstate (Created on "
                    + DATE_FORMAT_COMMENT.format(date) + ")\\n"
                    + "Software Engineering Chair, Saarland University";

            vizGraph.setAttribute("label", label);
            vizGraph.setAttribute("fontsize", "30");
            vizGraph.setAttribute("fontname", "Helvetica-Oblique");
        }

        return vizGraph;
    }

    /**
     * Set all attributes specified in the graphElement to the specified
     * vizElement.
     * 
     * @param vizElement
     *            a GraphViz graph element like Edge, Node or Graph.
     * @param attributes
     *            an iterator containing <code>GraphVizAttribute</code>s.
     */
    private void addAttributesToGraphElement(Element vizElement,
            Iterator attributes) {
        GraphVizAttribute attr = null;

        for (Iterator itAttr = attributes; itAttr.hasNext();) {
            attr = (GraphVizAttribute) itAttr.next();

            vizElement.setAttribute(attr.getName(), attr.getValue());
        }
    }

    private void registerNodeAllreadyVisit(IGraphNode graphNode, Node vizNode) {
        this.fAlreadyVisitMap.put(graphNode, vizNode);
    }

    private Node isAlreadyVisit(IGraphNode graphNode) {
        return (Node) this.fAlreadyVisitMap.get(graphNode);
    }

}