package org.deltadebugging.ddstate.visualizer;

/*
 * $Id: DDRunVisualizeListener.java,v 1.7 2005/09/28 09:20:35 lehmann Exp $
 * 
 * Copyright (C) 2004 Karsten Lehmann
 * 
 * Software Engineering Chair, Saarland University
 * Prof. Dr. Andreas Zeller
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 *
 * Created on 27.09.2004 16:37:17 by
 * @author Karsten Lehmann 
 * Mail to: lehmann@cs.uni-sb.de or KarstenLehmann@web.de
 */

import java.io.File;
import java.util.Date;
import java.util.HashMap;
import java.util.Map;

import org.deltadebugging.ddcore.DeltaSet;
import org.deltadebugging.ddstate.launcher.client.IJUnitTestDescription;
import org.deltadebugging.ddstate.launcher.ddrun.IDDRunDescription;
import org.deltadebugging.ddstate.launcher.ddrun.IDDRunListener;
import org.deltadebugging.ddstate.launcher.ddrun.IDDRunResult;
import org.deltadebugging.ddstate.logger.DDLogger;
import org.deltadebugging.ddstate.memorygraph.IRootNode;
import org.deltadebugging.ddstate.visualizer.vizgraph.exceptions.GraphVisualizerException;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Preferences;

/**
 * This listener observe the delta debugging run and visualize the memory graphs
 * if they are extraced and if they are compared.
 * 
 * @author <a href="mailto:KarstenLehmann@web.de">Karsten Lehmann </a>
 */
public class DDRunVisualizeListener implements IDDRunListener {

    /** Personal logger for this class */
    private final static DDLogger logger = VisualizerPlugin
            .getLogger(DDRunVisualizeListener.class);

    /** Tool to visualize a memory graph */
    private final GraphVisualizer fVisualizer = new GraphVisualizer();

    /** Tool to visualize memory graph differences */
    private final GraphDiffVisualizer fDiffVisualizer = new GraphDiffVisualizer();

    /** A map to remember the dd run description */
    private final Map fDDRunDesc = new HashMap();

    /*
     * (non-Javadoc)
     * 
     * @see org.deltadebugging.ddstate.launcher.ddrun.IDDRunListener#ddRunStarted(org.deltadebugging.ddstate.launcher.ddrun.IDDRunDescription)
     */
    public void ddRunStarted(IDDRunDescription desc) {
        this.fDDRunDesc.put(new Integer(desc.getDDRunID()), desc);
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.deltadebugging.ddstate.launcher.ddrun.IDDRunListener#memoryGraphsExtracted(int,
     *      org.deltadebugging.ddstate.memorygraph.IRootNode,
     *      org.deltadebugging.ddstate.memorygraph.IRootNode)
     */
    public void memoryGraphsExtracted(int ddRunID, IRootNode passGraph,
            IRootNode failGraph) {

        if (!isVisualizerEnabled())
            return;

        IDDRunDescription desc = (IDDRunDescription) this.fDDRunDesc
                .get(new Integer(ddRunID));

        Date currentTime = new Date(System.currentTimeMillis());
        String parsedDate = AbstractGraphVisualizer.DATE_FORMAT_DATE
                .format(currentTime);
        String parsedTime = AbstractGraphVisualizer.DATE_FORMAT_TIME
                .format(currentTime);

        String fileDir = "undefined" + File.separatorChar + parsedDate;
        String passFileName = ddRunID + "_PassingGraph.dot";
        String failFileName = ddRunID + "_FailingGraph.dot";

        if (desc != null) {
            IJUnitTestDescription passJUnitTestDesc = desc
                    .getPassTestDescription().getJUnitTestDescription();
            IJUnitTestDescription failJUnitTestDesc = desc
                    .getFailTestDescription().getJUnitTestDescription();

            fileDir = this.getShortClassName(failJUnitTestDesc
                    .getTestCaseClassName())
                    + "." + failJUnitTestDesc.getTestName();
            fileDir += File.separatorChar + parsedDate;
            passFileName = ddRunID + "_" + passJUnitTestDesc.getTestName()
                    + "_" + parsedTime + ".dot";

            failFileName = ddRunID + "_" + failJUnitTestDesc.getTestName()
                    + "_" + parsedTime + ".dot";
        }

        //
        // visualize passing graph
        //
        try {
            if (passGraph != null) {
                this.fVisualizer.visualizeAndOpen(passGraph, fileDir
                        + File.separatorChar + passFileName);
            }
        } catch (GraphVisualizerException e) {
            logger.log(IStatus.ERROR, "Failed to visualize passing graph");
        }

        //
        // visualize failing graph
        //
        try {
            if (failGraph != null) {
                this.fVisualizer.visualizeAndOpen(failGraph, fileDir
                        + File.separatorChar + failFileName);
            }
        } catch (GraphVisualizerException e) {
            logger.log(IStatus.ERROR, "Failed to visualize passing graph");
        }
    }

    private String getShortClassName(String className) {
        String result = className;

        if (result != null) {
            int lastIndex = result.lastIndexOf('.') + 1;
            result = result.substring(lastIndex, result.length());
        }

        return result;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.deltadebugging.ddstate.launcher.ddrun.IDDRunListener#ddRunTerminated(org.deltadebugging.ddstate.launcher.ddrun.IDDRunResult)
     */
    public void ddRunTerminated(IDDRunResult result) {
        this.fDDRunDesc.remove(new Integer(result.getDDRunDescription()
                .getDDRunID()));
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.deltadebugging.ddstate.launcher.ddrun.IDDRunListener#memoryGraphsCompared(int,
     *      org.deltadebugging.ddstate.memorygraph.IRootNode,
     *      org.deltadebugging.ddstate.memorygraph.IRootNode,
     *      org.deltadebugging.ddcore.DeltaSet)
     */
    public void memoryGraphsCompared(int ddRunID, IRootNode passGraph,
            IRootNode failGraph, DeltaSet deltas) {

        if (!isVisualizerEnabled())
            return;
        
        Date currentTime = new Date(System.currentTimeMillis());
        String parsedDate = AbstractGraphVisualizer.DATE_FORMAT_DATE
                .format(currentTime);
        String parsedTime = AbstractGraphVisualizer.DATE_FORMAT_TIME
                .format(currentTime);

        IDDRunDescription desc = (IDDRunDescription) this.fDDRunDesc
                .get(new Integer(ddRunID));

        IJUnitTestDescription failJUnitTestDesc = desc.getFailTestDescription()
                .getJUnitTestDescription();

        //
        // visualize passing graph
        //
        String fileDir = this.getShortClassName(failJUnitTestDesc
                .getTestCaseClassName())
                + "." + failJUnitTestDesc.getTestName();
        fileDir += File.separatorChar + parsedDate;
        String fileName = ddRunID +"_COMPARED_" + parsedTime + ".dot";

        try {
            this.fDiffVisualizer.visualizeAndOpen(passGraph, fileDir
                    + File.separatorChar + fileName);
        } catch (GraphVisualizerException e) {
            logger.log(IStatus.ERROR, "Failed to visualize compared graph for "
                    + failJUnitTestDesc.getTestCaseClassName());
        }
    }

    /**
     * Returns <code>true</code> if the graph visualization is enabled in the
     * preference page, otherwise <code>false</code>.
     * 
     * @return <code>true</code> if the graph visualization is enabled in the
     *         preference page, otherwise <code>false</code>.
     */
    private boolean isVisualizerEnabled() {
        boolean result = false;

        Preferences prefs = VisualizerPlugin.getDefault()
                .getPluginPreferences();
        result = prefs
                .getBoolean(IVisualizerPreferenceConstants.GRAPHVIZ_ENABLE_VISUALIZATION);

        return result;
    }

}