package org.deltadebugging.ddstate;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.MissingResourceException;
import java.util.ResourceBundle;

import org.deltadebugging.ddstate.comparator.IGraphComparatorListener;
import org.deltadebugging.ddstate.logger.DDLogger;
import org.eclipse.core.runtime.CoreException;
import org.eclipse.core.runtime.IConfigurationElement;
import org.eclipse.core.runtime.IExtensionPoint;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Platform;
import org.eclipse.jface.preference.IPreferenceStore;
import org.eclipse.ui.plugin.AbstractUIPlugin;
import org.osgi.framework.BundleContext;

/**
 * The main plugin class to be used in the desktop.
 */
public class DDStatePlugin extends AbstractUIPlugin {

	// The shared instance.
	private static DDStatePlugin plugin = null;

	// Resource bundle.
	private ResourceBundle resourceBundle;

	// use this variables for tracing
	public final static boolean DEBUG;

	public final static boolean DEBUG_GRAPH_EXTRACTOR;

	public final static boolean DEBUG_GRAPH_COMPARATOR;

	public final static boolean DEBUG_GRAPH_COMPARATOR_MCS;

	public final static boolean DEBUG_GRAPH_COMPARATOR_DELTACOMPARATOR;

	public final static boolean DEBUG_GRAPH_COMPARATOR_DELTACREATOR;

	public final static boolean DEBUG_GRAPH_DELTA;

	public final static boolean DEBUG_GRAPH_STATISTICS;

	public final static String PLUGIN_ID = "org.deltadebugging.ddstate"; //$NON-NLS-1$

	public static final String ID_EXTENSION_POINT_GRAPHEXTRACTOR_LISTENERS = PLUGIN_ID
			+ ".graphExtractorListeners"; //$NON-NLS-1$

	public static final String ID_EXTENSION_POINT_GRAPHCOMPARATOR_LISTENERS = PLUGIN_ID
			+ ".graphComparatorListeners"; //$NON-NLS-1$

	static {
		DEBUG = "true".equalsIgnoreCase(Platform.getDebugOption(PLUGIN_ID //$NON-NLS-1$
				+ "/debug")); //$NON-NLS-1$
		DEBUG_GRAPH_EXTRACTOR = "true".equalsIgnoreCase(Platform //$NON-NLS-1$
				.getDebugOption(PLUGIN_ID + "/debug/GraphExtractor")); //$NON-NLS-1$
		DEBUG_GRAPH_COMPARATOR = "true".equalsIgnoreCase(Platform
				.getDebugOption(PLUGIN_ID + "/debug/GraphComparator"));
		DEBUG_GRAPH_COMPARATOR_MCS = "true".equalsIgnoreCase(Platform
				.getDebugOption(PLUGIN_ID
						+ "/debug/GraphComparator/MCSComparator"));
		DEBUG_GRAPH_COMPARATOR_DELTACOMPARATOR = "true"
				.equalsIgnoreCase(Platform.getDebugOption(PLUGIN_ID
						+ "/debug/GraphComparator/DeltaComparator"));
		DEBUG_GRAPH_COMPARATOR_DELTACREATOR = "true".equalsIgnoreCase(Platform
				.getDebugOption(PLUGIN_ID
						+ "/debug/GraphComparator/DeltaCreator"));
		DEBUG_GRAPH_DELTA = "true".equalsIgnoreCase(Platform
				.getDebugOption(PLUGIN_ID + "/debug/GraphDelta"));
		DEBUG_GRAPH_STATISTICS = "true".equalsIgnoreCase(Platform
				.getDebugOption(PLUGIN_ID + "/debug/Statistics"));
	}

	/**
	 * List storing the registered graph extractor listeners.
	 */
	private List graphExtractorListeners = null;

	/**
	 * List storing the registered graph comperator listeners.
	 */
	private List graphComparatorListeners = null;

	/**
	 * The constructor.
	 */
	public DDStatePlugin() {
		super();
		plugin = this;

		try {
			this.resourceBundle = ResourceBundle
					.getBundle("org.deltadebugging.ddstate.DDStatePluginResources");
		} catch (MissingResourceException x) {
			this.resourceBundle = null;
		}
	}

	/**
	 * Return a logger for the given tracer class.
	 * 
	 * @param tracer
	 *            the tracer is the class that have to log or trace messages.
	 * @return a logger for the given tracer class.
	 */
	public static DDLogger getLogger(Class tracer) {
		return new DDLogger(DDStatePlugin.getDefault(), tracer);
	}

	/**
	 * This method is called upon plug-in activation
	 */
	public void start(BundleContext context) throws Exception {
		super.start(context);
	}

	/**
	 * This method is called when the plug-in is stopped
	 */
	public void stop(BundleContext context) throws Exception {
		super.stop(context);
	}

	/**
	 * Returns the shared instance.
	 */
	public static DDStatePlugin getDefault() {
		return plugin;
	}

	/**
	 * Returns the string from the plugin's resource bundle, or 'key' if not
	 * found.
	 */
	public static String getResourceString(String key) {
		ResourceBundle bundle = DDStatePlugin.getDefault().getResourceBundle();
		try {
			return (bundle != null) ? bundle.getString(key) : key;
		} catch (MissingResourceException e) {
			return key;
		}
	}

	/**
	 * Returns the plugin's resource bundle,
	 */
	public ResourceBundle getResourceBundle() {
		return this.resourceBundle;
	}

	/**
	 * Adds a GraphExtractor listener to the collection of listeners
	 */
	public void addGraphExtractorListener(IGraphExtractorListener newListener) {
		if (this.graphExtractorListeners == null)
			loadGraphExtractorListeners();
		for (Iterator iter = this.graphExtractorListeners.iterator(); iter
				.hasNext();) {
			Object o = iter.next();
			if (o == newListener)
				return;
		}
		this.graphExtractorListeners.add(newListener);
	}

	/**
	 * Adds a GraphComparator listener to the collection of listeners
	 */
	public void addGraphComparatorListener(IGraphComparatorListener newListener) {
		if (this.graphComparatorListeners == null)
			loadGraphExtractorListeners();
		for (Iterator iter = this.graphComparatorListeners.iterator(); iter
				.hasNext();) {
			Object o = iter.next();
			if (o == newListener)
				return;
		}
		this.graphComparatorListeners.add(newListener);
	}

	/**
	 * Returns an array of all GraphExtractor listeners
	 */
	public List getGraphExtractorListeners() {
		if (this.graphExtractorListeners == null) {
			loadGraphExtractorListeners();
		}
		return this.graphExtractorListeners;
	}

	/**
	 * Returns an array of all GraphComparator listeners
	 */
	public List getGraphComparatorListeners() {
		if (this.graphComparatorListeners == null) {
			loadGraphComparatorListeners();
		}
		return this.graphComparatorListeners;
	}

	/**
	 * Initializes GraphExtractor Listener extensions
	 */
	private void loadGraphExtractorListeners() {
		this.graphExtractorListeners = new ArrayList();
		IExtensionPoint extensionPoint = Platform.getExtensionRegistry()
				.getExtensionPoint(ID_EXTENSION_POINT_GRAPHEXTRACTOR_LISTENERS);
		if (extensionPoint == null) {
			return;
		}
		IConfigurationElement[] configs = extensionPoint
				.getConfigurationElements();

		for (int i = 0; i < configs.length; i++) {
			try {
				IGraphExtractorListener graphExtractorListener = (IGraphExtractorListener) configs[i]
						.createExecutableExtension("class"); //$NON-NLS-1$
				this.graphExtractorListeners.add(graphExtractorListener);
			} catch (CoreException e) {
				DDStatePlugin.getLogger(DDStatePlugin.class).log(IStatus.ERROR,
						"Could not load some graphExtractor extension points",
						e);
			}
		}
	}

	/**
	 * Initializes GraphComparator Listener extensions
	 */
	private void loadGraphComparatorListeners() {
		this.graphComparatorListeners = new ArrayList();
		IExtensionPoint extensionPoint = Platform
				.getExtensionRegistry()
				.getExtensionPoint(ID_EXTENSION_POINT_GRAPHCOMPARATOR_LISTENERS);
		if (extensionPoint == null) {
			return;
		}
		IConfigurationElement[] configs = extensionPoint
				.getConfigurationElements();

		for (int i = 0; i < configs.length; i++) {
			try {
				IGraphComparatorListener graphComparatorListener = (IGraphComparatorListener) configs[i]
						.createExecutableExtension("class"); //$NON-NLS-1$
				this.graphComparatorListeners.add(graphComparatorListener);
			} catch (CoreException e) {
				DDStatePlugin.getLogger(DDStatePlugin.class).log(IStatus.ERROR,
						"Could not load some graphComparator extension points",
						e);
			}
		}
	}

	/**
	 * Removes a GraphExtractor listener to the collection of listeners
	 */
	public void removeGraphExtractorListener(IGraphExtractorListener newListener) {
		if (this.graphExtractorListeners != null)
			this.graphExtractorListeners.remove(newListener);
	}

	/**
	 * Removes a GraphComparator listener to the collection of listeners
	 */
	public void removeGraphComparatorListener(
			IGraphComparatorListener newListener) {
		if (this.graphComparatorListeners != null)
			this.graphComparatorListeners.remove(newListener);
	}

	protected void initializeDefaultPreferences(IPreferenceStore store) {
		// GraphExtractorPreferencePage
		store.setDefault(
				IDDStatePreferenceConstants.GRAPHEXTRACTOR_ACTIVE_FILTERS_LIST,
				"org.deltadebugging.ddstate.*," + "org.apache.bcel.*,"
						+ "java.*,javax.*,sun.*,sunw.*,com.sun.*,junit.*");
		store
				.setDefault(
						IDDStatePreferenceConstants.GRAPHEXTRACTOR_INACTIVE_FILTERS_LIST,
						""); //$NON-NLS-1$
	}
}