/*
 * $Id: GraphComparator.java,v 1.7 2005/11/20 14:25:02 lehmann Exp $
 * 
 * Copyright (C) 2004 Karsten Lehmann
 * 
 * Software Engineering Chair, Saarland University
 * Prof. Dr. Andreas Zeller
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 *
 * Created on 01.09.2004 13:52:46 by
 * @author Karsten Lehmann 
 * Mail to: lehmann@cs.uni-sb.de or KarstenLehmann@web.de
 */
package org.deltadebugging.ddstate.comparator;

import java.util.Iterator;
import java.util.List;

import org.deltadebugging.ddcore.DeltaSet;
import org.deltadebugging.ddstate.DDStatePlugin;
import org.deltadebugging.ddstate.comparator.delta.GraphDecorator;
import org.deltadebugging.ddstate.comparator.delta.GraphDifferentMarker;
import org.deltadebugging.ddstate.comparator.exceptions.GraphComparatorException;
import org.deltadebugging.ddstate.comparator.mcs.MCSComparator;
import org.deltadebugging.ddstate.logger.DDLogger;
import org.deltadebugging.ddstate.memorygraph.IRootNode;
import org.deltadebugging.ddstate.memorygraph.exceptions.DeltaComputationException;
import org.deltadebugging.ddstate.memorygraph.util.MemoryGraphUtil;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.ISafeRunnable;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.NullProgressMonitor;
import org.eclipse.core.runtime.Platform;

/**
 * COMMENTME Comment type GraphComparator
 * 
 * @author <a href="mailto:KarstenLehmann@web.de">Karsten Lehmann </a>
 */
public class GraphComparator implements IGraphComparator {

	/**
	 * Personal logger for this class
	 */
	final static DDLogger logger = DDStatePlugin
			.getLogger(GraphComparator.class);

	/**
	 * The instance of the graph comparator. The graph comparator is a
	 * singleton. Therefor you have to call getInstance to get an instance of
	 * the graph comparator.
	 */
	private static GraphComparator instance = null;

	private GraphComparator() {
		// use getInstance to get an instance of the graph comparator
	}

	/**
	 * Return the single instance of the graph comparator.
	 * 
	 * @return the singel instance of the graph comparator.
	 */
	public static GraphComparator getInstance() {
		if (instance == null) {
			instance = new GraphComparator();
		}

		return instance;
	}

	public abstract class ListenerSafeRunnable implements ISafeRunnable {
		public void handleException(Throwable t) {
			logger.log(IStatus.ERROR, "Exception on notify listener", t);
		}
	}

	/**
	 * @see org.deltadebugging.ddstate.comparator.IGraphComparator#compare(int,
	 *      org.deltadebugging.ddstate.memorygraph.IRootNode,
	 *      org.deltadebugging.ddstate.memorygraph.IRootNode,
	 *      org.eclipse.core.runtime.IProgressMonitor)
	 */
	public DeltaSet compare(int ddRunID, IRootNode firstGraph,
			IRootNode secondGraph, IProgressMonitor monitor)
			throws GraphComparatorException {

		List listeners = DDStatePlugin.getDefault()
				.getGraphComparatorListeners();

		DeltaSet result = null;

		if (firstGraph == null || secondGraph == null)
			throw new IllegalArgumentException("Given graphs must not be null.");

		if (monitor == null) {
			monitor = new NullProgressMonitor();
		}

		try {
			monitor.beginTask("Compare memory graphs...", 2);

			long startTime = logger.traceAndTimeStart(
					DDStatePlugin.DEBUG_GRAPH_COMPARATOR, "" + ddRunID,
					"Start Comparing both graphs ...");

			// notify all listener that the comparison run is started
			logger.trace(DDStatePlugin.DEBUG_GRAPH_COMPARATOR,
					"Starting comparison run ...");
			this.notifyComparisonRunStarted(ddRunID, listeners, firstGraph,
					secondGraph);

			// find the maximum common subgraph (MCS) of the two graphs
			if (monitor.isCanceled()) {
				return result;
			}

			monitor.subTask("Compare memory graphs (find common subgraph)...");
			MCSComparator.getInstance().compare(firstGraph, secondGraph);
			monitor.worked(1);

			// Register the GraphDifferentMarker on the DeltaComparator
			if (monitor.isCanceled()) {
				return result;
			}

			logger.trace(DDStatePlugin.DEBUG_GRAPH_STATISTICS,
					"Statistics first graph");
			logger.trace(DDStatePlugin.DEBUG_GRAPH_STATISTICS, firstGraph
					.getStatistics().toString());

			logger.trace(DDStatePlugin.DEBUG_GRAPH_STATISTICS,
					"Statistics second graph");
			logger.trace(DDStatePlugin.DEBUG_GRAPH_STATISTICS, secondGraph
					.getStatistics().toString());

			GraphDecorator
					.addGraphDecoratorListener(new GraphDifferentMarker());

			try {
				monitor.subTask("Compare memory graphs (compute deltas)...");
				result = firstGraph.computeDeltaSet(secondGraph);
				monitor.worked(1);
			} catch (DeltaComputationException e) {
				throw new GraphComparatorException(
						"Exception on computing deltas of the two graphs", e);
			}

			long endTime = logger.traceAndTimeEnd(
					DDStatePlugin.DEBUG_GRAPH_COMPARATOR, "" + ddRunID,
					"Comparison run terminated ...");

			// notify all listener that the comparison run is terminated
			this.notifyComparisonRunTerminated(ddRunID, listeners, endTime
					- startTime);
		} finally {
			monitor.done();
		}

		assert MemoryGraphUtil.assertAllElementsAreCompared(firstGraph);
		assert MemoryGraphUtil.assertAllElementsAreCompared(secondGraph);

		return result;
	}

	// COMMENTME notifyComparisonRunStarted
	private void notifyComparisonRunStarted(final int ddRunID, List listeners,
			final IRootNode firstGraph, final IRootNode secondGraph) {
		for (Iterator itListeners = listeners.iterator(); itListeners.hasNext();) {
			final IGraphComparatorListener listener = (IGraphComparatorListener) itListeners
					.next();
			Platform.run(new ListenerSafeRunnable() {
				public void run() {
					listener.comparisonRunStarted(ddRunID, firstGraph,
							secondGraph);
				}
			});
		}
	}

	// COMMENTME notifyComparisonRunTerminated
	private void notifyComparisonRunTerminated(final int ddRunID,
			List listeners, final long elapsedTime) {
		for (Iterator itListeners = listeners.iterator(); itListeners.hasNext();) {
			final IGraphComparatorListener listener = (IGraphComparatorListener) itListeners
					.next();
			Platform.run(new ListenerSafeRunnable() {
				public void run() {
					listener.comparisonRunTerminated(ddRunID, elapsedTime);
				}
			});
		}
	}

}