/*
 * $Id: GraphEdgeComparator.java,v 1.5 2005/09/19 16:31:09 lehmann Exp $
 * 
 * Copyright (C) 2004 Karsten Lehmann
 * 
 * Software Engineering Chair, Saarland University
 * Prof. Dr. Andreas Zeller
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 *
 * Created on 13.12.2004 20:15:09 by
 * @author Karsten Lehmann 
 * Mail to: lehmann@cs.uni-sb.de or KarstenLehmann@web.de
 */
package org.deltadebugging.ddstate.comparator;

import java.util.Comparator;

import org.deltadebugging.ddstate.memorygraph.ILocalVariableEdge;
import org.deltadebugging.ddstate.memorygraph.IGraphEdge;
import org.deltadebugging.ddstate.memorygraph.IStackFrame;
import org.deltadebugging.ddstate.memorygraph.value.IArrayElementEdge;

/**
 * The GraphEdgeComparator compare two edges. Two edge are identical if the
 * names of the two graph edges are identical. The ordering thus controlled by
 * the name of the edges. For edges of an array reference we use the numerical
 * ordering on the array indices.
 * 
 * @author <a href="mailto:KarstenLehmann@web.de">Karsten Lehmann </a>
 */
public class GraphEdgeComparator implements Comparator {

    private static GraphEdgeComparator instance = null;

    /**
     * Designed as singleton.
     */
    private GraphEdgeComparator() {
        // empty constructor
    }

    public final static GraphEdgeComparator getInstance() {
        if (instance == null) {
            instance = new GraphEdgeComparator();
        }

        return instance;
    }

    /*
     * (non-Javadoc)
     * 
     * @see java.util.Comparator#compare(java.lang.Object, java.lang.Object)
     */
    public int compare(Object firstObj, Object secondObj) {

        if (!(firstObj instanceof IGraphEdge && secondObj instanceof IGraphEdge))
            throw new ClassCastException("Given objects must be from type "
                    + IGraphEdge.class.getName());

        IGraphEdge firstEdge = (IGraphEdge) firstObj;
        IGraphEdge secondEdge = (IGraphEdge) secondObj;

        int result = 0;

        // If the given edges are edge from an array use numerical sorting to
        // order the array elements
        if (firstEdge instanceof IArrayElementEdge
                && secondEdge instanceof IArrayElementEdge) {
            result = this.compareArrayElementEdges(
                    (IArrayElementEdge) firstEdge,
                    (IArrayElementEdge) secondEdge);

        } else if (firstEdge instanceof ILocalVariableEdge
                && secondEdge instanceof ILocalVariableEdge) {
            result = this.compareLocalVariableEdges(
                    (ILocalVariableEdge) firstEdge,
                    (ILocalVariableEdge) secondEdge);

        } else {

            // Two edge are equal if the operation of the edges are equal.
            // This requires, that the edges come from the same source node.
            result = firstEdge.getName().compareTo(secondEdge.getName());
        }

        return result;
    }

    /**
     * If we compare the edges to the local variables we use the informations
     * given by the stackframe and the variable name. For the former we compare
     * the inversive stackframe number. This is necessary because if we compare
     * the edges over many delta debugging runs (e.g. used in cause transition
     * determination) we have to ensure that the lowest frame in our stack is
     * the base. The base frame has the number 0. All frames above it get the
     * appropriate higher number. <br>
     * 
     * The result is the compare of two strings. We build the string by
     * concatenate the inverse stack frame number with the variable name.
     * 
     * @param firstEdge
     * @param secondEdge
     * @see java.lang.String#comapre(java.lang.Object, java.lang.Object)
     */
    public int compareLocalVariableEdges(ILocalVariableEdge firstEdge,
            ILocalVariableEdge secondEdge) {

        IStackFrame firstStackFrame = firstEdge.getStackFrame();
        IStackFrame secondStackFrame = secondEdge.getStackFrame();

        int firstInvFrameNr = firstStackFrame.getFramesCount() - 1
                - firstStackFrame.getFrameNumber();
        int secondInvFrameNr = secondStackFrame.getFramesCount() - 1
                - secondStackFrame.getFrameNumber();

        String firstKey = firstInvFrameNr + firstEdge.getVariableName();
        String secondKey = secondInvFrameNr + secondEdge.getVariableName();

        int result = 0;

        if (this.isThisVariable(firstEdge) && !this.isThisVariable(secondEdge)) {
            result = -1;
            
        } else if (!this.isThisVariable(firstEdge) && this.isThisVariable(secondEdge)) {
            result = 1;
            
        } else {
            result = firstKey.compareTo(secondKey);
        }

        return result;
    }

    private boolean isThisVariable(ILocalVariableEdge edge) {
        return edge.getName().endsWith("this");
    }

    /**
     * Compare the two array element edges on their indeces.
     * 
     * @param firstEdge
     * @param secondEdge
     * @see java.lang.Integer#comapre(java.lang.Object, java.lang.Object)
     */
    public int compareArrayElementEdges(IArrayElementEdge firstEdge,
            IArrayElementEdge secondEdge) {

        int result = 0;
        Integer firstIndex = new Integer(firstEdge.getArrayIndex());
        Integer secondIndex = new Integer(secondEdge.getArrayIndex());

        result = firstIndex.compareTo(secondIndex);

        return result;
    }
}
