/*
 * $Id: GraphEdgeDiffUtil.java,v 1.3 2005/11/28 11:12:49 lehmann Exp $
 * 
 * Copyright (C) 2005 Karsten Lehmann
 * 
 * Software Engineering Chair, Saarland University
 * Prof. Dr. Andreas Zeller
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 *
 * Created on 21.04.2005 12:18:32 by
 * @author Karsten Lehmann 
 * Mail to: lehmann@cs.uni-sb.de or KarstenLehmann@web.de
 */
package org.deltadebugging.ddstate.comparator.delta;

import java.util.Comparator;
import java.util.Iterator;

import org.deltadebugging.ddstate.comparator.GraphEdgeComparator;
import org.deltadebugging.ddstate.memorygraph.IGraphEdge;
import org.deltadebugging.ddstate.memorygraph.IGraphNode;

/**
 * The graph edge diff util computes a edge diff between two graph nodes. The
 * edge diff includes all edge only in the first, the second or both memory
 * graphs. To check the equality of to edge the util use the graph edge
 * comparator.
 * 
 * @see org.deltadebugging.ddstate.comparator.GraphEdgeComparator
 * 
 * @author <a href="mailto:KarstenLehmann@web.de">Karsten Lehmann </a>
 */
public class GraphEdgeDiffUtil {

	/**
	 * Comparator used to compare two graph edges for equality.
	 */
	private final static Comparator COMPARATOR = GraphEdgeComparator
			.getInstance();

	/**
	 * Compares the successors of the given two graph nodes and use the graph
	 * edge comparator to check the equality of the successors.
	 * 
	 * @param firstNode
	 * @param secondNode
	 * @return a container that includes all successors grouped by the
	 *         membership of the graph nodes.
	 */
	public static GraphEdgeDiff computeDiff(IGraphNode firstNode,
			IGraphNode secondNode) {
		GraphEdgeDiff result = new GraphEdgeDiff();

		Iterator itFirstSuccessors = firstNode.getChildren().iterator();
		Iterator itSecondSuccessors = secondNode.getChildren().iterator();

		IGraphEdge firstSuccEdge = null;
		IGraphEdge secondSuccEdge = null;

		// We have to look for every edge in the two iterators
		while (itFirstSuccessors.hasNext() || itSecondSuccessors.hasNext()) {

			// 1) If the first node have no more successors, we can add the
			// edges of the second node in the collection of edges that are only
			// in the second node available
			if (!itFirstSuccessors.hasNext()) {
				while (itSecondSuccessors.hasNext()) {
					result
							.addEdgeOnlyInSecondGraph((IGraphEdge) itSecondSuccessors
									.next());
				}
				return result;
			}

			// 2) ..., else if the second node has no successors, we can add the
			// edges of the first node in the collection of edges thar are only
			// in the first node available.
			if (!itSecondSuccessors.hasNext()) {
				while (itFirstSuccessors.hasNext()) {
					result
							.addEdgeOnlyInFirstGraph((IGraphEdge) itFirstSuccessors
									.next());
				}

				return result;
			}

			// 3) ..., else the first node and the second node has successors.
			firstSuccEdge = (IGraphEdge) itFirstSuccessors.next();
			secondSuccEdge = (IGraphEdge) itSecondSuccessors.next();

			// Because the successor are in an ascending order, we request
			// the respective iterator with the "lower" successor for the
			// next successor until we have two equal successors or no more
			// successors exists in one of the two iterators.
			int compResult = COMPARATOR.compare(firstSuccEdge, secondSuccEdge);

			for (; compResult != 0; compResult = COMPARATOR.compare(
					firstSuccEdge, secondSuccEdge)) {

				if (compResult < 0) {
					result.addEdgeOnlyInFirstGraph(firstSuccEdge);

					if (itFirstSuccessors.hasNext())
						firstSuccEdge = (IGraphEdge) itFirstSuccessors.next();
					else
						break;

				} else {
					result.addEdgeOnlyInSecondGraph(secondSuccEdge);

					if (itSecondSuccessors.hasNext())
						secondSuccEdge = (IGraphEdge) itSecondSuccessors.next();
					else
						break;
				}
			}

			if (compResult == 0) {
				result.addEdgesInBothGraph(firstSuccEdge, secondSuccEdge);
			}
		}

		return result;
	}

}