/*
 * $Id: MCSComparator.java,v 1.15 2005/11/20 14:24:25 lehmann Exp $
 * 
 * Copyright (C) 2004 Karsten Lehmann
 * 
 * Software Engineering Chair, Saarland University
 * Prof. Dr. Andreas Zeller
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 *
 * Created on 14.12.2004 11:39:47 by
 * @author Karsten Lehmann 
 * Mail to: lehmann@cs.uni-sb.de or KarstenLehmann@web.de
 */
package org.deltadebugging.ddstate.comparator.mcs;

import java.util.Collection;
import java.util.Iterator;

import org.deltadebugging.ddstate.DDStatePlugin;
import org.deltadebugging.ddstate.comparator.delta.GraphEdgeDiff;
import org.deltadebugging.ddstate.comparator.delta.GraphEdgeDiffUtil;
import org.deltadebugging.ddstate.comparator.delta.GraphEdgePair;
import org.deltadebugging.ddstate.logger.DDLogger;
import org.deltadebugging.ddstate.memorygraph.IGraphEdge;
import org.deltadebugging.ddstate.memorygraph.IGraphElement;
import org.deltadebugging.ddstate.memorygraph.IGraphNode;
import org.deltadebugging.ddstate.memorygraph.IRootNode;
import org.deltadebugging.ddstate.memorygraph.RootNode;
import org.deltadebugging.ddstate.memorygraph.value.IPrimitiveValueNode;

/**
 * The MCSComparator provide the service to find the maximum common subgraph of
 * two memory graph. The class use the functionality of the IMCSComparable
 * interface to set the associaten between the two graphs. Two graphs that want
 * to compare each other with the MCSComparator have to implement the
 * IMCSComparable interface for their graph elements.
 * 
 * @author <a href="mailto:KarstenLehmann@web.de">Karsten Lehmann </a>
 */
public class MCSComparator {

	/** Personal logger for this class */
	private final static DDLogger logger = DDStatePlugin
			.getLogger(MCSComparator.class);

	private static MCSComparator instance = null;

	/**
	 * Designed the comparator as a singleton
	 * 
	 */
	private MCSComparator() {
		// empty constructor
	}

	public final static MCSComparator getInstance() {
		if (instance == null) {
			instance = new MCSComparator();
		}

		return instance;
	}

	public void compare(IRootNode firstGraph, IRootNode secondGraph) {

		assert firstGraph instanceof RootNode;
		assert secondGraph instanceof RootNode;

		String key = "compareMCS" + firstGraph.getDDRunID();
		logger.traceAndTimeStart(DDStatePlugin.DEBUG_GRAPH_COMPARATOR, key,
				"Compute MCS");

		// set the association between the two root nodes
		this.setAssociation(firstGraph, secondGraph);

		this.findMCS(firstGraph, secondGraph);

		logger.traceAndTimeEnd(DDStatePlugin.DEBUG_GRAPH_COMPARATOR, key,
				"MCS computed");
	}

	/**
	 * Set the association between the two graph elements. The precondition is
	 * that the each of the given two graph elements have no association so far.
	 * 
	 * @param firstElement
	 *            first graph element associated to the second graph element.
	 * @param secondElement
	 *            second graph element associated to the first graph element.
	 */
	private void setAssociation(IGraphElement firstElement,
			IGraphElement secondElement) {
		assert !firstElement.hasAssociatedElement();
		assert !secondElement.hasAssociatedElement();

		firstElement.setAssociatedElement(secondElement);
		secondElement.setAssociatedElement(firstElement);

		if (DDStatePlugin.DEBUG_GRAPH_STATISTICS) {
			if (firstElement instanceof IGraphNode) {
				firstElement.getRootNode().getStatistics().numberOfEqualNodes++;
				secondElement.getRootNode().getStatistics().numberOfEqualNodes++;
			} else {
				firstElement.getRootNode().getStatistics().numberOfEqualEdges++;
				secondElement.getRootNode().getStatistics().numberOfEqualEdges++;
			}

		}
	}

	/**
	 * This function find the maximum common subgraph of the given graphs.
	 * 
	 * @param firstGraph
	 *            the first graph to compare.
	 * @param secondGraph
	 *            the second graph to compare.
	 */
	private void findMCS(IGraphNode firstGraph, IGraphNode secondGraph) {
		// The given graphs node have to be associated to each other. Because we
		// start at the root node and the root nodes are per definition equal
		// and therefor associated to each other.
		assert firstGraph.isAssociatedWith(secondGraph);

		logger.trace(DDStatePlugin.DEBUG_GRAPH_COMPARATOR_MCS,
				"Compare graphs \n\t1. " + firstGraph.toString() + "\n\t2. "
						+ secondGraph.toString());

		GraphEdgeDiff diff = GraphEdgeDiffUtil.computeDiff(firstGraph,
				secondGraph);

		IGraphEdge firstSuccEdge = null;
		IGraphEdge secondSuccEdge = null;

		Collection graphPairs = diff.getEdgesInBothGraph();

		for (Iterator itGraphPairs = graphPairs.iterator(); itGraphPairs
				.hasNext();) {
			GraphEdgePair graphPair = (GraphEdgePair) itGraphPairs.next();
			firstSuccEdge = graphPair.getFirstEdge();
			secondSuccEdge = graphPair.getSecondEdge();

			logger.trace(DDStatePlugin.DEBUG_GRAPH_COMPARATOR_MCS,
					"Found equal edges \n\t1. " + firstSuccEdge.getName()
							+ "\n\t2. " + secondSuccEdge.getName());

			if (firstSuccEdge.graphEquals(secondSuccEdge)) {
				this.setAssociation(firstSuccEdge, secondSuccEdge);

				IGraphNode firstTargetNode = firstSuccEdge.getTargetNode();
				IGraphNode secondTargetNode = secondSuccEdge.getTargetNode();

				if (firstTargetNode instanceof IPrimitiveValueNode
						&& secondTargetNode instanceof IPrimitiveValueNode) {
					// avoid assertion. if the two primtives are associated with
					// each other
					if (!firstTargetNode.isAssociatedWith(secondTargetNode)) {
						this.setAssociation(firstTargetNode, secondTargetNode);
					}

				} else 
					if (firstTargetNode.graphEquals(secondTargetNode)) {

					logger.trace(DDStatePlugin.DEBUG_GRAPH_COMPARATOR_MCS,
							"Found equal nodes \n\t1. "
									+ firstTargetNode.toString() + "\n\t2. "
									+ secondTargetNode.toString());

					// avoid cycles
					if (!firstTargetNode.hasAssociatedElement()
							&& !secondTargetNode.hasAssociatedElement()) {
						this.setAssociation(firstTargetNode, secondTargetNode);

						this.findMCS(firstTargetNode, secondTargetNode);
					}

					// else {
					// System.out.println("References changed");
					// }
				}
			}
		}
	}
}
