/*
 * $Id: GraphNode.java,v 1.57 2006/01/24 19:45:32 lehmann Exp $
 * 
 * Copyright (C) 2004 Karsten Lehmann
 * 
 * Software Engineering Chair, Saarland University
 * Prof. Dr. Andreas Zeller
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 *
 * Created on 26.08.2004 by
 * @author Karsten Lehmann 
 * Mail to: lehmann@cs.uni-sb.de or KarstenLehmann@web.de
 */
package org.deltadebugging.ddstate.memorygraph;

import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;
import java.util.SortedSet;
import java.util.TreeSet;
import java.util.Vector;

import org.deltadebugging.ddstate.DDStatePlugin;
import org.deltadebugging.ddstate.comparator.GraphEdgeComparator;
import org.deltadebugging.ddstate.memorygraph.exceptions.MemoryGraphExtractionException;
import org.deltadebugging.ddstate.memorygraph.type.IGlobalVariableEdge;
import org.deltadebugging.ddstate.memorygraph.value.IArrayElementEdge;
import org.deltadebugging.ddstate.memorygraph.value.IObjectFieldEdge;
import org.deltadebugging.ddstate.memorygraph.value.IObjectNode;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.debug.core.DebugException;
import org.eclipse.jdt.debug.core.IJavaObject;
import org.eclipse.jdt.debug.core.IJavaValue;
import org.eclipse.jdt.debug.core.IJavaVariable;

/**
 * COMMENTME Comment type ***
 * 
 * @author <a href="mailto:KarstenLehmann@web.de">Karsten Lehmann </a>
 */
public abstract class GraphNode extends GraphElement implements IGraphNode {

    /**
     * A List of all direct successors of this graph node. It is necessary for
     * the graph comparison that the successors are in a specified order.
     * Therefor we use internally the TreeSet. The TreeSet use the
     * <code>compareTo</code> function impelemented by the GraphEdge to bring
     * the elements in required the order.
     * 
     * @see org.deltadebugging.ddstate.memorygraph.GraphEdge#compareTo(Object)
     */
    private SortedSet fChildren = new TreeSet(GraphEdgeComparator.getInstance());

    private Set fParents = new HashSet();

    /**
     * If <code>isUnfolded</code> is <code>true</code> this graph node was
     * always unfolded.
     */
    private boolean isUnfolded = false;

    public GraphNode(IRootNode rootNode) {
        super(rootNode);

        if (DDStatePlugin.DEBUG_GRAPH_STATISTICS && rootNode != null) {
            rootNode.getStatistics().numberOfGraphNodes++;
        }
    }

    /**
     * Looks for the mirror in this graph node and create all direct successors
     * for the referenced objects in the mirror. For example the ArrayNode hold
     * a mirror of the type ArrayNode. When you call this method all fields
     * referenced by the array will be created as successor of the ArrayNode.
     * 
     * @throws MemoryGraphExtractionException
     *             if something went wrong on the exctraction of the memory
     *             graph.
     */
    protected abstract void createChildren()
            throws MemoryGraphExtractionException;

    /*
     * (non-Javadoc)
     * 
     * @see org.deltadebugging.ddstate.memorygraph.IGraphElement#unfoldSubgraph(IProgressMonitor
     *      monitor)
     */
    public void unfoldSubgraph(IProgressMonitor monitor)
            throws MemoryGraphExtractionException {

        assert !this.isUnfolded;

        if (monitor.isCanceled()) {
            return;
        }

        if (this.shouldUnfoldSubgraph() && !this.isUnfolded) {
            Iterator itChildren = this.fChildren.iterator();

            this.isUnfolded = true;

            while (itChildren.hasNext()) {
                GraphEdge graphEdge = (GraphEdge) itChildren.next();

                GraphNode targetNode = (GraphNode) graphEdge.getTargetNode();

                // Prevent that the target node was more than one time unfolded.
                // If we don't check if the targetNode is already unfolded and
                // this node has more than one reference to the target node, the
                // target node would unfolded twice and more.
                if (!targetNode.isUnfolded) {
                    targetNode.createChildren();
                    targetNode.unfoldSubgraph(monitor);
                }
            }
        }
    }

    /**
     * Override the method, if you want that a graph node does not unfold his
     * successors.
     * 
     * @return the default resault is true.
     */
    public boolean shouldUnfoldSubgraph() {
        return true;
    }

    protected IGraphNode createAndRegisterNode(IJavaVariable variable)
            throws MemoryGraphExtractionException {
        IGraphNode result = null;

        try {
            IJavaValue value = (IJavaValue) variable.getValue().getAdapter(
                    IJavaValue.class);

            // Look if the value of the given variable is an object and already
            // registered. Then we return the registered node for this value. So
            // we ensure that never two nodes exists for the same object in our
            // memory graph.
            if (value instanceof IJavaObject
                    && this.getRootNode().existObjectNode((IJavaObject) value)) {
                result = this.getRootNode().getObjectNode((IJavaObject) value);

            } else {
                // If not an node exists for the value we have to create one.
                result = GraphElementUtil
                        .createValueNode(this.fRootNode, value);

                // Represents the value an object in the memory we register the
                // node which represents this value.
                if (result instanceof IObjectNode) {
                    this.getRootNode().registerObjectNode((IObjectNode) result);
                }
            }
        } catch (DebugException e) {
            throw new MemoryGraphExtractionException(
                    "Failure on check if the given java variable already has a node.",
                    e);
        }

        return result;
    }

    /**
     * Add a new graph edge (successor) to the list of the existing successors
     * (graph edges). The means that this graph node points to the target node
     * defined in the given graph edge.
     * 
     * @param edge
     *            the graph edge to be added.
     * @throws IllegalArgumentException
     *             if the graph edge is null.
     */
    public void addSuccessor(IGraphEdge edge) {
        if (edge == null)
            throw new IllegalArgumentException("Graphedge must not be null.");

        this.fChildren.add(edge);
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.deltadebugging.ddstate.memorygraph.IGraphNode#getSuccessors()
     */
    public SortedSet getChildren() {
        return this.fChildren;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.deltadebugging.ddstate.memorygraph.IGraphNode#addPredeccessor(org.deltadebugging.ddstate.memorygraph.IGraphEdge)
     */
    public void addParent(IGraphEdge graphEdge) {
        if (graphEdge == null)
            throw new NullPointerException("GraphEdge must not be null.");

        if (!this.fParents.contains(graphEdge)) {
            this.fParents.add(graphEdge);
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.deltadebugging.ddstate.memorygraph.IGraphNode#getParents()
     */
    public Collection getParents() {
        return this.fParents;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.deltadebugging.ddstate.memorygraph.IGraphNode#isLocalVariableNode()
     */
    public boolean isLocalVariable() {
        boolean result = false;

        Iterator itPredecessors = this.getParents().iterator();
        while (itPredecessors.hasNext() && result == false) {
            IGraphEdge edge = (IGraphEdge) itPredecessors.next();

            if (edge instanceof ILocalVariableEdge) {
                result = true;
            }
        }

        return result;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.deltadebugging.ddstate.memorygraph.IGraphNode#getLocalVariablePrecessors()
     */
    public ILocalVariableEdge[] getLocalVariablePrecessors() {
        ILocalVariableEdge[] result = null;

        Iterator itPredecessors = this.getParents().iterator();
        Collection colResult = new Vector();

        while (itPredecessors.hasNext()) {
            IGraphEdge edge = (IGraphEdge) itPredecessors.next();

            if (edge instanceof ILocalVariableEdge) {
                colResult.add(edge);
            }
        }

        result = (ILocalVariableEdge[]) colResult
                .toArray(new ILocalVariableEdge[colResult.size()]);

        return result;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.deltadebugging.ddstate.memorygraph.IGraphNode#getGlobalVariablePredecessors()
     */
    public IGlobalVariableEdge[] getGlobalVariablePredecessors() {
        IGlobalVariableEdge[] result = null;

        Iterator itPredecessors = this.getParents().iterator();
        Collection colResult = new Vector();

        while (itPredecessors.hasNext()) {
            IGraphEdge edge = (IGraphEdge) itPredecessors.next();

            if (edge instanceof IGlobalVariableEdge) {
                colResult.add(edge);
            }
        }

        result = (IGlobalVariableEdge[]) colResult
                .toArray(new IGlobalVariableEdge[colResult.size()]);

        return result;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.deltadebugging.ddstate.memorygraph.IGraphNode#isGlobalVariableNode()
     */
    public boolean isGlobalVariable() {
        boolean result = false;

        Iterator itPredecessors = this.getParents().iterator();
        while (itPredecessors.hasNext() && result == false) {
            IGraphEdge edge = (IGraphEdge) itPredecessors.next();

            if (edge instanceof IGlobalVariableEdge) {
                result = true;
            }
        }

        return result;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.deltadebugging.ddstate.memorygraph.IGraphNode#isArrayElementVariable()
     */
    public boolean isArrayElementVariable() {
        boolean result = false;

        Iterator itPredecessors = this.getParents().iterator();
        while (itPredecessors.hasNext() && result == false) {
            IGraphEdge edge = (IGraphEdge) itPredecessors.next();

            if (edge instanceof IArrayElementEdge) {
                result = true;
            }
        }

        return result;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.deltadebugging.ddstate.memorygraph.IGraphNode#getArrayElementPredecessors()
     */
    public IArrayElementEdge[] getArrayElementPredecessors() {
        IArrayElementEdge[] result = null;

        Iterator itPredecessors = this.getParents().iterator();
        Collection colResult = new Vector();

        while (itPredecessors.hasNext()) {
            IGraphEdge edge = (IGraphEdge) itPredecessors.next();

            if (edge instanceof IArrayElementEdge) {
                colResult.add(edge);
            }
        }

        result = (IArrayElementEdge[]) colResult
                .toArray(new IArrayElementEdge[colResult.size()]);

        return result;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.deltadebugging.ddstate.memorygraph.IGraphNode#isObjectReferenceFieldVariable()
     */
    public boolean isObjectReferenceFieldVariable() {
        boolean result = false;

        Iterator itPredecessors = this.getParents().iterator();
        while (itPredecessors.hasNext() && result == false) {
            IGraphEdge edge = (IGraphEdge) itPredecessors.next();

            if (edge instanceof IObjectFieldEdge) {
                result = true;
            }
        }

        return result;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.deltadebugging.ddstate.memorygraph.IGraphNode#getObjectReferenceFieldPredecessors()
     */
    public IObjectFieldEdge[] getObjectReferenceFieldPredecessors() {
        IObjectFieldEdge[] result = null;

        Iterator itPredecessors = this.getParents().iterator();
        Collection colResult = new Vector();

        while (itPredecessors.hasNext()) {
            IGraphEdge edge = (IGraphEdge) itPredecessors.next();

            if (edge instanceof IObjectFieldEdge) {
                colResult.add(edge);
            }
        }

        result = (IObjectFieldEdge[]) colResult
                .toArray(new IObjectFieldEdge[colResult.size()]);

        return result;
    }
}