/*
 * $Id: RootNode.java,v 1.46 2006/01/03 21:02:54 lehmann Exp $
 * 
 * Copyright (C) 2004 Karsten Lehmann
 * 
 * Software Engineering Chair, Saarland University
 * Prof. Dr. Andreas Zeller
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 *
 * Created on 26.08.2004 by
 * @author Karsten Lehmann 
 * Mail to: lehmann@cs.uni-sb.de or KarstenLehmann@web.de
 */
package org.deltadebugging.ddstate.memorygraph;

import java.util.Hashtable;
import java.util.Iterator;

import org.deltadebugging.ddcore.DeltaSet;
import org.deltadebugging.ddstate.DDStatePlugin;
import org.deltadebugging.ddstate.comparator.delta.GraphDecorator;
import org.deltadebugging.ddstate.comparator.delta.GraphEdgeDiff;
import org.deltadebugging.ddstate.comparator.delta.GraphEdgeDiffUtil;
import org.deltadebugging.ddstate.comparator.delta.GraphEdgePair;
import org.deltadebugging.ddstate.logger.DDLogger;
import org.deltadebugging.ddstate.memorygraph.delta.DeltaTreeFactory;
import org.deltadebugging.ddstate.memorygraph.exceptions.DeltaComputationException;
import org.deltadebugging.ddstate.memorygraph.exceptions.MemoryGraphExtractionException;
import org.deltadebugging.ddstate.memorygraph.type.IReferenceTypeNode;
import org.deltadebugging.ddstate.memorygraph.value.IObjectNode;
import org.deltadebugging.ddstate.memorygraph.value.IValueNode;
import org.eclipse.core.runtime.IProgressMonitor;
import org.eclipse.core.runtime.IStatus;
import org.eclipse.core.runtime.Status;
import org.eclipse.core.runtime.SubProgressMonitor;
import org.eclipse.debug.core.DebugException;
import org.eclipse.jdt.debug.core.IJavaObject;
import org.eclipse.jdt.debug.core.IJavaType;
import org.eclipse.jdt.debug.core.IJavaValue;
import org.eclipse.jdt.debug.core.IJavaVariable;

/**
 * The root node is the anchor of the memory graph. He has no predecessors only
 * successors. The successors of the root node are the values of the base
 * variables of the actuals memory state. The base variables includes all local
 * variables in the different stack frames however the class objects and the
 * <tt>this</tt> objects. <br>
 * 
 * To compare two memory graphs you have to begin with the two root nodes of the
 * graphs. Two root nodes are per definition equal.
 * 
 * @author <a href="mailto:KarstenLehmann@web.de">Karsten Lehmann </a>
 */
public class RootNode extends GraphNode implements IRootNode {

	/** Personal logger for this class */
	private final static DDLogger logger = DDStatePlugin
			.getLogger(RootNode.class);

	/** Name for the root node */
	public final static String ROOT_NODE_NAME = "<<ROOT>>";

	/**
	 * A hashtable wich contains references to all found non primitive object
	 * for the memory graph. We need this hashtable to avoid creating two nodes
	 * of the same object.
	 */
	private final Hashtable fObjectReferences = new Hashtable();

	/**
	 * Remember if the graph is the graph for the passing or the failing Run.
	 */
	private boolean fIsPassingGraph = true;

	/**
	 * Set the unique id for every delta debuggin run
	 */
	private int fDDRunID = -1;

	/**
	 * Some interesting statistics about this memory graph.
	 */
	private Statistics fStatistics = new Statistics();

	private RootNode() {
		super(null);

		// set the rootNode to itself
		this.fRootNode = this;

		this.fStatistics.numberOfGraphNodes = 1;
	}

	public RootNode(boolean isPassingGraph, int ddRunID) {
		this();

		this.fIsPassingGraph = isPassingGraph;
		this.fDDRunID = ddRunID;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.memorygraph.IRootNode#isPassingGraph()
	 */
	public boolean isPassingGraph() {
		return this.fIsPassingGraph;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.memorygraph.IRootNode#getDDRunID()
	 */
	public int getDDRunID() {
		return this.fDDRunID;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.memorygraph.GraphNode#createSuccessors()
	 */
	protected void createChildren() {
		throw new IllegalAccessError(
				"This method is not supported by the Root node.");
	}

	/**
	 * Register a ObjectNode in the object reference map with the value of the
	 * Graphnode as key.
	 * 
	 * @param graphNode
	 *            the Graphnode to register.
	 * @throws IllegalArgumentException
	 *             if the Graphnode is null.
	 */
	public void registerObjectNode(IObjectNode graphNode) {

		if (graphNode == null)
			throw new IllegalArgumentException(
					"The graph node must not be null.");

		IJavaValue value = graphNode.getValue();

		logger.trace(DDStatePlugin.DEBUG_GRAPH_EXTRACTOR,
				"Try to add object reference (" + value + ") to map (key = "
						+ value + ")");
		this.fObjectReferences.put(value, graphNode);
	}

	public boolean existObjectNode(IJavaObject objectReference) {
		boolean result = false;

		if (objectReference != null)
			result = this.fObjectReferences.containsKey(objectReference);

		return result;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.memorygraph.IRootNode#getObjectNode(org.eclipse.jdt.debug.core.IJavaObject)
	 */
	public IObjectNode getObjectNode(IJavaObject objectReference) {
		if (objectReference == null)
			throw new IllegalArgumentException("Given key must not be null");

		return (IObjectNode) this.fObjectReferences.get(objectReference);

	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.memorygraph.IGraphNode#unfoldSubgraph(IProgressMonitor
	 *      monitor)
	 */
	public void unfoldSubgraph(IProgressMonitor monitor)
			throws MemoryGraphExtractionException {
		String graphSpec = this.fIsPassingGraph ? "passing" : "failing";

		// overide only to add the debug output
		String key = "unfoldSubgraph" + this.fDDRunID + graphSpec;
		logger.traceAndTimeStart(DDStatePlugin.DEBUG, key, "Unfold "
				+ graphSpec + " graph (DDRunID = " + this.fDDRunID + ") ...");

		SubProgressMonitor subMonitor = new SubProgressMonitor(monitor, 1);
		subMonitor.beginTask("Unfold " + graphSpec + " Graph",
				IProgressMonitor.UNKNOWN);

		super.unfoldSubgraph(monitor);
		subMonitor.worked(IProgressMonitor.UNKNOWN);

		if (subMonitor.isCanceled()) {
			logger.traceAndTimeEnd(DDStatePlugin.DEBUG, key, "Unfolding "
					+ graphSpec + " graph canceld");
		} else {
			logger.traceAndTimeEnd(DDStatePlugin.DEBUG, key, graphSpec
					+ " graph unfolded");
		}
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.memorygraph.IGraphElement#computeDeltaSet(org.deltadebugging.ddstate.memorygraph.IGraphElement)
	 */
	public DeltaSet computeDeltaSet(IGraphElement element)
			throws DeltaComputationException {

		if (!this.isMaskedAs(GraphDecorator.NOT_COMPARED)) {
			return new DeltaSet();
		}

		if (element == null) {
			throw new NullPointerException("Given element is null");
		}

		assert element instanceof IRootNode;

		// Root nodes are per definition equal
		GraphDecorator.decorateGraphElementsAsEqual(this, element);

		DeltaSet result = new DeltaSet();

		GraphEdgeDiff diff = GraphEdgeDiffUtil.computeDiff(this,
				(IRootNode) element);

		if (diff.getEdgesOnlyInFirstGraph().size() > 0) {
			for (Iterator itNewEdges = diff.getEdgesOnlyInFirstGraph()
					.iterator(); itNewEdges.hasNext();) {
				IGraphEdge edge = (IGraphEdge) itNewEdges.next();
				GraphDecorator.decorateGraphEdgeAs(edge,
						GraphDecorator.ONLY_IN_FIRSTGRAPH, true);

				IStatus status = new Status(IStatus.INFO,
						DDStatePlugin.PLUGIN_ID, IStatus.INFO,
						"The edge with the name '"
								+ edge.getName()
								+ "' exists only in the "
								+ (this.isPassingGraph() ? "passing"
										: "failing") + " graph.", null);
				logger.log(status);
			}
		}

		if (diff.getEdgesOnlyInSecondGraph().size() > 0) {
			for (Iterator itNewEdges = diff.getEdgesOnlyInSecondGraph()
					.iterator(); itNewEdges.hasNext();) {
				IGraphEdge edge = (IGraphEdge) itNewEdges.next();
				GraphDecorator.decorateGraphEdgeAs(edge,
						GraphDecorator.ONLY_IN_SECONDGRAPH, true);

				IStatus status = new Status(
						IStatus.INFO,
						DDStatePlugin.PLUGIN_ID,
						IStatus.INFO,
						"The edge with the name '"
								+ edge.getName()
								+ "' exists only in the "
								+ (((IRootNode) element).isPassingGraph() ? "passing"
										: "failing") + " graph.", null);
				logger.log(status);
			}
		}

		try {
			for (Iterator itEdgePairs = diff.getEdgesInBothGraph().iterator(); itEdgePairs
					.hasNext();) {
				GraphEdgePair pair = (GraphEdgePair) itEdgePairs.next();
				result.addAll(pair.getFirstEdge().computeDeltaSet(
						pair.getSecondEdge()));
			}

		} finally {
			// clear the cache of the DeltaTreeFactory
			DeltaTreeFactory.clearCache(this.fDDRunID);
		}

		return result;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.memorygraph.IGraphElement#graphEquals(org.deltadebugging.ddstate.memorygraph.IGraphElement)
	 */
	public boolean graphEquals(IGraphElement element) {
		if (!(element instanceof IRootNode)) {
			return false;
		}

		return true;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.memorygraph.IRootNode#addLocalVariable(org.eclipse.jdt.debug.core.IJavaVariable,
	 *      org.deltadebugging.ddstate.memorygraph.IStackFrame)
	 */
	public void addLocalVariable(IJavaVariable variable, IStackFrame stackFrame)
			throws MemoryGraphExtractionException {
		String successorName = null;
		try {
			successorName = variable.getName();
			
		} catch (DebugException e) {
			throw new MemoryGraphExtractionException(
					"Cannot retrieve the name of the local variable", e);
		}

		logger.trace(DDStatePlugin.DEBUG_GRAPH_EXTRACTOR,
				"Add local variable successor to RootNode\n"
						+ "\t- stackframe    = '" + stackFrame + "'\n"
						+ "\t- variable name = '" + successorName + "'");

		IGraphNode successor = this.createAndRegisterNode(variable);

		assert successor instanceof IValueNode;

		IGraphEdge newSuccessor = null;
		try {
			newSuccessor = GraphElementUtil.createLocalVarEdge(this.fRootNode,
					(IValueNode) successor, stackFrame, variable.getName());
		} catch (DebugException e) {
			throw new MemoryGraphExtractionException(
					"Cannot read the variable name", e);
		}

		this.addSuccessor(newSuccessor);
		successor.addParent(newSuccessor);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.memorygraph.IRootNode#addType(org.eclipse.jdt.debug.core.IJavaType)
	 */
	public void addType(IJavaType type) throws MemoryGraphExtractionException {

		try {
			logger.trace(DDStatePlugin.DEBUG_GRAPH_EXTRACTOR,
					"Add type successor to RootNode\n" + "\t- type name = '"
							+ type.getName() + "'");
		} catch (DebugException doNothing) {
			// do nothing
		}

		IGraphNode typeNode = null;

		typeNode = GraphElementUtil.createTypeNode(this.fRootNode, type);

		assert typeNode instanceof IReferenceTypeNode;

		IGraphEdge newSuccessor = GraphElementUtil.createTypeEdge(this,
				(IReferenceTypeNode) typeNode);

		this.addSuccessor(newSuccessor);
		typeNode.addParent(newSuccessor);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.memorygraph.IRootNode#getStatistics()
	 */
	public Statistics getStatistics() {
		return this.fStatistics;
	}
}