/*
 * $Id: UpdateArraySizeDelta.java,v 1.9 2006/01/24 19:43:55 lehmann Exp $
 * 
 * Copyright (C) 2005 Karsten Lehmann
 * 
 * Software Engineering Chair, Saarland University
 * Prof. Dr. Andreas Zeller
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 *
 * Created on 01.08.2005 10:15:26 by
 * @author Karsten Lehmann 
 * Mail to: lehmann@cs.uni-sb.de or KarstenLehmann@web.de
 */
package org.deltadebugging.ddstate.memorygraph.delta.update;

import java.util.Collection;
import java.util.Iterator;
import java.util.Vector;

import org.deltadebugging.ddstate.launcher.client.ITestRunner;
import org.deltadebugging.ddstate.memorygraph.IGraphEdge;
import org.deltadebugging.ddstate.memorygraph.delta.IMemoryGraphDelta;
import org.deltadebugging.ddstate.memorygraph.delta.IUpdateDelta;
import org.deltadebugging.ddstate.memorygraph.delta.location.IDeltaLocation;
import org.deltadebugging.ddstate.memorygraph.delta.location.PathLocation;
import org.deltadebugging.ddstate.memorygraph.exceptions.DeltaExecutionException;
import org.deltadebugging.ddstate.memorygraph.util.MemoryGraphUtil;
import org.deltadebugging.ddstate.memorygraph.value.IArrayNode;
import org.eclipse.debug.core.DebugException;
import org.eclipse.jdt.debug.core.IJavaArray;
import org.eclipse.jdt.debug.core.IJavaArrayType;
import org.eclipse.jdt.debug.core.IJavaThread;
import org.eclipse.jdt.debug.core.IJavaVariable;

/**
 * This delta update the array size of an array in the memory.
 * 
 * @author <a href="mailto:KarstenLehmann@web.de">Karsten Lehmann </a>
 */
public class UpdateArraySizeDelta extends UpdateObjectValueDelta implements
        IUpdateDelta {

    private Collection fLocationsToChange = new Vector();

    public UpdateArraySizeDelta(IArrayNode oldValue, IArrayNode newValue) {
        super(oldValue, newValue);

        // compute the all paths to this array that have to changes
        Iterator itPred = oldValue.getParents().iterator();
        while (itPred.hasNext()) {
            IGraphEdge graphEdge = (IGraphEdge) itPred.next();
            this.fLocationsToChange.add(new PathLocation(MemoryGraphUtil
                    .searchPathToRoot(graphEdge, false)));
        }
    }

    /**
     * Returns the length of the old array node.
     * 
     * @return the length of the old array node.
     */
    public int getOldArrayLength() {
        int result = 0;

        IArrayNode arrayNode = (IArrayNode) this.getOldValue();

        if (arrayNode != null)
            result = arrayNode.getLength();

        return result;
    }

    /**
     * Returns the length of the new array node.
     * 
     * @return the length of the new array node.
     */
    public int getNewArrayLength() {
        int result = 0;

        IArrayNode arrayNode = (IArrayNode) this.getNewValue();

        if (arrayNode != null)
            result = arrayNode.getLength();

        return result;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.deltadebugging.ddstate.memorygraph.delta.IMemoryGraphDelta#apply(org.eclipse.jdt.debug.core.IJavaThread,
     *      org.deltadebugging.ddstate.launcher.client.ITestRunner)
     */
    public void apply(IJavaThread thread, ITestRunner testRunner)
            throws DeltaExecutionException {

        IJavaVariable oldArrayVar = this.getLocation().getVariableToLocation(
                thread, testRunner);
        IJavaArray oldArrayValue = null;
        try {
            oldArrayValue = (IJavaArray) oldArrayVar.getValue();
        } catch (DebugException e) {
            throw new DeltaExecutionException(
                    "Cannot retrieve the value of the old array", e);
        }

        IArrayNode newArray = (IArrayNode) this.getNewValue();

        IJavaArray newArrayValue = null;
        try {
            newArrayValue = this.getNewArrayInstance(oldArrayValue, newArray
                    .getLength());
        } catch (DebugException e) {
            throw new DeltaExecutionException("Cannot create new array", e);
        }

        // Copy the array elements of the old array to the new one
        try {
            this.copyValues(oldArrayValue, newArrayValue);
        } catch (DebugException e) {
            throw new DeltaExecutionException(
                    "Cannot copy the elements from the old array value to the new one",
                    e);
        }

        Iterator itLocToChange = this.fLocationsToChange.iterator();

        // Replace the old array with the new one
        try {
            while (itLocToChange.hasNext()) {
                IDeltaLocation location = (IDeltaLocation) itLocToChange.next();
                IJavaVariable varToChange = location.getVariableToLocation(
                        thread, testRunner);
                varToChange.setValue(newArrayValue);
            }
        } catch (DebugException e) {
            throw new DeltaExecutionException("Cannot replace the old array", e);
        }
    }

    /**
     * Create a new array with the given size from the same type of the old
     * array. If the value of old array is <code>null</code> or the value has
     * not the recommended <code>IJavaArrayType</code> as type this functions
     * returns <code>null</code>.
     * 
     * @param oldArray
     *            the java value of the old array.
     * @param size
     *            the size of the new array.
     * @return the new created array (as java value).
     * @throws DebugException
     *             if this method fails. Reasons include:
     *             <ul>
     *             <li>Failure communicating with the VM. The DebugException's
     *             status code contains the underlying exception responsible for
     *             the failure.</li>
     *             </ul>
     */
    private IJavaArray getNewArrayInstance(IJavaArray oldArray, int size)
            throws DebugException {

        IJavaArray result = null;

        if (oldArray != null) {
            IJavaArrayType type = (IJavaArrayType) oldArray.getJavaType();
            if (type != null) {
                result = type.newInstance(size);
            }
        }

        return result;
    }

    /**
     * Fill the target array with the values of the source array. Copy until you
     * reached either the limit of the old array or the limit of the new array.
     * 
     * @param sourceArray
     *            the source array
     * @param targetArray
     *            the target array
     * @throws DebugException
     *             if this method fails. Reasons include:
     *             <ul>
     *             <li>Failure communicating with the VM. The DebugException's
     *             status code contains the underlying exception responsible for
     *             the failure.</li>
     *             </ul>
     */
    private void copyValues(IJavaArray sourceArray, IJavaArray targetArray)
            throws DebugException {
        IArrayNode arrOldValue = (IArrayNode) this.getOldValue();

        IArrayNode arrNewValue = (IArrayNode) this.getNewValue();

        int limit = Math.min(arrOldValue.getLength(), arrNewValue.getLength());

        for (int i = 0; i < limit; i++) {
            targetArray.setValue(i, sourceArray.getValue(i));
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.deltadebugging.ddstate.memorygraph.delta.IMemoryGraphDelta#ctsEqual(org.deltadebugging.ddstate.memorygraph.delta.IMemoryGraphDelta)
     */
    public boolean ctsEqual(IMemoryGraphDelta delta) {

        if (!(delta instanceof UpdateArraySizeDelta)) {
            return false;
        }

        return this.getLocation().equals(delta.getLocation());
    }

}
