/*
 * $Id: UpdateObjectReferenceDelta.java,v 1.9 2005/11/21 10:42:04 lehmann Exp $
 * 
 * Copyright (C) 2005 Karsten Lehmann
 * 
 * Software Engineering Chair, Saarland University
 * Prof. Dr. Andreas Zeller
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 *
 * Created on 01.08.2005 11:05:00 by
 * @author Karsten Lehmann 
 * Mail to: lehmann@cs.uni-sb.de or KarstenLehmann@web.de
 */
package org.deltadebugging.ddstate.memorygraph.delta.update;

import org.deltadebugging.ddstate.launcher.client.ITestRunner;
import org.deltadebugging.ddstate.memorygraph.IGraphEdge;
import org.deltadebugging.ddstate.memorygraph.delta.IMemoryGraphDelta;
import org.deltadebugging.ddstate.memorygraph.delta.IUpdateDelta;
import org.deltadebugging.ddstate.memorygraph.delta.MemoryGraphDelta;
import org.deltadebugging.ddstate.memorygraph.delta.create.CreateObjectDelta;
import org.deltadebugging.ddstate.memorygraph.delta.location.CreateObjectDeltaLocation;
import org.deltadebugging.ddstate.memorygraph.delta.location.IDeltaLocation;
import org.deltadebugging.ddstate.memorygraph.delta.location.PathLocation;
import org.deltadebugging.ddstate.memorygraph.exceptions.DeltaExecutionException;
import org.deltadebugging.ddstate.memorygraph.value.IObjectNode;
import org.eclipse.debug.core.DebugException;
import org.eclipse.debug.core.model.IValue;
import org.eclipse.jdt.debug.core.IJavaThread;
import org.eclipse.jdt.debug.core.IJavaVariable;

/**
 * This delta updates an object reference
 * 
 * @author <a href="mailto:KarstenLehmann@web.de">Karsten Lehmann </a>
 */
public class UpdateObjectReferenceDelta extends MemoryGraphDelta implements
		IUpdateDelta {

	/** The location where the object reference have to point to */
	private final IDeltaLocation fNewLocation;

	/**
	 * The temporary available variable for the old location.
	 */
	private IJavaVariable fOldTmpVariable = null;

	/**
	 * The temporary available variable for the new location.
	 */
	private IJavaVariable fNewTmpVariable = null;

	public UpdateObjectReferenceDelta(IGraphEdge edgeToChange,
			IObjectNode targetNode) {
		super(new PathLocation(edgeToChange, true));
		this.fNewLocation = new PathLocation(targetNode, true);
	}

	public UpdateObjectReferenceDelta(IGraphEdge edgeToChange,
			CreateObjectDelta targetDelta) {
		super(new PathLocation(edgeToChange, true));
		this.fNewLocation = new CreateObjectDeltaLocation(targetDelta);

		this.addDependeny(targetDelta);
	}

	public UpdateObjectReferenceDelta(CreateObjectDelta sourceDelta,
			IGraphEdge newEdge, IObjectNode targetObject) {
		super(new CreateObjectDeltaLocation(sourceDelta, newEdge));
		this.fNewLocation = new PathLocation(targetObject, true);

		this.addDependeny(sourceDelta);
	}

	public UpdateObjectReferenceDelta(CreateObjectDelta sourceDelta,
			IGraphEdge edgeToChange, CreateObjectDelta targetDelta) {
		super(new CreateObjectDeltaLocation(sourceDelta, edgeToChange));
		this.fNewLocation = new CreateObjectDeltaLocation(targetDelta);

		this.addDependeny(sourceDelta);
		this.addDependeny(targetDelta);
	}

	/**
	 * Load the temporary variable of the old and the new location.
	 * 
	 * @param thread
	 * @param testRunner
	 * @throws DeltaExecutionException
	 */
	public void prepare(IJavaThread thread, ITestRunner testRunner)
			throws DeltaExecutionException {

		this.fNewTmpVariable = this.fNewLocation.getVariableToLocation(thread,
				testRunner);
		this.fOldTmpVariable = this.getLocation().getVariableToLocation(thread,
				testRunner);
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.memorygraph.delta.IMemoryGraphDelta#apply(org.eclipse.jdt.debug.core.IJavaThread,
	 *      org.deltadebugging.ddstate.launcher.client.ITestRunner)
	 */
	public void apply(IJavaThread thread, ITestRunner testRunner)
			throws DeltaExecutionException {

		try {
			IValue newValue = this.fNewTmpVariable.getValue();
			this.fOldTmpVariable.setValue(newValue);

		} catch (DebugException e) {
			throw new DeltaExecutionException("Cannot apply delta", e);

		} catch (Throwable e) {
			throw new DeltaExecutionException("Cannot apply delta", e);
		}

	}

	/**
	 * Returns the new location of the object reference.
	 * 
	 * @return the new location of the object reference.
	 */
	public IDeltaLocation getNewLocation() {
		return this.fNewLocation;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see org.deltadebugging.ddstate.memorygraph.delta.IMemoryGraphDelta#ctsEqual(org.deltadebugging.ddstate.memorygraph.delta.IMemoryGraphDelta)
	 */
	public boolean ctsEqual(IMemoryGraphDelta delta) {

		if (!(delta instanceof UpdateObjectReferenceDelta)) {
			return false;
		}

		UpdateObjectReferenceDelta upObjRefDelta = (UpdateObjectReferenceDelta) delta;

		return this.getLocation().equals(upObjRefDelta.getLocation())
				&& this.getNewLocation().equals(upObjRefDelta.getNewLocation());
	}

}
