/*
 * $Id: MemoryGraphUtil.java,v 1.19 2006/01/24 19:45:33 lehmann Exp $
 * 
 * Copyright (C) 2005 Karsten Lehmann
 * 
 * Software Engineering Chair, Saarland University
 * Prof. Dr. Andreas Zeller
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 *
 * Created on 17.02.2005 13:09:05 by
 * @author Karsten Lehmann 
 * Mail to: lehmann@cs.uni-sb.de or KarstenLehmann@web.de
 */
package org.deltadebugging.ddstate.memorygraph.util;

import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import org.deltadebugging.ddstate.DDStatePlugin;
import org.deltadebugging.ddstate.comparator.GraphEdgeComparator;
import org.deltadebugging.ddstate.comparator.delta.GraphDecorator;
import org.deltadebugging.ddstate.logger.DDLogger;
import org.deltadebugging.ddstate.memorygraph.IGraphEdge;
import org.deltadebugging.ddstate.memorygraph.IGraphElement;
import org.deltadebugging.ddstate.memorygraph.IGraphNode;
import org.deltadebugging.ddstate.memorygraph.IRootNode;
import org.eclipse.core.runtime.IStatus;

/**
 * This util provides some usefull methods for the memory graph operation such
 * as finding a path to the root node from every graph node.
 * 
 * @author <a href="mailto:KarstenLehmann@web.de">Karsten Lehmann </a>
 */
public class MemoryGraphUtil {

    /** Personal logger for this class */
    final static DDLogger logger = DDStatePlugin
            .getLogger(MemoryGraphUtil.class);

    private MemoryGraphUtil() {
        // util class is not initializable
    }

    /**
     * Search a path from the root node to the given graph element. If the
     * element is an graph edge this edge is part of the path (last element in
     * the path).
     * 
     * @param element
     *            the end point of the path.
     * @param associatedNodes
     *            if true the search and therefor the result includes only
     *            element that are in both memory graphs. This means that the
     *            search only regard nodes, which have an associated element.
     * @return the path from the root node to the given graph element.
     * @throws IllegalStateException
     *             if no path was found.
     */
    public static Path searchPathToRoot(IGraphElement element,
            boolean associatedNodes) {

        Path result = null;

        if (element instanceof IGraphNode) {
            result = new Path();
            MemoryGraphUtil.searchPathToNode((IGraphNode) element, result,
                    new HashSet(), associatedNodes);

        } else if (element instanceof IGraphEdge) {
            IGraphEdge edge = (IGraphEdge) element;

            IGraphNode sourceNode = edge.getSourceNode();
            result = searchPathToRoot(sourceNode, associatedNodes);
            result.addLastGraphEdge(edge);
        }

        if (result == null
                || (result.size() == 0 && !(element instanceof IRootNode))) {
            logger.log(IStatus.ERROR, "Path for '" + element.toString()
                    + "' to the root node not found");

            throw new IllegalStateException("Path for '" + element.toString()
                    + "' to the root node not found");
        }

        return result;
    }

    /**
     * Search a path from the root node to the given graph node
     * 
     * @param node
     *            the end point of the path.
     * @param pathToNode
     *            the path to the given node. If we start the search this is an
     *            empty path.
     * @param visitedNodes
     *            a set of all node we have visited during the search.
     * @param associatedNodes
     *            if true the search includes only element that are in both
     *            memory graphs. This means that the search only regard nodes,
     *            which have an associated element.
     * @return true if the search was succesful, otherwise false.
     */
    private static boolean searchPathToNode(IGraphNode node, Path pathToNode,
            Set visitedNodes, boolean associatedNodes) {
        if (node instanceof IRootNode) {
            return true;
        }

        if (!visitedNodes.add(node)) {
            return false;
        }

        IGraphNode predeccessorNode = null;
        IGraphEdge predeccessorEdge = null;
        Iterator itPredeccessors = getSortedPredecessors(node, associatedNodes);

        while (itPredeccessors.hasNext()) {
            predeccessorEdge = (IGraphEdge) itPredeccessors.next();
            predeccessorNode = predeccessorEdge.getSourceNode();

            // look for cycles
            if (pathToNode.containsEdge(predeccessorEdge)) {
                // stop recursion
                return false;
            }

            if (associatedNodes && !predeccessorEdge.hasAssociatedElement()) {
                return false;
            }

            // else find the path to root
            pathToNode.addFirstGraphEdge(predeccessorEdge);

            boolean findRootPath = MemoryGraphUtil
                    .searchPathToNode(predeccessorNode, pathToNode,
                            visitedNodes, associatedNodes);

            if (findRootPath)
                return true;

            // else search an other path and remove this element (the first in
            // the path)
            pathToNode.removeFirstPathElement();
        }

        return false;
    }

    /**
     * Returns the predecessors of the given node sorted by the
     * GraphEdgeComparator.
     * 
     * @param node
     * @param associatedNodes
     *            if true the result contains only all predecessors that have an
     *            associated element.
     * @return the predecessors of the given node sorted by the
     *         GraphEdgeComparator.
     * @see GraphEdgeComparator
     */
    private static Iterator getSortedPredecessors(IGraphNode node,
            boolean associatedNodes) {
        Set setResult = new HashSet();

        Iterator itPredecessors = node.getParents().iterator();
        while (itPredecessors.hasNext()) {
            IGraphEdge edge = (IGraphEdge) itPredecessors.next();

            if (!associatedNodes
                    || (edge.hasAssociatedElement() && edge.getTargetNode()
                            .isAssociatedWith(
                                    ((IGraphEdge) edge.getAssociatedElement())
                                            .getTargetNode()))) {
                setResult.add(edge);
            }
        }

        return setResult.iterator();
    }

    /**
     * Count the number of graph edges and nodes reachable from the given node
     * (the result also includes the node).
     * 
     * @param node
     *            the start node for the search.
     * @return the number of graph edges and nodes reachable from the given
     *         node.
     */
    public static CountResult countNodeAndEdges(IGraphNode node) {
        final CountResult result = new CountResult();

        IGraphNodeVisitor nodeVisitor = new IGraphNodeVisitor() {
            public void graphNodeVisited(IGraphNode graphNode) {
                result.numberOfNodes++;
            }
        };

        IGraphEdgeVisitor edgeVisitor = new IGraphEdgeVisitor() {
            public void graphEdgeVisited(IGraphEdge graphEdge) {
                result.numberOfEdges++;
            }
        };

        MemoryGraphUtil.traverseGraph(node, nodeVisitor, edgeVisitor);

        return result;
    }

    /**
     * Starting from the given node this method reset all informations in the
     * graph set on comparing two memory graphs.
     * 
     * @param node
     *            the start node.
     */
    public static void resetCompareInformatios(IGraphNode node) {

        node.getRootNode().getStatistics().numberOfEqualEdges = 0;
        node.getRootNode().getStatistics().numberOfEqualNodes = 0;

        IGraphNodeVisitor nodeVisitor = new IGraphNodeVisitor() {
            public void graphNodeVisited(IGraphNode graphNode) {
                graphNode.resetAssociatedElement();
                graphNode.maskAs(GraphDecorator.NOT_COMPARED);
            }
        };

        IGraphEdgeVisitor edgeVisitor = new IGraphEdgeVisitor() {
            public void graphEdgeVisited(IGraphEdge graphEdge) {
                graphEdge.resetAssociatedElement();
                graphEdge.maskAs(GraphDecorator.NOT_COMPARED);
            }
        };

        MemoryGraphUtil.traverseGraph(node, nodeVisitor, edgeVisitor);
    }

    /**
     * Check if all elements in the graph are compared. This means that all
     * elements are not masked as <code>GraphDecorator.NOT_COMPARED</code>.
     * 
     * @param rootNode
     * @return true if all elements in the graph are compared, otherwise false.
     * 
     * @see GraphDecorator#NOT_COMPARED
     * @see org.deltadebugging.ddstate.comparator.delta.IDeltaComparable
     */
    public static boolean assertAllElementsAreCompared(IRootNode rootNode) {
        boolean result = true;

        final Set notComparedElements = new HashSet();

        IGraphNodeVisitor nodeVisitor = new IGraphNodeVisitor() {
            public void graphNodeVisited(IGraphNode graphNode) {
                if (graphNode.isMaskedAs(GraphDecorator.NOT_COMPARED)) {
                    notComparedElements.add(graphNode);
                }
            }
        };

        IGraphEdgeVisitor edgeVisitor = new IGraphEdgeVisitor() {
            public void graphEdgeVisited(IGraphEdge graphEdge) {
                if (graphEdge.isMaskedAs(GraphDecorator.NOT_COMPARED)) {
                    notComparedElements.add(graphEdge);
                }
            }
        };

        MemoryGraphUtil.traverseGraph(rootNode, nodeVisitor, edgeVisitor);

        result = notComparedElements.size() == 0;

        return result;
    }

    /**
     * Traverse the full graph starting from the given node. Everytime a node is
     * visited the nodeVisitor will notified, if a graph edge was visited the
     * edgeVisitor will notified.
     * 
     * @param node
     *            the start node.
     * @param nodeVisitor
     * @param edgeVisitor
     */
    public static void traverseGraph(IGraphNode node,
            IGraphNodeVisitor nodeVisitor, IGraphEdgeVisitor edgeVisitor) {
        MemoryGraphUtil.traverseGraph(node, nodeVisitor, edgeVisitor,
                new HashSet());
    }

    private static void traverseGraph(IGraphNode node,
            IGraphNodeVisitor nodeVisitor, IGraphEdgeVisitor edgeVisitor,
            Set visitedNodes) {

        if (!visitedNodes.add(node)) {
            return;
        }

        nodeVisitor.graphNodeVisited(node);

        IGraphNode successorNode = null;
        IGraphEdge successorEdge = null;

        Iterator itSuccessors = node.getChildren().iterator();

        while (itSuccessors.hasNext()) {
            successorEdge = (IGraphEdge) itSuccessors.next();
            edgeVisitor.graphEdgeVisited(successorEdge);

            successorNode = successorEdge.getTargetNode();

            traverseGraph(successorNode, nodeVisitor, edgeVisitor, visitedNodes);
        }
    }

    public interface IGraphNodeVisitor {
        public void graphNodeVisited(IGraphNode graphNode);
    }

    public interface IGraphEdgeVisitor {
        public void graphEdgeVisited(IGraphEdge graphEdge);
    }

    public static class CountResult {

        /**
         * The number of graph nodes.
         */
        public long numberOfNodes = 0;

        /**
         * The number of graph edges
         */
        public long numberOfEdges = 0;
    }

}
