/*
 * $Id: Path.java,v 1.7 2005/08/24 19:03:21 lehmann Exp $
 * 
 * Copyright (C) 2005 Karsten Lehmann
 * 
 * Software Engineering Chair, Saarland University
 * Prof. Dr. Andreas Zeller
 *
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Common Public License v1.0
 * which accompanies this distribution, and is available at
 * http://www.eclipse.org/legal/cpl-v10.html
 *
 * Created on 15.03.2005 14:03:44 by
 * @author Karsten Lehmann 
 * Mail to: lehmann@cs.uni-sb.de or KarstenLehmann@web.de
 */
package org.deltadebugging.ddstate.memorygraph.util;

import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;

import org.deltadebugging.ddstate.memorygraph.GraphEdge;
import org.deltadebugging.ddstate.memorygraph.IGraphEdge;

/**
 * The path is a path between two nodes in the memory graph. The path describes
 * a list you must pass to go from the source node to the destination node. This
 * path is not immutable. All operations on this path have direct impact to this
 * graph.
 * 
 * @author <a href="mailto:KarstenLehmann@web.de">Karsten Lehmann </a>
 */
public class Path {

	private final LinkedList fEdgeList;

	/**
	 * Create a new empty path.
	 */
	public Path() {
		this.fEdgeList = new LinkedList();
	}

	/**
	 * Create this path as a copy of the given path.
	 * 
	 * @param path
	 */
	public Path(Path path) {
		this.fEdgeList = new LinkedList();

		for (Iterator itGraphEdge = path.graphEdges(); itGraphEdge.hasNext();) {
			IGraphEdge edge = (IGraphEdge) itGraphEdge.next();

			this.fEdgeList.addLast(edge);
		}
	}

	/**
	 * Create this path of the given list of edges.
	 * 
	 * @param edgeList
	 *            a list of graph edges.
	 */
	public Path(List edgeList) {
		this.fEdgeList = new LinkedList(edgeList);
	}

	/**
	 * Inserts the given edge at the beginning of the path. If the path always
	 * contains elements the target node of the given edge must equal to the
	 * source node of the first edge in the path.
	 * 
	 * @param edge
	 *            the edge to be inserted at the beginning of the path.
	 */
	public void addFirstGraphEdge(IGraphEdge edge) {
		if (edge == null) {
			throw new NullPointerException("Given edge must not be null");
		}

		// Check that the given edge is a real predecessor of the first edge
		if (!this.isEmpty()) {
			IGraphEdge firstEdge = this.getFirstPathElement();
			if (!edge.getTargetNode().equals(firstEdge.getSourceNode())) {
				throw new IllegalArgumentException(
						"The target node of the given edge "
								+ "must equal to the source node "
								+ "of the first edge in the path.");
			}
		}

		this.fEdgeList.addFirst(edge);
	}

	/**
	 * Inserts the given edge at the end of the path. If the path always
	 * contains elements the source node of the given edge must equal to the
	 * target node of the last edge in the path.
	 * 
	 * @param edge
	 *            the edge to be inserted at the end of the path.
	 */
	public void addLastGraphEdge(IGraphEdge edge) {
		if (edge == null) {
			throw new NullPointerException("Given edge must not be null");
		}

		// Check that the given edge is a real successor of the last edge
		if (!this.isEmpty()) {
			IGraphEdge lastEdge = this.getLastPathElement();
			if (!edge.getSourceNode().equals(lastEdge.getTargetNode())) {
				throw new IllegalArgumentException(
						"The source node of the given edge "
								+ "must equal to the target node "
								+ "of the last edge in the path.");
			}
		}

		this.fEdgeList.addLast(edge);
	}

	/**
	 * Returns <tt>true</tt> if this path contains the specified graph edges.
	 * More formally, returns <tt>true</tt> if and only if this path contains
	 * at least one graph edge <tt>e</tt> such that <tt>(edge==null ? e==null
	 * : edge.equals(e))</tt>.
	 * 
	 * @param edge
	 *            graph edge whose presence in this path is to be tested.
	 * @return <tt>true</tt> if this path contains the specified graph edge.
	 */
	public boolean containsEdge(IGraphEdge edge) {

		return this.fEdgeList.contains(edge);
	}

	/**
	 * Removes and returns the first graph edge from this path.
	 * 
	 * @return the first graph edge from this path.
	 * @throws NoSuchElementException
	 *             if this path is empty.
	 */
	public IGraphEdge removeFirstPathElement() {
		return (IGraphEdge) this.fEdgeList.removeFirst();
	}

	/**
	 * Returns the first graph edge in this path.
	 * 
	 * @return the first graph edge in this path.
	 * @throws NoSuchElementException
	 *             if this path is empty.
	 */
	public IGraphEdge getFirstPathElement() {
		return (IGraphEdge) this.fEdgeList.getFirst();
	}

	/**
	 * Returns the last graph edge in this path.
	 * 
	 * @return the last graph edge in this path.
	 * @throws NoSuchElementException
	 *             if this path is empty.
	 */
	public IGraphEdge getLastPathElement() {
		return (IGraphEdge) this.fEdgeList.getLast();
	}

	/**
	 * Returns an iterator over the graph edges in this path (in proper
	 * sequence).
	 * <p>
	 * 
	 * This implementation merely returns a list iterator over the path.
	 * 
	 * @return an iterator over the graph edges in this path (in proper
	 *         sequence).
	 */
	public Iterator graphEdges() {
		return this.fEdgeList.iterator();
	}

	/**
	 * Returns the number of graph edges in this path.
	 * 
	 * @return the number of graph edges in this path.
	 */
	public int size() {
		return this.fEdgeList.size();
	}

	/**
	 * Returns the graph edge at the specified position in this path.
	 * 
	 * @param index
	 *            index of graph edge to return.
	 * @return the graph edge at the specified position in this path.
	 * @throws IndexOutOfBoundsException
	 *             if the specified index is is out of range (<tt>index &lt; 0 || index &gt;= size()</tt>).
	 */
	public IGraphEdge get(int index) {
		return (IGraphEdge) this.fEdgeList.get(index);
	}

	/**
	 * Returns <tt>true</tt> if this path contains no graph edges.
	 * <p>
	 * 
	 * This implementation returns <tt>size() == 0</tt>.
	 * 
	 * @return <tt>true</tt> if this path contains no graph.
	 */
	public boolean isEmpty() {
		return this.fEdgeList.isEmpty();
	}

	/**
	 * Returns this path as list (the graph edges in proper sequence).
	 * 
	 * @return this path as list (the graph edges in proper sequence).
	 */
	public List getAsList() {
		return this.fEdgeList;
	}

	/*
	 * (non-Javadoc)
	 * 
	 * @see java.lang.Object#toString()
	 */
	public String toString() {
		StringBuffer result = null;

		if (!this.fEdgeList.isEmpty()) {
			result = new StringBuffer();

			for (int i = 0; i < this.fEdgeList.size(); i++) {
				GraphEdge edge = (GraphEdge) this.fEdgeList.get(i);
				result.append(edge.toString());
				if (i < this.fEdgeList.size() - 1) {
					result.append("\n");
				}
			}
		} else {
			result = new StringBuffer("Empty Path");
		}

		return result.toString();
	}

}